
/*
Created by: Peeyush
Created on: September 16 2021
Email: j.k.peeyush@gmail.com
Description: Connect to DB and provide utility fuctions
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
"use strict";

/*
Function to create model instance.
Returns 'express-cassandra' client instance
*/
const createInstance = (cassandra, ContactPoints, LocalDataCenter, Port, Keyspace, requestTracker,user,pwd) => {
    const client = new cassandra.Client({
        applicationName: 'PRASI LABS',
        queryOptions: {
            autoPage: true,
            prepare: true,
            consistency: cassandra.types.consistencies.localOne
        },
        encoding: {
            set: Set
        },
        profiles: [
            new cassandra.ExecutionProfile('oltp', {
                consistency: cassandra.types.consistencies.localOne
            })
        ],
        requestTracker: requestTracker,
        contactPoints: ContactPoints,
        localDataCenter: LocalDataCenter,
        protocolOptions: { port: Port },
        keyspace: Keyspace,
        credentials: {
            username: user,
            password: pwd
        }
    });
    return client;
};

//Check databse connectivity
const queryDummyData = async (client, mapper, options) => {
    return client.connect()
        .then(function () {
            return client.execute(options.cql, options.quryHint);
        }).then(function (result) {
            const row = result.rows[0];
            return row;
        }).then(function (row) {
            return row;
        }).catch((err) => {
            return err;
        })
}

/*
Function to query data from the database
*/
const queryData = async (client, mapper, options, args) => {
    return client.connect()
        .then(function () {
            return client.execute(options.cql, args, options.quryHint);
        }).then(function (result) {
            return result.rows;
        }).catch((err) => {
            return err;
        })
}

/*
Function to eecute queries for write operation (update insert delete) to the database
*/
const writeData = async (client, mapper, options, args) => {
    return client.connect()
        .then(function () {
            return client.execute(options.cql, args, options.quryHint);
        }).then(function (result) {
            return result;
        }).catch((err) => {
            return err;
        })
}

/*
Function to query data from the database using mapper function
*/
const findData = async (objMapper, filterObject, selectObject, optionObject) => {
    return objMapper.find(filterObject, selectObject, optionObject)
        .then(function (result) {
            return result
        }).catch((err) => {
            return err;
        })
}

/*
Function to query single document from the database using mapper function
*/
const findSingleData = async (objMapper, filterObject, selectObject, optionObject) => {
    return objMapper.get(filterObject, selectObject, optionObject)
        .then(function (result) {
            selectObject, optionObject
            return result
        }).catch((err) => {
            return err;
        })
}

/*
Function to query all data from the database using mapper function
*/
const findAllData = async (objMapper, filterObject, selectObject, optionObject) => {
    return objMapper.findAll(selectObject, optionObject)
        .then(function (result) {
            return result
        }).catch((err) => {
            return err;
        })
}

/*
Function to insert data to the database using mapper function
*/
const insertData = async (objMapper, filterObject, selectObject, optionObject) => {
    return objMapper.insert(selectObject, optionObject)
        .then(function (result) {
            return result
        }).catch((err) => {
            return err;
        })
}

/*
Function to insert data to the database using mapper function
*/
const insertDataIfNotPresent = async (objMapper, filterObject, selectObject, optionObject) => {

    optionObject.ifNotExists = true;
    return objMapper.insert(selectObject, optionObject)
        .then(function (result) {
            return result
        }).catch((err) => {
            return err;
        })
}

/*
Function to update data into the database using mapper function
*/

const updateData = async (objMapper, filterObject, selectObject, optionObject) => {
    selectObject.ifExists = true
    return objMapper.update(filterObject, selectObject, optionObject)
        .then(function (result) {
            return result
        }).catch((err) => {
            return err;
        })
}

/*
Function to upsert data into the database using mapper function
*/
const upsertData = async (objMapper, filterObject, selectObject, optionObject) => {
    selectObject.ifExists = false
    return objMapper.update(filterObject, selectObject, optionObject)
        .then(function (result) {
            return result
        }).catch((err) => {
            return err;
        })
}

/*
Function to delete socument from the database using mapper function
*/
const deleteData = async (objMapper, filterObject, selectObject, optionObject) => {
    return objMapper.remove(filterObject, selectObject)
        .then(function (result) {
            return result
        }).catch((err) => {
            return err;
        })
}


/*
Object mapper with query to get record from id
*/
class mapperFunc {
    #mapper = {}
    constructor(mapper, table) {
        this.#mapper = mapper
        this.#mapper.getResultC = mapper.mapWithQuery(`SELECT * FROM ${table} WHERE id=?`, objArg => [objArg.id]);

    }
    async getResults(id) {
        return this.#mapper.getResultC({ id }).then(
            result => { return result }
        ).catch(err => { return err })
    }
}

exports.createInstance = createInstance;
exports.queryDummyData = queryDummyData;
exports.queryData = queryData;
exports.writeData = writeData;
exports.findData = findData;
exports.findSingleData = findSingleData;
exports.findAllData = findAllData;
exports.insertData = insertData;
exports.updateData = updateData;
exports.deleteData = deleteData;
exports.upsertData = upsertData;
exports.insertDataIfNotPresent = insertDataIfNotPresent;
exports.mapperFunc = mapperFunc;