/*
Created by: Peeyush
Created on: July 16 2021
Email: j.k.peeyush@gmail.com
Description: declares all global vaiables
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
"use strict";
const path = require('path');
const { callbackify } = require('util');
//import database functions
const DBredirects = require(path.resolve('middleware/', './DBredirect'));
const system_user = 'SYSTEM_USER'
const audit_all_fields = 'ALL_FIELDS'
const { accessSync, constants, readFileSync } = require('fs')


/*************************************
 *************************************
 * 
 * Local functions
 * 
 *************************************
 * ***********************************/
const createInvoice = require('./invoicecreate');
const updateInvoice = require('./invoiceupdate')
/*************************************
 *************************************
 * global operations
 * 
 *************************************
 * ***********************************/
// Get Cassandra client
exports.connectDB = function (cassandra, host, datacenter, port, keyspace, requestTracker, user, pwd) {
  try {
    if (cassandra === undefined || host === undefined ||
      datacenter === undefined ||
      port === undefined ||
      keyspace === undefined
    ) {
      return new Error("Please provide database connection information");
    }
    return DBredirects.createInstance(cassandra, host, datacenter, port, keyspace, requestTracker, user, pwd);
  } catch (err) {
    return err;
  }
};

// Check dummy connection
exports.checkDB = async function (client, mapper, options) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    return await DBredirects.queryDummyData(client, mapper, options);
  } catch (err) {
    return err;
  }
};

//Get all produts
exports.getAllProducts = async function (client, mapper, options, param, body, headers) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    options.cql = `SELECT ${mapper.global.PROD_QUERY_FIELD} FROM PRODUCT;`
    return (await DBredirects.queryData(client, mapper, options, [])).filter(prod => { return prod.status == 'ACTIVE' });

  } catch (err) {
    return err;
  }
}


//Get all products be search term
exports.getProductsBySearch = async function (client, mapper, options, param, body, headers) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    //  options.cql = `SELECT ${mapper.global.PROD_QUERY_FIELD} FROM PRODUCT WHERE STATUS=?;`
    return (await DBredirects.findAllData(mapper.mapper.forModel('Product'), {}, {}, { consistency: options.consistency }))
      ._rs.rows.filter(i => { return i.status == 'ACTIVE' })
      .filter(obj => {
        return obj.productname.toUpperCase() == param.search?.toUpperCase()
          || obj.casno.toUpperCase() == param.search?.toUpperCase()
          || obj.type.toUpperCase() == param.search?.toUpperCase()
          || obj.chemicalname.toUpperCase() == param.search?.toUpperCase()
          || obj.makecountry.toUpperCase() == mapper.global.CLIST.filter(c => {
            return (c.Name.toUpperCase() == param.search?.toUpperCase() || c.ISO2.toUpperCase() == param.search?.toUpperCase())
          })[0]?.ISO2.toUpperCase()
          || obj.makecompany.toUpperCase() == param.search?.toUpperCase()
          || obj.type.toUpperCase() == param.category?.toUpperCase()
          ;
      });
  } catch (err) {
    return err;
  }
}

// search by name
exports.getProductsByName = async function (client, mapper, options, param, body, headers) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    // options.cql = `SELECT ${mapper.global.PROD_QUERY_FIELD} FROM PRODUCT WHERE STATUS=? AND productname=?;`
    return (await DBredirects.findAllData(mapper.mapper.forModel('Product'), {}, {}, { consistency: options.consistency }))
      ._rs.rows.filter(i => { return i.status == 'ACTIVE' }).filter(obj => {
        return obj.productname.toUpperCase() == param.name.toUpperCase()
      });
  } catch (err) {
    return err;
  }
}

//Search by country
exports.getProductsByCountry = async function (client, mapper, options, param, body, headers) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }

    // options.cql = `SELECT ${mapper.global.PROD_QUERY_FIELD} FROM PRODUCT WHERE STATUS=?;`
    return (await DBredirects.findAllData(mapper.mapper.forModel('Product'), {}, {}, { consistency: options.consistency })
    )._rs.rows.filter(i => { return i.status == 'ACTIVE' })
      .filter(obj => {
        return obj.makecountry.toUpperCase() == mapper.global.CLIST.filter(c => {
          return (c.Name.toUpperCase() == param.country.toUpperCase() || c.ISO2.toUpperCase() == param.country.toUpperCase())
        })[0]?.ISO2.toUpperCase()
          ;
      });
  } catch (err) {
    return err;
  }
}

//Search by partner
exports.getProductsByPartner = async function (client, mapper, options, param, body, headers) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }

    // options.cql = `SELECT ${mapper.global.PROD_QUERY_FIELD} FROM PRODUCT WHERE STATUS=?;`
    return (await DBredirects.findAllData(mapper.mapper.forModel('Product'), {}, {}, { consistency: options.consistency }))
      ._rs.rows.filter(i => { return i.status == 'ACTIVE' }).filter(obj => {
        return obj.makecompany.toUpperCase() == param.partner.toUpperCase()
          ;
      });
  } catch (err) {
    return err;
  }
}

//Search by CAS No
exports.getProductsByCasno = async function (client, mapper, options, param, body, headers) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    //options.cql = `SELECT ${PROD_QUERY_FIELD} FROM PRODUCT WHERE STATUS=?;`
    return (await DBredirects.findAllData(mapper.mapper.forModel('Product'), {}, {}, { consistency: options.consistency }))
      ._rs.rows.filter(i => { return i.status == 'ACTIVE' }).filter(obj => {
        return obj.casno.toUpperCase() == param.casno.toUpperCase()

      });
  } catch (err) {
    return err;
  }
}

//Search by category
exports.getProductsByCategory = async function (client, mapper, options, param, body, headers) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    //  options.cql = `SELECT ${mapper.global.PROD_QUERY_FIELD} FROM PRODUCT WHERE STATUS=?;`
    return (await DBredirects.findAllData(mapper.mapper.forModel('Product'), {}, {}, { consistency: options.consistency }))
      ._rs.rows.filter(i => { return i.status == 'ACTIVE' }).filter(obj => {
        return obj.type.toUpperCase() == param.category.toUpperCase()

      });

  } catch (err) {
    return err;
  }
}
// Funtion to return catefory list of products
exports.getCategoryList = async (client, mapper, options, param, body, headers) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    return mapper.global.LOV.category_list;
  } catch (err) {
    return err;
  }
}

//Function to get partner list
exports.getPartnerList = async (client, mapper, options, param, body, headers) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    return mapper.global.LOV.partner_list;
  } catch (err) {
    return err;
  }
}

//Function to get country list
exports.getCountryList = async (client, mapper, options, param, body, headers) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    return mapper.global.LOV.country_list;
  } catch (err) {
    return err;
  }
}

//Function to get currency list
exports.getCurrencyList = async (client, mapper, options, param, body, headers) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    return mapper.global.LOV.currency_list;
  } catch (err) {
    return err;
  }
}

//Function to get currency exchange list
exports.getXchangeList = async (client, mapper, options, param, body, headers) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }

    return mapper.global.XRATE;

  } catch (err) {
    return err;
  }
}

//Function to get gst list
exports.getGSTList = async (client, mapper, options, param, body, headers) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }

    return mapper.global.GSTMAP;

  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * Login functionlity 
 * 
 * ***********************************/
/*returns
{ "id": "",
 "userid": "",
 "lastlogin": "",
 "logintoken": "",
 "apitoken": "",
 "type": ""}
*/
exports.login = async function (client, mapper, options, param, body, headers, auditFunc) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    if (body.System.action == 'loginpassword') {
      let data = await DBredirects.findData(mapper.mapper.forModel('User'), { userid: body.System.user }, { fields: ["userid", "emailid", "id", "password", "status", "type", "accntid"] }, { consistency: options.consistency })

      //check for email
      if (data._rs.rows.length < 1) {
        data = await DBredirects.findAllData(mapper.mapper.forModel('User'), {}, { fields: ["userid", "emailid", "id", "password", "status", "type", "accntid"] }, { consistency: options.consistency })

        data._rs.rows = data._rs.rows.filter(val => {
          return val.emailid == body.System.user && val.status == 'ACTIVE'
        })

        if (
          data._rs.rows.filter(val => {
            return val.emailid == body.System.user && val.status == 'ACTIVE'
          }).length != 1
        ) {

          return new Error('Provided userid or password is wrong! It may possible that your id is not active.');

        }
      }

      if (data._rs.rows.filter(val => {
        return val.status == 'ACTIVE'
      }
      ).length == 1) {

        if (
          (await mapper.global.validateT(mapper, data._rs.rows[0].password, body.System.user, 'PASSWORD', DBredirects)).data != headers.authorization
        ) {
          return new Error('Provided userid or password is wrong! It may possible that your id is not active.');
        }
        let usr = { accntid: data._rs.rows[0].accntid, id: data._rs.rows[0].id, userid: data._rs.rows[0].userid, lastlogin: (new Date()).toISOString() }
        //callback audit
        let value = { 'login_data': JSON.stringify(usr), 'action_result': 'success' }
        auditFunc(value, 'CURRENT', 'LOGIN', 'User', audit_all_fields, mapper, DBredirects, options, system_user, 'User');
        //Generate new token
        let tokens = await mapper.global.validateT(mapper, headers.authorization, body.System.user, 'GENERATE', DBredirects);
        usr.logintoken = tokens.logintoken
        usr.apitoken = tokens.apitoken
        let d = await DBredirects.updateData(mapper.mapper.forModel('User'), usr, {}, { consistency: options.consistency })
        //console.log(d._rs.rows)

        //return to client
        usr.type = data._rs.rows[0].type

        let com = (await mapper.global.communicationEntry('LOGIN', data._rs.rows[0].accntid, body.System.user, `Logged in successfully at ${new Date().toLocaleString()}`, mapper, DBredirects, options, system_user))
        let v = {
          'new_data': JSON.stringify(com.data),
          'action_result': JSON.stringify(
            com.rs
          )
        }
        //callback audit
        auditFunc(v, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

        return usr
      }
      else {
        return new Error('Provided userid or password is wrong! It may possible that your id is not active.');
      }
    } else {
      let x = await mapper.global.validateT(mapper, headers.authorization, body.System.user, 'LOGIN', DBredirects);
      if (!x) {
        return new Error("Wrong credentials. Please try using password")
      }
      return x
    }
  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * logout functionlity 
 * 
 * ***********************************/
//returns user id
exports.logout = async function (client, mapper, options, param, body, headers, auditFunc) {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
    let d = await DBredirects.updateData(mapper.mapper.forModel('User'), { id: body.System.id, userid: body.System.user, apitoken: '0', logintoken: '0' }, { fields: ["id", "userid", "apitoken", "logintoken"] }, { consistency: options.consistency })
    //callback audit
    let value = { 'logout_data': JSON.stringify({ userid: body.System.user, password: body.System.logintoken }), 'action_result': JSON.stringify(d._rs.rows) }
    auditFunc(value, 'CURRENT', 'UPDATE', 'User', audit_all_fields, mapper, DBredirects, options, system_user, 'User');
    let com = (await mapper.global.communicationEntry('LOGOUT', userR.accntid, body.System.user, `Logged out successfully at ${new Date().toLocaleString()}`, mapper, DBredirects, options, system_user))
    let v = {
      'new_data': JSON.stringify(com.data),
      'action_result': JSON.stringify(
        com.rs
      )
    }
    //callback audit
    auditFunc(v, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
    return body.System.user
  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * Any communicationn entry
 * 
 *************************************/
exports.postCommunication = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }

    let checkEmail = await DBredirects.findAllData(mapper.mapper.forModel('Email'), {}, { fields: ['eid', 'status', 'email', 'id'] }, { consistency: options.consistency })

    let e = checkEmail._rs.rows.filter(obj => {
      return (body.Communication.accntid == obj.email) && obj.status == 'ACTIVE'
    })

    if (e.length == 1) {
      body.Communication.type.toUpperCase() == 'SUBSCRIPTION' ? e[0].subscription = true : e[0].subscription = false;
    }
    else if (e.length < 1) {
      let email = mapper.global.Schema.Email;
      email.email = body.Communication.accntid
      email.subscription = body.Communication.type.toUpperCase() == 'SUBSCRIPTION' ? true : false
      email.status = 'ACTIVE'
      let check = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings({
        data: [{
          Email: email
        }], System: {
          user: system_user
        }
      }), 'Email');

      if (!check.status) {
        return new Error("Something Went wrong")
      }
      check.data[0].Email.created = check.data[0].Email.updated
      check.data[0].Email.createdby = check.data[0].Email.updatedby
      e[0] = check.data[0].Email
    }
    await DBredirects.upsertData(mapper.mapper.forModel('Email'), e[0], {}, { consistency: options.consistency, prepare: true });
    //insert communication
    let com = (
      await mapper.global.communicationEntry(body.Communication.type, body.Communication.accntid,
        body.Communication.user, body.Communication.message
        , mapper, DBredirects, options, 'ANONYMOUS')
    )
    //callback audit
    let value = {
      'new_data': JSON.stringify(com.data),
      'action_result': JSON.stringify(
        com.rs
      )
    }
    auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')


    const SE = require('./mail');

    let x = await SE(body.Communication.accntid,
      body.Communication.type.toUpperCase() == 'FEEDBACK' ? mapper.global.ENV.FEEDBACK : mapper.global.ENV.SUBSCRIPTION
      , body.Communication.type, `
            Dear PRASI LABS,
            <br/><br/>
            ${body.Communication.message}
            <br/><br/>
            Thanks and Regards,<br/>
            ${body.Communication.accntid}
          `, mapper);

    return com.rs
  } catch (err) {
    return err;
  }

}

/*************************************
 * 
 * user functionlity 
 * 
 * ***********************************/
//Functon to get user specific communication
exports.getCommunicationList = async (client, mapper, options, param, body, headers) => {
  try {

    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    return (await DBredirects.findData(mapper.mapper.forModel('Communication'), { accntid: userR.accntid }, {}, { consistency: options.consistency, prepare: true, fetchSize: 1000 }))._rs.rows.filter(item => {
      return item.type != 'LOGIN' && item.type != 'LOGOUT'
    }).reduce((pre, curr) => {
      pre.push({ updated: curr.updated, message: curr.message, type: curr.type });
      return pre
    }, []);
  } catch (err) {
    return err;
  }
}

//Functon to get user specific SR
exports.getSRList = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
    let i
    userR.type == 'ADMIN' && body.listaccount ? i = body.listaccount : i = userR.accntid

    return (await DBredirects.findData(mapper.mapper.forModel('Servicerequest'), { accntid: i }, {}, { consistency: options.consistency, prepare: true }))._rs.rows
  } catch (err) {
    return err;
  }
}

//Functon to get user specific statistics
exports.getStat = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
    //account status
    let accountstatus = (
      await DBredirects.findData(
        mapper.mapper.forModel('Account'),
        { customerid: userR.accntid },
        {},
        { consistency: options.consistency, prepare: true }
      )
    ).first().status;

    //last login time
    let lastlogin = (
      await DBredirects.findData(mapper.mapper.forModel('User'),
        { userid: body.System.user }, {},
        { consistency: options.consistency, prepare: true })).first().lastlogin;
    //total srs
    options.cql = `SELECT count(*) FROM Servicerequest where accntid=?;`
    let totalSR = (await DBredirects.queryData(client, mapper, options, [userR.accntid]))[0].count.low;

    //total document uploaded
    let noofdoc = 0
    //calulation of total amount and succesful orders.
    let tmporders = (await DBredirects.findData(mapper.mapper.forModel('Invoice'),
      { accntid: userR.accntid }, {},
      { consistency: options.consistency, prepare: true }))._rs.rows.filter(i => {
        return (i.type == 'ORIGIONAL') && (i.status = 'NEW' || i.status == 'FINAL')
      });

    let totalCompletedOrders = tmporders.length

    let totalamount = tmporders.reduce((prev, curr) => {

      prev = prev + mapper.global.GST.calculteXrate(curr.finalamount, curr.currency, 'INR', mapper.global.XRATE)
      return prev
    }, 0)

    let totaldiscount = tmporders.reduce((prev, curr) => {
      prev = prev + curr.item.reduce((pre, cur) => {

        pre = pre + mapper.global.GST.calculteXrate(JSON.parse(cur).itemdiscount, JSON.parse(cur).currency, 'INR', mapper.global.XRATE)
        return pre
      }, 0)
      return prev
    }, 0)

    //calculation of 30 days data
    let today = new Date();
    let priorDate = new Date(new Date().setDate(today.getDate() - 30));

    let last30days = tmporders.filter(i => {
      return new Date(i.updated) > priorDate
    })

    let totalamount30 = last30days.reduce((prev, curr) => {
      prev = prev + mapper.global.GST.calculteXrate(curr.finalamount, curr.currency, 'INR', mapper.global.XRATE)
      return prev
    }, 0)

    let totaldiscount30 = last30days.reduce((prev, curr) => {
      prev = prev + curr.item.reduce((prev, curr) => {
        prev = prev + mapper.global.GST.calculteXrate(JSON.parse(curr).itemdiscount, JSON.parse(curr).currency, 'INR', mapper.global.XRATE)
        return prev
      }, 0)
      return prev
    }, 0)

    return {
      totaldiscount: Math.round(totaldiscount),
      totalamount: Math.round(totalamount),
      totalorders: totalCompletedOrders,
      accountstatus: accountstatus,
      lastlogin: lastlogin,
      totalSR: totalSR,
      noofdoc: noofdoc,
      totalamount30: Math.round(totalamount30),
      totaldiscount30: Math.round(totaldiscount30)
    }

  } catch (err) {
    return err;
  }
}

//Functon to get user specific incoice
exports.getInvoiceList = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    return (await DBredirects.findData(mapper.mapper.forModel('Invoice'), { accntid: userR.accntid }, {}, { consistency: options.consistency, prepare: true }))._rs.rows.filter(inv => {
      return (inv.status == 'NEW' || inv.status == 'FINAL') && inv.type == 'ORIGIONAL'
    })
  } catch (err) {
    return err;
  }
}

//Functon to get user specific order list
exports.getOrderList = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
    let i
    userR.type == 'ADMIN' && body.listaccount ? i = body.listaccount : i = userR.accntid
    let orderArray = (await DBredirects.findData(mapper.mapper.forModel('Orders'), { accntid: i }, {}, { consistency: options.consistency, prepare: true }))._rs.rows
    let orderItemArray = (await DBredirects.findData(mapper.mapper.forModel('Orderitem'), { accntid: i }, {}, { consistency: options.consistency, prepare: true }))._rs.rows
    let retOrderArray = []
    orderArray.forEach(order => {
      retOrderArray.push({ Orders: order })
    }
    )
    orderItemArray.filter(item => {
      retOrderArray.push({ Orderitem: item })
    })

    return retOrderArray
  } catch (err) {
    return err;
  }
}

//Functon to get user specific login detail list
exports.getLoginList = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    return (await DBredirects.findData(mapper.mapper.forModel('Communication'), { accntid: userR.accntid }, {}, { consistency: options.consistency, prepare: true }))._rs.rows.filter(item => {
      return item.type == 'LOGIN'
    });
  } catch (err) {
    return err;
  }
}

//Get list of registered account
//Admin use is also applicable
exports.getAccountList = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
    if (userR.type == 'ADMIN' && body.listaccount && body.listaccount.toUpperCase() == 'ALL') {
      options.cql = `SELECT * FROM ACCOUNT;`
      return await DBredirects.queryData(client, mapper, options, []);
    }

    return (await DBredirects.findData(mapper.mapper.forModel('Account'), { customerid: userR.accntid }, {}, { consistency: options.consistency, prepare: true }))._rs.rows;
  } catch (err) {
    return err;
  }
}

//Get list of registered contact
//Admin use is also applicable
exports.getContactList = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    let account;

    if (userR.type == 'ADMIN' && body.listaccount) {
      account = (await DBredirects.findData(mapper.mapper.forModel('Account'),
        { customerid: body.listaccount }, {}, { consistency: options.consistency, prepare: true })
      ).first();
    } else {
      account = (
        await DBredirects.findData(mapper.mapper.forModel('Account'),
          { customerid: userR.accntid }, {}, { consistency: options.consistency, prepare: true })
      ).first();

    }



    let conObj = [];

    let conlist = [];
    if (!account.contact) {
      return [{
        Account: { customerid: account.customerid },
        Contact: {},
        Email: {},
        Phone: {}
      }]
    }

    if (account.contact['PRIMARY'])
      conlist.push(account.contact['PRIMARY']);

    if (account.contact['BILLING'])
      conlist.push(account.contact['BILLING']);

    if (account.contact['SHIPPING'])
      conlist.push(account.contact['SHIPPING']);

    if (account.contact['GENERAL'])
      account.contact['GENERAL'].split(',').forEach(part => {
        if (part)
          conlist.push(part)
      })


    for (let con = 0; con < conlist.length; con++) {

      let c = (await DBredirects.findData(mapper.mapper.forModel('Contact'),
        { contactid: conlist[con] }, {}, { consistency: options.consistency, prepare: true })).first();

      let e = (await DBredirects.findData(mapper.mapper.forModel('Email'),
        { eid: c.emaillist }, {}, { consistency: options.consistency, prepare: true })).first();

      let p = (await DBredirects.findData(mapper.mapper.forModel('Phone'),
        { pid: c.phonelist }, {}, { consistency: options.consistency, prepare: true })).first();

      if (c.status != 'INACTIVE') {
        conObj.push({
          Contact: c,
          Email: e,
          Phone: p,
          Account: { customerid: account.customerid }
        })
      }

    }

    return conObj;
  } catch (err) {
    return err;
  }
}

// get list of registered address
//Admin use is also applicable
exports.getAddressList = async (client, mapper, options, param, body, headers) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    let account;

    if (userR.type == 'ADMIN' && body.listaccount) {
      account = (await DBredirects.findData(mapper.mapper.forModel('Account'),
        { customerid: body.listaccount }, {}, { consistency: options.consistency, prepare: true })
      ).first();
    } else {
      account = (
        await DBredirects.findData(mapper.mapper.forModel('Account'),
          { customerid: userR.accntid }, {}, { consistency: options.consistency, prepare: true })
      ).first();
    }


    let adrObj = [];

    let adrlist = [];
    if (!account.address) {
      return [{
        Account: { customerid: account.customerid },
        Address: {},
        Zip: {}
      }]
    }

    if (account.address['PRIMARY'])
      adrlist.push(account.address['PRIMARY']);

    if (account.address['BILLING'])
      adrlist.push(account.address['BILLING']);

    if (account.address['SHIPPING'])
      adrlist.push(account.address['SHIPPING']);

    if (account.address['GENERAL'])
      account.address['GENERAL'].split(',').forEach(part => {
        if (part)
          adrlist.push(part)
      })


    for (let adr = 0; adr < adrlist.length; adr++) {

      let a = (await DBredirects.findData(mapper.mapper.forModel('Address'),
        { addressid: adrlist[adr] }, {}, { consistency: options.consistency, prepare: true })).first();

      let z = (await DBredirects.findData(mapper.mapper.forModel('Zip'),
        { zid: a.ziplist }, {}, { consistency: options.consistency, prepare: true })).first();
      if (a.status != 'INACTIVE') {
        adrObj.push({
          Address: a,
          Zip: z,
          Account: { customerid: account.customerid }
        })
      }
    }

    return adrObj;
  } catch (err) {
    return err;
  }
}
/*************************************
 * 
 * Account create functionlity 
 * 
 * ***********************************/
/* returns
{
        "userid": "",
        "logintoken": "0",
        "apitoken": "0",
        "type": ""
    }
*/
exports.postAccount = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    if (client === undefined) {
      return new Error("Client can not be null");
    }

    //Generate key value pair to match client and database column definition and verify the payload
    body.System.user = 'ANONYMOUS';
    let accntobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Account');

    //Return error if no object is present
    if (!accntobj.status) {
      return new Error(`Validation failed. ${accntobj.msg}`);
    }

    //set null to unset to avoid tombstone if missed anywhere
    accntobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, accntobj.data);

    //validate data before inserting into Database. Check if account already exists.
    let checkCustomer = await DBredirects.findAllData(mapper.mapper.forModel('Account'), {}, { fields: ['customerid', 'status', 'email', 'id'] }, { consistency: options.consistency });
    if ((checkCustomer._rs.rows.filter(obj => {
      return (obj['customerid'] == accntobj.data[0]['Account']['customerid'] || obj['email'] == accntobj.data[0]['Account']['email'])
    })).length != 0) {
      return new Error('Account already exists. It may not be active. Please contact us in such cases');
    } else {

      //Email validation      
      await mapper.global.validateEmail(accntobj.data[0], mapper, options, auditFunc, 'CREATE-ACCOUNT', DBredirects, audit_all_fields, system_user);
      //Phone validtion
      await mapper.global.validatePhone(accntobj.data[0], mapper, options, auditFunc, 'CREATE-ACCOUNT', DBredirects, audit_all_fields, system_user);
      //zip validtion
      await mapper.global.validateZip(accntobj.data[0], mapper, options, auditFunc, 'CREATE-ACCOUNT', DBredirects, audit_all_fields, system_user);

      accntobj.data[0].Account.created = accntobj.data[0].Account.updated
      accntobj.data[0].Account.createdby = accntobj.data[0].Account.upatedby
      accntobj.data[0].Account.status = 'UNVERIFIED'
      let insertS = (await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Account'), {}, accntobj.data[0].Account, { consistency: options.consistency }))._rs.rows

      //callback audit
      let value = { 'new_data': JSON.stringify(accntobj.data[0].Account), 'action_result': JSON.stringify(insertS) }
      auditFunc(value, 'CURRENT', 'CREATE', 'Account', audit_all_fields, mapper, DBredirects, options, system_user, 'Account');

      if (insertS[0]["[applied]"] == true) {
        let User = mapper.global.Schema.User;
        User.status = 'ACTIVE'
        User.type = 'GEN'
        User.accntid = accntobj.data[0].Account.customerid
        let usrobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings({
          data: [{
            User: User
          }], System: {
            user: accntobj.data[0].Account.email
          }
        }), 'User');

        let tokens = await mapper.global.validateT(mapper, accntobj.data[0].Account.email, usrobj.data[0].User.userid, 'GENERATE', DBredirects);
        usrobj.data[0].User.password = tokens.passwordtoken
        usrobj.data[0].User.logintoken = tokens.logintoken
        usrobj.data[0].User.apitoken = tokens.apitoken
        usrobj.data[0].User.created = accntobj.data[0].Account.updated
        usrobj.data[0].User.emailid = accntobj.data[0].Account.email
        usrobj.data[0].User.createdby = accntobj.data[0].Account.updatedby
        usrobj.data[0].User.lastlogin = accntobj.data[0].Account.updated
        let t

        try {
          t = (await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('User'),
            {}, usrobj.data[0].User, { consistency: options.consistency }))._rs.rows
          //  console.log(t)
          //If user id creation fails delete account
          if (!(t[0]["[applied]"] == true)) {
            let d = (await DBredirects.deleteData(mapper.mapper.forModel('Account'), accntobj.data[0].Account, {}, {}))
            let value = { 'old_data': JSON.stringify(accntobj.data[0].Account), 'action_result': JSON.stringify(d) }
            auditFunc(value, 'HISTORY', 'DELETE', 'Account', audit_all_fields, mapper, DBredirects, options, system_user, 'Account');

          }
        } catch (error) {
          console.log(error)
          let d = (await DBredirects.deleteData(mapper.mapper.forModel('Account'), accntobj.data[0].Account, {}, {}))
          let value = { 'old_data': JSON.stringify(accntobj.data[0].Account), 'action_result': JSON.stringify(d) }
          auditFunc(value, 'HISTORY', 'DELETE', 'Account', audit_all_fields, mapper, DBredirects, options, system_user, 'Account - Delete');
          return new Error('Account creation failed.')
        }
        //callback audit
        let value = { 'new_data': JSON.stringify(usrobj.data[0].User), 'action_result': JSON.stringify(t) }
        auditFunc(value, 'CURRENT', 'CREATE', 'User', audit_all_fields, mapper, DBredirects, options, system_user, 'User');

        //insert communication
        if (t[0]["[applied]"] == true) {
          let msg = `Your account has been successfully created. Please note, your cusomer id is ${accntobj.data[0].Account.customerid} and user id is ${usrobj.data[0].User.userid}. Password is the email provided during resigtration. Please change it on first login.`
          let com = (
            await mapper.global.communicationEntry('INBOUND', accntobj.data[0].Account.customerid, accntobj.data[0].Account.updatedby, msg, mapper, DBredirects, options, system_user)
          )
          //callback audit
          value = {
            'new_data': JSON.stringify(com.data),
            'action_result': JSON.stringify(
              com.rs
            )
          }

          //callback audit
          auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

          const SE = require('./mail');

          let x = await SE(mapper.global.ENV.EMAILOTHER, accntobj.data[0].Account.email, 'DO NOT REPLY. Account created!', `
            Dear ${accntobj.data[0].Account.fullname},
            <br/><br/>
            Your Account has been successfully created. Your user id is <b> ${usrobj.data[0].User.userid} </b>and default password is your registered email.
            <br/>
            You can use either user id or registered email as USERNAME to login. We advice you to change your password on your first login.
            <br/><br/>
            Thanks and Regards,<br/>
            PRASI LABS
            <br/><br/>
            <b>Note: Do not reply to this email. It is not monitored.</b>
          `, mapper);

          return { userid: usrobj.data[0].User.userid, type: usrobj.data[0].User.type }
        }
        else {
          return new Error("Something went wrong. Please try again")
        }
      } else {
        return new Error("Something went wrong. Please try again")
      }
    }
  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * Account update functionlity 
 * 
 * ***********************************/
/*
returns updated fields
{
        "fieldsU": [
            "field1 ",
            "field2 ",...
        ]
    }
*/
exports.updateAccount = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    //Generate key value pair to match client and database column definition and verify the payload
    let accntobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Account');
    //Return error if no object is present

    if (!accntobj.status) {
      return new Error(`Validation failed. ${accntobj.msg}`);
    }



    //set null to unset to avoid tombstone if missed anywhere
    accntobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, accntobj.data);

    //validate data before inserting into Database. Check if account already exists.
    let checkCustomer = await DBredirects.findSingleData(mapper.mapper.forModel('Account'), { customerid: accntobj.data[0]['Account']['customerid'], email: accntobj.data[0]['Account']['email'], id: accntobj.data[0]['Account']['id'] }, {}, { consistency: options.consistency });
    if (([checkCustomer].filter(obj => {
      return (obj && obj.status != 'INACTIVE')
    }).length == 0)) {


      return new Error('Account does not exist or inactive');


    } else {

      if (userR.type != 'ADMIN' && userR.accntid != accntobj.data[0]['Account']['customerid']) {
        return new Error('Only admin is allowed to change other account.')
      }
      //Phone validtion
      await mapper.global.validatePhone(accntobj.data[0], mapper, options, auditFunc, 'UPDATE-ACCOUNT', DBredirects, audit_all_fields, system_user);
      //zip validtion
      await mapper.global.validateZip(accntobj.data[0], mapper, options, auditFunc, 'UPDATE-ACCOUNT', DBredirects, audit_all_fields, system_user);

      checkCustomer.dob = checkCustomer.dob.toString()

      let fieldsU = mapper.global.VALIDATE.matchObj(checkCustomer, accntobj.data[0].Account, mapper.global.TRAVERSEJSON)
      if (fieldsU.length == 0) {
        return new Error("No change detected.");
      }
      let updateS = (await DBredirects.updateData(mapper.mapper.forModel('Account'), accntobj.data[0].Account, {}, { consistency: options.consistency }))._rs.rows

      //callback audit
      let value = { 'new_data': JSON.stringify(accntobj.data[0].Account), 'action_result': JSON.stringify(updateS) }
      auditFunc(value, 'CHANGE', 'UPDATE', 'Account', `${fieldsU}`, mapper, DBredirects, options, system_user, 'Account');

      //insert communication
      let msg = `Your account has been successfully updated. Following entities have been updated ${fieldsU}.`
      let com = (
        await mapper.global.communicationEntry('INBOUND', accntobj.data[0].Account.customerid, accntobj.data[0].Account.updatedby, msg, mapper, DBredirects, options, system_user)
      )
      //callback audit
      value = {
        'new_data': JSON.stringify(com.data),
        'action_result': JSON.stringify(
          com.rs
        )
      }

      auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

      return { fieldsU }
    }
  } catch (err) {
    return err;
  }
}


/*************************************
 * 
 * Contact upsert functionlity 
 * 
 * ***********************************/
exports.postContact = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    //Generate key value pair to match client and database column definition and verify the payload
    let conobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Contact');

    //Return error if no object is present
    if (!conobj.status) {
      return new Error(`Validation failed. ${conobj.msg}`);
    }

    //set null to unset to avoid tombstone if missed anywhere
    conobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, conobj.data);

    //loop
    for (let i = 0; i < conobj.data.length; i++) {


      if (userR.type != 'ADMIN') {
        if (conobj.data[i]['Account'] && conobj.data[i]['Account'].length != 0 && userR.accntid != conobj.data[i]['Account']['customerid']) {
          return new Error('Only admin is allowed to create or update other contact.')
        } else {
          conobj.data[i]['Account'] = { customerid: '' }
          conobj.data[i]['Account']['customerid'] = userR.accntid
        }
      } else {
        if (!conobj.data[i]['Account']) {
          conobj.data[i]['Account'] = { customerid: '' }
          conobj.data[i]['Account']['customerid'] = userR.accntid
        }
      }

      let accnt = (
        await DBredirects.findData(mapper.mapper.forModel('Account'),
          { customerid: conobj.data[i]['Account']['customerid'] }, { fields: ["customerid", "email", "id", "contact"] }, {})).first()

      //Phone validtion
      let p = await mapper.global.validatePhone(conobj.data[i], mapper, options, auditFunc, 'UPSERT-CONTACT', DBredirects, audit_all_fields, system_user);
      //email validtion
      let e = await mapper.global.validateEmail(conobj.data[i], mapper, options, auditFunc, 'UPSERT-CONTACT', DBredirects, audit_all_fields, system_user);

      //Update the reference of the phone and email

      conobj.data[i].Contact.phonelist = p ? p.pid : conobj.data[i].Phone.pid;
      conobj.data[i].Contact.emaillist = e ? e.eid : conobj.data[i].Email.eid;


      //getContact
      let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('Contact'), { contactid: conobj.data[i].Contact.contactid, id: conobj.data[i].Contact.id }, {}, { consistency: options.consistency });
      let activeOldData = [oldData].filter(obj => {
        return (obj && obj.status != 'INACTIVE')
      })

      let fieldsU = [];
      if (activeOldData.length != 0) {
        fieldsU = mapper.global.VALIDATE.matchObj(oldData, conobj.data[i].Contact, mapper.global.TRAVERSEJSON)
        if (fieldsU.length == 0) {
          return new Error("No change detected.");
        }

      } else {
        conobj.data[i].Contact.created = conobj.data[i].Contact.updated;
        conobj.data[i].Contact.createdby = conobj.data[i].Contact.updatedby
      }
      let upsertS = (await DBredirects.upsertData(mapper.mapper.forModel('Contact'), conobj.data[i].Contact, {}, { consistency: options.consistency }))
      let value = {
        'new_data': JSON.stringify(conobj.data[i].Contact),
        'action_result': JSON.stringify(upsertS._rs)
      }

      if (activeOldData.length != 0) {
        auditFunc(value, 'CHANGE', 'UPDATE', 'Contact', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-CONTACT');
      } else {
        auditFunc(value, 'CURRENT', 'CREATE', 'Contact', audit_all_fields, mapper, DBredirects, options, system_user, 'CREATE-CONTACT');
      }

      switch (conobj.data[i].Contact.type) {
        case 'PRIMARY':
          if (!accnt.contact)
            accnt.contact = { 'PRIMARY': conobj.data[i].Contact.contactid }
          else if (!(accnt.contact['PRIMARY'] && accnt.contact['PRIMARY'] != '')) {
            accnt.contact['PRIMARY'] = conobj.data[i].Contact.contactid
          }
          else if (!(accnt.contact['PRIMARY'] == conobj.data[i].Contact.contactid)) {

            //update type
            let old = (await DBredirects.findData(mapper.mapper.forModel('Contact'), { contactid: accnt.contact['PRIMARY'] }, {}, { consistency: options.consistency })).first()
            old.type = 'GENERAL'
            let u = (await DBredirects.updateData(mapper.mapper.forModel('Contact'), old, {}, { consistency: options.consistency }))
            let v = { 'new_data': JSON.stringify(old), 'action_result': JSON.stringify(u) }
            auditFunc(v, 'CHANGE', 'UPDATE', 'Contact', `type`, mapper, DBredirects, options, system_user, 'UPDATE-CONTACT');
            if (!(accnt.contact['GENERAL'] && accnt.contact['GENERAL'] != '')) {
              accnt.contact['GENERAL'] = conobj.data[i].Contact.contactid
            }
            else if (accnt.contact['GENERAL'].split(',').filter(item => { return item == accnt.contact['PRIMARY'] }).length < 1)
              accnt.contact['GENERAL'] = accnt.contact['PRIMARY'] + ',' + (accnt.contact['GENERAL'] ? accnt.contact['GENERAL'] : '')
            accnt.contact['PRIMARY'] = conobj.data[i].Contact.contactid

            if (accnt.contact['SHIPPING'] == conobj.data[i].Contact.contactid) {
              accnt.contact['SHIPPING'] = ''
            }
            if (accnt.contact['BILLING'] == conobj.data[i].Contact.contactid) {
              accnt.contact['BILLING'] = ''
            }

            accnt.contact['GENERAL'] = accnt.contact['GENERAL'].split(',').filter(item => { return item != conobj.data[i].Contact.contactid }).join(',');

          } break;
        case 'SHIPPING':
          if (!accnt.contact)
            accnt.contact = { 'SHIPPING': conobj.data[i].Contact.contactid }
          else if (!(accnt.contact['SHIPPING'] && accnt.contact['SHIPPING'] != '')) {
            accnt.contact['SHIPPING'] = conobj.data[i].Contact.contactid
          }
          else if (!(accnt.contact['SHIPPING'] == conobj.data[i].Contact.contactid)) {

            //update type
            let old = (await DBredirects.findData(mapper.mapper.forModel('Contact'), { contactid: accnt.contact['SHIPPING'] }, {}, { consistency: options.consistency })).first()
            old.type = 'GENERAL'
            let u = (await DBredirects.updateData(mapper.mapper.forModel('Contact'), old, {}, { consistency: options.consistency }))
            let v = { 'new_data': JSON.stringify(old), 'action_result': JSON.stringify(u) }
            auditFunc(v, 'CHANGE', 'UPDATE', 'Contact', `type`, mapper, DBredirects, options, system_user, 'UPDATE-CONTACT');
            if (!(accnt.contact['GENERAL'] && accnt.contact['GENERAL'] != '')) {
              accnt.contact['GENERAL'] = conobj.data[i].Contact.contactid
            }
            else if (accnt.contact['GENERAL'].split(',').filter(item => { return item == accnt.contact['SHIPPING'] }).length < 1)
              accnt.contact['GENERAL'] = accnt.contact['SHIPPING'] + ',' + (accnt.contact['GENERAL'] ? accnt.contact['GENERAL'] : '')
            accnt.contact['SHIPPING'] = conobj.data[i].Contact.contactid

            if (accnt.contact['PRIMARY'] == conobj.data[i].Contact.contactid) {
              accnt.contact['PRIMARY'] = ''
            }
            if (accnt.contact['BILLING'] == conobj.data[i].Contact.contactid) {
              accnt.contact['BILLING'] = ''
            }
            accnt.contact['GENERAL'] = accnt.contact['GENERAL'].split(',').filter(item => { return item != conobj.data[i].Contact.contactid }).join(',');
          }

          break;
        case 'BILLING':
          if (!accnt.contact)
            accnt.contact = { 'BILLING': conobj.data[i].Contact.contactid }
          else if (!(accnt.contact['BILLING'] && accnt.contact['BILLING'] != '')) {
            accnt.contact['BILLING'] = conobj.data[i].Contact.contactid
          }
          else if (!(accnt.contact['BILLING'] == conobj.data[i].Contact.contactid)) {
            //update type
            let old = (await DBredirects.findData(mapper.mapper.forModel('Contact'), { contactid: accnt.contact['BILLING'] }, {}, { consistency: options.consistency })).first()
            old.type = 'GENERAL'
            let u = (await DBredirects.updateData(mapper.mapper.forModel('Contact'), old, {}, { consistency: options.consistency }))
            let v = { 'new_data': JSON.stringify(old), 'action_result': JSON.stringify(u) }
            auditFunc(v, 'CHANGE', 'UPDATE', 'Contact', `type`, mapper, DBredirects, options, system_user, 'UPDATE-CONTACT');

            if (!(accnt.contact['GENERAL'] && accnt.contact['GENERAL'] != '')) {
              accnt.contact['GENERAL'] = conobj.data[i].Contact.contactid
            }
            else if (accnt.contact['GENERAL'].split(',').filter(item => { return item == accnt.contact['BILLING'] }).length < 1)
              accnt.contact['GENERAL'] = accnt.contact['BILLING'] + ',' + (accnt.contact['GENERAL'] ? accnt.contact['GENERAL'] : '')
            accnt.contact['BILLING'] = conobj.data[i].Contact.contactid


            if (accnt.contact['PRIMARY'] == conobj.data[i].Contact.contactid) {
              accnt.contact['PRIMARY'] = ''
            }
            if (accnt.contact['SHIPPING'] == conobj.data[i].Contact.contactid) {
              accnt.contact['SHIPPING'] = ''
            }
            accnt.contact['GENERAL'] = accnt.contact['GENERAL'].split(',').filter(item => { return item != conobj.data[i].Contact.contactid }).join(',');

          }

          break;
        default:
          if (!accnt.contact)
            accnt.contact = { 'GENERAL': conobj.data[i].Contact.contactid }
          else if (!(accnt.contact['GENERAL'] && accnt.contact['GENERAL'] != '')) {
            accnt.contact['GENERAL'] = conobj.data[i].Contact.contactid
          }
          else if (accnt.contact['GENERAL'].split(',').filter(item => { return item.trim() == conobj.data[i].Contact.contactid.trim() }).length < 1) {
            accnt.contact['GENERAL'] = conobj.data[i].Contact.contactid + ',' + (accnt.contact['GENERAL'] ? accnt.contact['GENERAL'] : '')
            if (accnt.contact['PRIMARY'] == conobj.data[i].Contact.contactid) {
              accnt.contact['PRIMARY'] = ''
            }
            if (accnt.contact['SHIPPING'] == conobj.data[i].Contact.contactid) {
              accnt.contact['SHIPPING'] = ''
            }
            if (accnt.contact['BILLING'] == conobj.data[i].Contact.contactid) {
              accnt.contact['BILLING'] = ''
            }
          }
      }
      let updateS = (await DBredirects.updateData(mapper.mapper.forModel('Account'), accnt, {}, { consistency: options.consistency }))._rs.rows

      //callback audit
      value = { 'new_data': JSON.stringify(accnt), 'action_result': JSON.stringify(updateS) }
      auditFunc(value, 'CHANGE', 'UPDATE', 'Account', `contact`, mapper, DBredirects, options, system_user, 'UPSERT-CONTACT Account');

      //insert communication
      let msg = `Your contact ${conobj.data[i].Contact.contactid} has been successfully updated. Following entities have been updated ${fieldsU}.`
      let com = (
        await mapper.global.communicationEntry('INBOUND', userR.accntid, conobj.data[i].Contact.updatedby, msg, mapper, DBredirects, options, system_user)
      )
      //callback audit
      value = {
        'new_data': JSON.stringify(com.data),
        'action_result': JSON.stringify(
          com.rs
        )
      }
      auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
      //lop end
    }
    return conobj.data.length

  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * Address upsert functionlity 
 * 
 * ***********************************/
exports.postAddress = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    //Generate key value pair to match client and database column definition and verify the payload
    let adrobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Address');

    //Return error if no object is present
    if (!adrobj.status) {
      return new Error(`Validation failed. ${adrobj.msg}`);
    }

    //set null to unset to avoid tombstone if missed anywhere
    adrobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, adrobj.data);

    //loop
    for (let i = 0; i < adrobj.data.length; i++) {

      if (userR.type != 'ADMIN') {
        if (adrobj.data[i]['Account'] && adrobj.data[i]['Account'].length != 0 && userR.accntid != adrobj.data[i]['Account']['customerid']) {
          return new Error('Only admin is allowed to create or update other address.')
        } else {
          adrobj.data[i]['Account'] = { customerid: '' }
          adrobj.data[i]['Account']['customerid'] = userR.accntid
        }
      } else {
        if (!adrobj.data[i]['Account']) {
          adrobj.data[i]['Account'] = { customerid: '' }
          adrobj.data[i]['Account']['customerid'] = userR.accntid
        }
      }

      let accnt = (
        await DBredirects.findData(mapper.mapper.forModel('Account'),
          { customerid: adrobj.data[i]['Account']['customerid'] }, { fields: ["customerid", "email", "id", "address"] }, {})).first()

      //zip validtion
      let z = await mapper.global.validateZip(adrobj.data[i], mapper, options, auditFunc, 'UPSERT-ADDRESS', DBredirects, audit_all_fields, system_user);

      //Update the reference of zip
      adrobj.data[i].Address.ziplist = z ? z.zid : adrobj.data[i].Zip.zid;


      //getAddress
      let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('Address'), { addressid: adrobj.data[i].Address.addressid, id: adrobj.data[i].Address.id }, {}, { consistency: options.consistency });
      let activeOldData = [oldData].filter(obj => {
        return (obj && obj.status != 'INACTIVE')
      })

      let fieldsU = [];
      if (activeOldData.length != 0) {
        fieldsU = mapper.global.VALIDATE.matchObj(oldData, adrobj.data[i].Address, mapper.global.TRAVERSEJSON)
        if (fieldsU.length == 0) {
          return new Error("No change detected.");
        }

      } else {
        adrobj.data[i].Address.created = adrobj.data[i].Address.updated;
        adrobj.data[i].Address.createdby = adrobj.data[i].Address.updatedby
      }
      let upsertS = (await DBredirects.upsertData(mapper.mapper.forModel('Address'), adrobj.data[i].Address, {}, { consistency: options.consistency }))
      let value = {
        'new_data': JSON.stringify(adrobj.data[i].Address),
        'action_result': JSON.stringify(upsertS._rs)
      }

      if (activeOldData.length != 0) {
        auditFunc(value, 'CHANGE', 'UPDATE', 'Address', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-ADDRESS');
      } else {
        auditFunc(value, 'CURRENT', 'CREATE', 'Address', audit_all_fields, mapper, DBredirects, options, system_user, 'CREATE-ADDRESS');
      }

      switch (adrobj.data[i].Address.type) {
        case 'PRIMARY':
          if (!accnt.address)
            accnt.address = { 'PRIMARY': adrobj.data[i].Address.addressid }
          else if (!(accnt.address['PRIMARY'] && accnt.address['PRIMARY'] != '')) {
            accnt.address['PRIMARY'] = adrobj.data[i].Address.addressid
          }
          else if (!(accnt.address['PRIMARY'] == adrobj.data[i].Address.addressid)) {
            //update type
            let old = (await DBredirects.findData(mapper.mapper.forModel('Address'), { addressid: accnt.address['PRIMARY'] }, {}, { consistency: options.consistency })).first()
            old.type = 'GENERAL'
            let u = (await DBredirects.updateData(mapper.mapper.forModel('Address'), old, {}, { consistency: options.consistency }))
            let v = { 'new_data': JSON.stringify(old), 'action_result': JSON.stringify(u) }
            auditFunc(v, 'CHANGE', 'UPDATE', 'Address', `type`, mapper, DBredirects, options, system_user, 'UPDATE-ADDRESS');

            if (!(accnt.address['GENERAL'] && accnt.address['GENERAL'] != '')) {
              accnt.address['GENERAL'] = adrobj.data[i].Address.addressid
            }
            else if (accnt.address['GENERAL'].split(',').filter(item => { return item == accnt.address['PRIMARY'] }).length < 1)
              accnt.address['GENERAL'] = accnt.address['PRIMARY'] + ',' + (accnt.address['GENERAL'] ? accnt.address['GENERAL'] : '')
            accnt.address['PRIMARY'] = adrobj.data[i].Address.addressid

            if (accnt.address['BILLING'] == adrobj.data[i].Address.addressid) {
              accnt.address['BILLING'] = ''
            }
            if (accnt.address['SHIPPING'] == adrobj.data[i].Address.addressid) {
              accnt.address['SHIPPING'] = ''
            }
            accnt.address['GENERAL'] = accnt.address['GENERAL'].split(',').filter(item => { return item != adrobj.data[i].Address.addressid }).join(',');

          }

          break;
        case 'SHIPPING':
          if (!accnt.address)
            accnt.address = { 'SHIPPING': adrobj.data[i].Address.addressid }
          else if (!(accnt.address['SHIPPING'] && accnt.address['SHIPPING'] != '')) {
            accnt.address['SHIPPING'] = adrobj.data[i].Address.addressid
          }
          else if (!(accnt.address['SHIPPING'] == adrobj.data[i].Address.addressid)) {
            //update type
            let old = (await DBredirects.findData(mapper.mapper.forModel('Address'), { addressid: accnt.address['SHIPPING'] }, {}, { consistency: options.consistency })).first()
            old.type = 'GENERAL'
            let u = (await DBredirects.updateData(mapper.mapper.forModel('Address'), old, {}, { consistency: options.consistency }))
            let v = { 'new_data': JSON.stringify(old), 'action_result': JSON.stringify(u) }
            auditFunc(v, 'CHANGE', 'UPDATE', 'Address', `type`, mapper, DBredirects, options, system_user, 'UPDATE-ADDRESS');
            if (!(accnt.address['GENERAL'] && accnt.address['GENERAL'] != '')) {
              accnt.address['GENERAL'] = adrobj.data[i].Address.addressid
            }
            else if (accnt.address['GENERAL'].split(',').filter(item => { return item == accnt.address['SHIPPING'] }).length < 1)
              accnt.address['GENERAL'] = accnt.address['SHIPPING'] + ',' + (accnt.address['GENERAL'] ? accnt.address['GENERAL'] : '')
            accnt.address['SHIPPING'] = adrobj.data[i].Address.addressid

            if (accnt.address['BILLING'] == adrobj.data[i].Address.addressid) {
              accnt.address['BILLING'] = ''
            }
            if (accnt.address['PRIMARY'] == adrobj.data[i].Address.addressid) {
              accnt.address['PRIMARY'] = ''
            }
            accnt.address['GENERAL'] = accnt.address['GENERAL'].split(',').filter(item => { return item != adrobj.data[i].Address.addressid }).join(',');


          }

          break;
        case 'BILLING':
          if (!accnt.address)
            accnt.address = { 'BILLING': adrobj.data[i].Address.addressid }
          else if (!(accnt.address['BILLING'] && accnt.address['BILLING'] != '')) {
            accnt.address['BILLING'] = adrobj.data[i].Address.addressid
          }
          else if (!(accnt.address['BILLING'] == adrobj.data[i].Address.addressid)) {
            //update type
            let old = (await DBredirects.findData(mapper.mapper.forModel('Address'), { addressid: accnt.address['BILLING'] }, {}, { consistency: options.consistency })).first()
            old.type = 'GENERAL'
            let u = (await DBredirects.updateData(mapper.mapper.forModel('Address'), old, {}, { consistency: options.consistency }))
            let v = { 'new_data': JSON.stringify(old), 'action_result': JSON.stringify(u) }
            auditFunc(v, 'CHANGE', 'UPDATE', 'Address', `type`, mapper, DBredirects, options, system_user, 'UPDATE-ADDRESS');

            if (!(accnt.address['GENERAL'] && accnt.address['GENERAL'] != '')) {
              accnt.address['GENERAL'] = adrobj.data[i].Address.addressid
            }
            else if (accnt.address['GENERAL'].split(',').filter(item => { return item == accnt.address['BILLING'] }).length < 1)
              accnt.address['GENERAL'] = accnt.address['BILLING'] + ',' + (accnt.address['GENERAL'] ? accnt.address['GENERAL'] : '')
            accnt.address['BILLING'] = adrobj.data[i].Address.addressid

            if (accnt.address['SHIPPING'] == adrobj.data[i].Address.addressid) {
              accnt.address['SHIPPING'] = ''
            }
            if (accnt.address['PRIMARY'] == adrobj.data[i].Address.addressid) {
              accnt.address['PRIMARY'] = ''
            }

            accnt.address['GENERAL'] = accnt.address['GENERAL'].split(',').filter(item => { return item != adrobj.data[i].Address.addressid }).join(',');

          }

          break;
        default:
          if (!accnt.address)
            accnt.address = { 'GENERAL': adrobj.data[i].Address.addressid }
          else {
            if (!(accnt.address['GENERAL'] && accnt.address['GENERAL'] != '')) {
              accnt.address['GENERAL'] = adrobj.data[i].Address.addressid
            }
            else if (accnt.address['GENERAL'].split(',').filter(item => { return item == adrobj.data[i].Address.addressid }).length < 1)
              accnt.address['GENERAL'] = adrobj.data[i].Address.addressid + ',' + (accnt.address['GENERAL'] ? accnt.address['GENERAL'] : '')

            if (accnt.address['SHIPPING'] == adrobj.data[i].Address.addressid) {
              accnt.address['SHIPPING'] = ''
            }
            if (accnt.address['PRIMARY'] == adrobj.data[i].Address.addressid) {
              accnt.address['PRIMARY'] = ''
            }
            if (accnt.address['BILLING'] == adrobj.data[i].Address.addressid) {
              accnt.address['BILLING'] = ''
            }


          }

      }
      let updateS = (await DBredirects.updateData(mapper.mapper.forModel('Account'), accnt, {}, { consistency: options.consistency }))._rs.rows

      //callback audit
      value = { 'new_data': JSON.stringify(accnt), 'action_result': JSON.stringify(updateS) }
      auditFunc(value, 'CHANGE', 'UPDATE', 'Account', `address`, mapper, DBredirects, options, system_user, 'UPSERT-ADDRESS Account');

      //insert communication
      let msg = `Your address ${adrobj.data[i].Address.addressid} has been successfully updated. Following entities have been updated ${fieldsU}.`
      let com = (
        await mapper.global.communicationEntry('INBOUND', userR.accntid, adrobj.data[i].Address.updatedby, msg, mapper, DBredirects, options, system_user)
      )
      //callback audit
      value = {
        'new_data': JSON.stringify(com.data),
        'action_result': JSON.stringify(
          com.rs
        )
      }
      auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
      //loop end
    }
    return adrobj.data.length

  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * Service request upsert functionlity 
 * 
 * ***********************************/
exports.postSR = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    //Generate key value pair to match client and database column definition and verify the payload
    let srobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Servicerequest');

    //Return error if no object is present
    if (!srobj.status) {
      return new Error(`Validation failed. ${srobj.msg}`);
    }

    //set null to unset to avoid tombstone if missed anywhere
    srobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, srobj.data);

    //loop
    for (let i = 0; i < srobj.data.length; i++) {

      if (userR.type != 'ADMIN') {
        if (srobj.data[i]['Servicerequest'] && srobj.data[i]['Servicerequest'].length != 0 && userR.accntid != srobj.data[i]['Servicerequest']['accntid']) {
          return new Error('Only admin is allowed to update other sr')
        }
      }

      //getSR
      let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('Servicerequest'), { srno: srobj.data[i].Servicerequest.srno, accntid: srobj.data[i].Servicerequest.accntid, type: srobj.data[i].Servicerequest.type, id: srobj.data[i].Servicerequest.id }, {}, { consistency: options.consistency });
      let activeOldData = [oldData].filter(obj => {
        return (obj && obj.status != 'INACTIVE')
      })

      let fieldsU = [];
      if (activeOldData.length != 0) {
        let y = {}
        Object.keys(oldData.interaction).forEach(key => {
          y[key] = oldData.interaction[key]
        })
        y[new Date().toISOString()] = srobj.data[i].Servicerequest.interaction
        srobj.data[i].Servicerequest.interaction = y
        fieldsU = mapper.global.VALIDATE.matchObj(oldData, srobj.data[i].Servicerequest, mapper.global.TRAVERSEJSON)
        if (fieldsU.length == 0) {
          return new Error("No change detected.");
        }

      } else {

        let x = {}
        x[new Date().toISOString()] = srobj.data[i].Servicerequest.interaction
        srobj.data[i].Servicerequest.interaction = x
        srobj.data[i].Servicerequest.created = srobj.data[i].Servicerequest.updated;
        srobj.data[i].Servicerequest.createdby = srobj.data[i].Servicerequest.updatedby
      }



      let upsertS = (await DBredirects.upsertData(mapper.mapper.forModel('Servicerequest'), srobj.data[i].Servicerequest, {}, { consistency: options.consistency }))
      let value = {
        'new_data': JSON.stringify(srobj.data[i].Servicerequest),
        'action_result': JSON.stringify(upsertS._rs)
      }

      if (activeOldData.length != 0) {
        auditFunc(value, 'CHANGE', 'UPDATE', 'Servicerequest', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-SR');
      } else {
        auditFunc(value, 'CURRENT', 'CREATE', 'Servicerequest', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-SR');
      }

      //insert communication
      let msg = `Your service request ${srobj.data[i].Servicerequest.srno} has been successfully updated. Following entities have been updated ${fieldsU}.`
      let com = (
        await mapper.global.communicationEntry('INBOUND', userR.accntid, srobj.data[i].Servicerequest.updatedby, msg, mapper, DBredirects, options, system_user)
      )
      //callback audit
      value = {
        'new_data': JSON.stringify(com.data),
        'action_result': JSON.stringify(
          com.rs
        )
      }
      auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
      //lop end
    }
    return srobj.data.length
  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * Order upsert functionlity 
 * 
 * ***********************************/
exports.postOrder = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)


    let orderitemobj = body.data.reduce((oldval, newval) => {
      if (newval.hasOwnProperty('Orderitem'))
        oldval.push({ 'Orderitem': newval['Orderitem'] })
      return oldval
    }, [])

    let orderobj = body.data.reduce((oldval, newval) => {
      if (newval.hasOwnProperty('Orders'))
        oldval.push({ 'Orders': newval['Orders'] })
      return oldval
    }, [])


    //Generate key value pair to match client and database column definition and verify the payload
    orderobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper,
      new mapper.global.VALIDATE.Settings({
        data: orderobj,
        System: body.System
      }),
      'Orders');


    //Return error if no object is present
    if (!orderobj.status) {
      return new Error(`Validation failed. ${orderobj.msg}`);
    }

    //Generate key value pair to match client and database column definition and verify the payload
    orderitemobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper,
      new mapper.global.VALIDATE.Settings({
        data: orderitemobj,
        System: body.System
      }),
      'Orderitem');

    //Return error if no object is present
    if (!orderitemobj.status) {
      return new Error(`Validation failed. ${orderitemobj.msg}`);
    }

    //set null to unset to avoid tombstone if missed anywherebody
    orderobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, orderobj.data);

    //  console.log( orderobj.data )

    orderitemobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, orderitemobj.data);

    orderitemobj = orderitemobj.data

    //console.log(orderobj.data);

    //loop
    for (let i = 0; i < orderobj.data.length; i++) {

      if (orderobj.data[i].Orders.accntid != userR.accntid && userR.type != 'ADMIN') {
        return new Error('Only admin is allowed to update other order')
      }

      let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('Orders'), { accntid: orderobj.data[i].Orders.accntid, type: orderobj.data[i].Orders.type, id: orderobj.data[i].Orders.id, orderid: orderobj.data[i].Orders.orderid, ordernum: orderobj.data[i].Orders.ordernum }, {}, { consistency: options.consistency });

      let activeOldData = [oldData].filter(obj => {
        return (obj && obj.status)
      })

      let Iorderitemobj = orderitemobj.filter(obj => {

        return obj.Orderitem.ordernum == orderobj.data[i].Orders.ordernum && obj.Orderitem.accntid == orderobj.data[i].Orders.accntid
      })

      if (Iorderitemobj.length == 0) {
        return new Error('Order is invalid. No product could be found')
      }


      orderobj.data[i].Orders.oredritemdetail = new Set();
      Iorderitemobj.forEach(obj => {

        if (obj.Orderitem.ordernum == orderobj.data[i].Orders.ordernum) {
          orderobj.data[i].Orders.oredritemdetail.add(obj.Orderitem.itemid);
        }
      })

      let fieldsU = [];
      if (activeOldData.length != 0) {
        if (!mapper.global.VALIDATE.orderValidation(oldData, userR.type)) {
          return new Error('Order data validation failed.')
        }

        fieldsU = mapper.global.VALIDATE.matchObj(oldData, orderobj.data[i].Orders, mapper.global.TRAVERSEJSON)
        if (fieldsU.length == 0) {
          return new Error("No change detected.");
        }

      } else {
        orderobj.data[i].Orders.created = orderobj.data[i].Orders.updated;
        orderobj.data[i].Orders.createdby = orderobj.data[i].Orders.updatedby
      }

      if (orderobj.data[i].Orders.accntid != userR.accntid && userR.type != 'ADMIN') {
        return new Error('User account and submitted account did not match. only admin is allowed to do such operation.')
      }

      let gstin = (await DBredirects.findAllData(mapper.mapper.forModel('Account'), {}, { fields: ['customerid', 'idnum', 'idname'] },
        { consistency: options.consistency }))._rs.rows.filter(a => {
          return orderobj.data[i].Orders.accntid == a.customerid
        })

      if (gstin.length != 0) {
        orderobj.data[i].Orders.bill['GSTIN'] = gstin[0].idnum;
      } else {
        orderobj.data[i].Orders.bill['GSTIN'] = 'N/A';
      }

      if (orderobj.data[i].Orders.direction == 'OUT')
        orderobj.data[i].Orders.poordernum = mapper.global.UTILITY.generatePOrderId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY) + orderobj.data[i].Orders.poordernum


      let upsertS = (await DBredirects.upsertData(mapper.mapper.forModel('Orders'), orderobj.data[i].Orders, {}, { consistency: options.consistency }))
      let value = {
        'new_data': JSON.stringify(orderobj.data[i].Orders),
        'action_result': JSON.stringify(upsertS._rs)
      }

      if (activeOldData.length != 0) {
        auditFunc(value, 'CHANGE', 'UPDATE', 'Orders', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-ORDER');
      } else {
        auditFunc(value, 'CURRENT', 'CREATE', 'Orders', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-ORDER');
      }

      //insert order item
      let itemChanges = []
      let invoiceItem = []
      for (let k = 0; k < Iorderitemobj.length; k++) {
        //for invoice function
        Iorderitemobj[k].Orderitem.orderid = orderobj.data[i].Orders.orderid
        Iorderitemobj[k].Orderitem.created = Iorderitemobj[k].Orderitem.updated;
        Iorderitemobj[k].Orderitem.createdby = Iorderitemobj[k].Orderitem.updatedby
        invoiceItem.push(Iorderitemobj[k].Orderitem)

        if (orderobj.data[i].Orders.accntid != Iorderitemobj[k].Orderitem.accntid) {
          return new Error('Something has been wrong. Account at order and item levels are different. Aborting!')
        }

        if (!mapper.global.VALIDATE.orderItemValidation(Iorderitemobj[k].Orderitem, userR.type)) {
          return new Error('Order item data validation failed.')
        }

        //creating array for batching update
        //directly accessing mapping function
        itemChanges.push(mapper.mapper.forModel('Orderitem')
          .batching.update(Iorderitemobj[k].Orderitem, { ifExists: false }
            , { consistency: options.consistency }))


      }
      //batch call
      let upsertI = await mapper.mapper.batch(itemChanges);

      let valueI = {
        'new_data': JSON.stringify(Iorderitemobj),
        'action_result': JSON.stringify(upsertI._rs)
      }
      auditFunc(valueI, 'CHANGE', 'UPSERT', 'Orderitem', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-ORDERITEM');

      //insert communication
      let msg = `Your Order ${orderobj.data[i].Orders.orderid} has been successfully updated. Following entities have been updated ${fieldsU}.`
      let com = (
        await mapper.global.communicationEntry('INBOUND', userR.accntid, orderobj.data[i].Orders.updatedby, msg, mapper, DBredirects, options, system_user)
      )
      //callback audit
      value = {
        'new_data': JSON.stringify(com.data),
        'action_result': JSON.stringify(
          com.rs
        )
      }
      auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
      //lop end

      if (body.signal == 'accept') {
        let x = await updateInvoice(mapper, options, { Orders: orderobj.data[i].Orders, Orderitem: invoiceItem }, body.signal, auditFunc, DBredirects, audit_all_fields, system_user);
        console.log(x)
      }
      else if (body.signal == 'complete') {
        let x = await updateInvoice(mapper, options, { Orders: orderobj.data[i].Orders, Orderitem: invoiceItem }, body.signal, auditFunc, DBredirects, audit_all_fields, system_user);
        console.log(x)
      }
      else if (body.signal == 'cancel') {
        let x = await updateInvoice(mapper, options, { Orders: orderobj.data[i].Orders, Orderitem: invoiceItem }, body.signal, auditFunc, DBredirects, audit_all_fields, system_user);
        console.log(x)
      }
      else if (body.signal == 'edit') {
        let x = await updateInvoice(mapper, options, { Orders: orderobj.data[i].Orders, Orderitem: invoiceItem }, body.signal, auditFunc, DBredirects, audit_all_fields, system_user);
        console.log(x);
        //console.log(orderobj.data[i].Orders,invoiceItem)
        let y = await createInvoice(mapper, options, { Orders: orderobj.data[i].Orders, Orderitem: invoiceItem }, body.System, auditFunc, DBredirects, audit_all_fields, system_user);
        console.log(y);
      }
      else {

        let x = await createInvoice(mapper, options, { Orders: orderobj.data[i].Orders, Orderitem: invoiceItem }, body.System, auditFunc, DBredirects, audit_all_fields, system_user);
        console.log(x)
      }

      const SE = require('./mail');

      let x = await SE(mapper.global.ENV.EMAILOTHER, mapper.global.ENV.ENQUIRY, 'DO NOT REPLY. Order Update!', `
        Dear PRASI LABS,
        <br/><br/>
        Customer Account <b>${userR.accntid}</b> has created/updated the order <b>${orderobj.data[i].Orders.orderid}</b>. Please check this order and it's corresponding invoice.
        <br/><br/>
        Thanks and Regards,<br/>
        PRASI LABS Email Robot
        <br/><br/>
        <b>Note: Do not reply to this email. It is not monitored.</b>
      `, mapper);
    }

    return orderobj.data.length
  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * User update functionlity 
 * 
 * ***********************************/
exports.postUser = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)

    //Generate key value pair to match client and database column definition and verify the payload
    let usrobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'User');

    //Return error if no object is present
    if (!usrobj.status) {
      return new Error(`Validation failed. ${usrobj.msg}`);
    }
    //set null to unset to avoid tombstone if missed anywhere
    usrobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, usrobj.data);

    //loop
    for (let i = 0; i < usrobj.data.length; i++) {

      let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('User'), { userid: usrobj.data[i].User.userid, id: usrobj.data[i].User.id }, {}, { consistency: options.consistency });

      let fieldsU = [];

      if ([oldData].length != 0) {

        fieldsU = mapper.global.VALIDATE.matchObj(oldData, usrobj.data[i].User, mapper.global.TRAVERSEJSON)
        if (fieldsU.length == 0) {
          return new Error("No change detected.");
        }
      }

      if (body.System.user == usrobj.data[i].User.userid && usrobj.data[i].User.password.trim()) {

        //check old password.

        if (body.Password != (await mapper.global.validateT(mapper, oldData.password, body.System.user, 'PASSWORD', DBredirects)).data) {

          return new Error('Old password is wrong.')
        }
        //update new password.
        let tokens = await mapper.global.validateT(mapper, usrobj.data[i].User.password, body.System.user, 'GENERATE', DBredirects);
        usrobj.data[i].User.password = tokens.passwordtoken;
        usrobj.data[i].User.logintoken = tokens.logintoken;
        usrobj.data[i].User.apitoken = tokens.apitoken;

      }
      else if (userR.type != 'ADMIN') {
        return new Error('User is only allowed to update password.')
      }

      let updateS = (await DBredirects.updateData(mapper.mapper.forModel('User'), usrobj.data[i].User, {}, { consistency: options.consistency }))
      let value = {
        'new_data': JSON.stringify(usrobj.data[i].User),
        'action_result': JSON.stringify(updateS._rs)
      }

      auditFunc(value, 'CHANGE', 'UPDATE', 'User', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-USER');

      //insert communication
      let msg = `User ${JSON.stringify(usrobj.data[i].User.userid)} has been successfully updated. Following entities have been updated ${fieldsU}.`
      let com = (
        await mapper.global.communicationEntry('INBOUND', userR.accntid, usrobj.data[i].User.updatedby, msg, mapper, DBredirects, options, system_user)
      )
      //callback audit
      value = {
        'new_data': JSON.stringify(com.data),
        'action_result': JSON.stringify(
          com.rs
        )
      }
      auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
      //lop end
    }
    return usrobj.data.length
  } catch (err) {
    return err;
  }
}

/*************************************
 * 
 * File download
 * 
 * ***********************************/
exports.downloadInvoice = async (client, mapper, options, param, body, headers, auditFunc) => {
  try {
    //let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
    console.log(body.name)
    accessSync(path.resolve('fs', './', body.name.replace(/\//g, '_') + '.pdf',), constants.R_OK);
    return path.resolve('fs', './', body.name.replace(/\//g, '_') + '.pdf')

  } catch (error) {
    return error
  }
}