/*
Created by: Peeyush
Created on: March 23 2022
Email: j.k.peeyush@gmail.com
Description: declares all global vaiables
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
"use strict";
const path = require('path');
const { callbackify } = require('util');
//import database functions
const DBredirects = require(path.resolve('middleware/', './DBredirect'));
const system_user = 'SYSTEM_USER'
const audit_all_fields = 'ALL_FIELDS'
const jwt = require('jsonwebtoken');
const { accessSync, constants, readFileSync, writeFileSync, statSync } = require('fs')


/*************************************
 *************************************
 * 
 * Local functions
 * 
 *************************************
 * ***********************************/

//Cancel any new invoice. already final invoice can not be cancelled.
const invalidateOldInvoice = async (mapper, options, obj, bodySystem, auditFunc) => {
    console.log('Cancelling any existing new invoice...')
    let invData = (await DBredirects.findData(mapper.mapper.forModel('Invoice'), {
        accntid: obj.accntid, ordernum: obj.orderid
    }, {
    }, { consistency: options.consistency }))._rs.rows.filter(i => {
        return i.status.toUpperCase() != 'CANCELLED'
    })

    let finalInv = invData.filter(i => {
        return i.status.toUpperCase() == 'FINAL'
    })

    if (finalInv.length != 0) {
        throw 'Unable to generate new invoice. There is already a final invoice on this order. Please generate manually.'
    } else {
        for (let i = 0; i < invData.length; i++) {
            invData[i].status = 'CANCELLED';

            let update = (await DBredirects.UpdateData(mapper.mapper.forModel('Invoice'), invData[i],
                {}, { consistency: options.consistency }))._rs.rows

            let value = { 'new_data': JSON.stringify(invData[i]), 'action_result': JSON.stringify(update) }
            auditFunc(value, 'HISTORY', 'UPDATE', 'Invoice', audit_all_fields, mapper, DBredirects, options, system_user, 'Invoice');

            //create communication
            let msg = `Your old invoice has been cancelled and a fresh invoice has been generted.`
            let com = (
                await mapper.global.communicationEntry('INBOUND', invData[i].invid, obj.updatedby, msg, mapper, DBredirects, options, system_user)
            )
            //callback audit
            value = {
                'new_data': JSON.stringify(com.data),
                'action_result': JSON.stringify(
                    com.rs
                )
            }
            auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
        }
    }

    console.log('Cancellation completed...')
}

//Invoice creation
const createInvoice = require('./invoicecreate');
/**************************
 * 
 * exported admin fuctions
 * 
 **************************/

//Get all users
exports.getUserAdmin = async function (client, mapper, options, param, body, headers) {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        options.cql = `SELECT * FROM User;`
        return await DBredirects.queryData(client, mapper, options, []);

    } catch (err) {
        return err;
    }
}

//Get all keys
exports.getKeyAdmin = async function (client, mapper, options, param, body, headers) {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        options.cql = `SELECT ${mapper.global.KEY_QUERY_FIELD} FROM KEY;`
        return await DBredirects.queryData(client, mapper, options, []);

    } catch (err) {
        return err;
    }
}

//Get all lov
exports.getLovAdmin = async function (client, mapper, options, param, body, headers) {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        options.cql = `SELECT ${mapper.global.LOV_QUERY_FIELD} FROM LOV;`
        return await DBredirects.queryData(client, mapper, options, []);

    } catch (err) {
        return err;
    }
}
//Get all companyDetail
exports.getCompanyAdmin = async function (client, mapper, options, param, body, headers) {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        options.cql = `SELECT ${mapper.global.COMP_QUERY_FIELD} FROM COMPANY;`
        return await DBredirects.queryData(client, mapper, options, []);

    } catch (err) {
        return err;
    }
}

//for getting all products
exports.getAllAdminProducts = async function (client, mapper, options, param, body, headers) {
    try {

        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        options.cql = `SELECT * FROM PRODUCT;`
        return await DBredirects.queryData(client, mapper, options, []);

    } catch (err) {
        return err;
    }
}

//get GST
exports.getGST = async function (client, mapper, options, param, body, headers) {
    try {

        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        options.cql = `SELECT PRODID,PRODUCTNAME,CASNO,currency FROM PRODUCT;`
        let p = (await DBredirects.queryData(client, mapper, options, []));
        let g = mapper.global.GSTMAP

        let gst = p.map(product => {
            return {
                prodid: product.prodid,
                casno: product.casno,
                productname: product.productname,
                currency: product.currency,
                cgst: g[product.prodid] ? g[product.prodid]['cgst'] : 0,
                sgst: g[product.prodid] ? g[product.prodid]['sgst'] : 0,
                igst: g[product.prodid] ? g[product.prodid]['igst'] : 0,
                tax: g[product.prodid] ? g[product.prodid]['tax'] : 0,
                discount: g[product.prodid] ? g[product.prodid]['discount'] : 0
            }
        })

        return gst
    } catch (err) {
        return err;
    }
}

exports.getCommunicationListAdmin = async (client, mapper, options, param, body, headers) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        let account = (await DBredirects.findAllData(mapper.mapper.forModel('User'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows

        let com = (await DBredirects.findAllData(mapper.mapper.forModel('Communication'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows

        let returnCOM = com.map((c) => {
            let acname = account.filter(a => { return a.userid == c.user })

            c['accntname'] = acname[0] ? acname[0].userid : '';
            return c
        })

        return returnCOM
    } catch (err) {
        return err;
    }
}

//fetch all the service request
exports.getSRListAdmin = async (client, mapper, options, param, body, headers) => {
    try {

        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        let account = (await DBredirects.findAllData(mapper.mapper.forModel('Account'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows
        let sr = (await DBredirects.findAllData(mapper.mapper.forModel('Servicerequest'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows

        // console.log(account,sr)
        let returnSR = sr.map((c) => {
            let acname = account.filter(a => { return a.customerid == c.accntid })
            c['accntname'] = acname[0] ? acname[0].fullname : '';
            return c
        })

        //console.log(returnSR)

        return returnSR;
    } catch (err) {
        return err;
    }
}

//fetch all the service request
exports.getXchangeListAdmin = async (client, mapper, options, param, body, headers) => {
    try {

        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        return mapper.global.XRATE
    } catch (err) {
        return err;
    }
}

exports.getStatAdmin = async (client, mapper, options, param, body, headers) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        //account status
        let accountstatus = (
            await DBredirects.findData(
                mapper.mapper.forModel('Account'),
                { customerid: userR.accntid },
                {},
                { consistency: options.consistency, prepare: true }
            )
        ).first().status;

        //last login time
        let lastlogin = (
            await DBredirects.findData(mapper.mapper.forModel('User'),
                { userid: body.System.user }, {},
                { consistency: options.consistency, prepare: true })).first().lastlogin;
        //total srs
        options.cql = `SELECT count(*) FROM Servicerequest where accntid=?;`
        let totalSR = (await DBredirects.queryData(client, mapper, options, [userR.accntid]))[0].count.low;

        //total document uploaded
        let noofdoc = 0
        //calulation of total amount and succesful orders.
        let tmporders = (await DBredirects.findData(mapper.mapper.forModel('Invoice'),
            { accntid: userR.accntid }, {},
            { consistency: options.consistency, prepare: true }))._rs.rows.filter(i => {
                return (i.type == 'ORIGIONAL') && (i.status = 'NEW' || i.status == 'FINAL')
            });

        let totalCompletedOrders = tmporders.length

        let totalamount = tmporders.reduce((prev, curr) => {

            prev = prev + mapper.global.GST.calculteXrate(curr.finalamount, curr.currency, 'USD', mapper.global.XRATE)
            return prev
        }, 0)

        let totaldiscount = tmporders.reduce((prev, curr) => {
            prev = prev + curr.item.reduce((pre, cur) => {

                pre = pre + mapper.global.GST.calculteXrate(JSON.parse(cur).itemdiscount, JSON.parse(cur).currency, 'USD', mapper.global.XRATE)
                return pre
            }, 0)
            return prev
        }, 0)

        //calculation of 30 days data
        let today = new Date();
        let priorDate = new Date(new Date().setDate(today.getDate() - 30));

        let last30days = tmporders.filter(i => {
            return new Date(i.updated) > priorDate
        })

        let totalamount30 = last30days.reduce((prev, curr) => {
            prev = prev + mapper.global.GST.calculteXrate(curr.finalamount, curr.currency, 'USD', mapper.global.XRATE)
            return prev
        }, 0)

        let totaldiscount30 = last30days.reduce((prev, curr) => {
            prev = prev + curr.item.reduce((prev, curr) => {
                prev = prev + mapper.global.GST.calculteXrate(JSON.parse(curr).itemdiscount, JSON.parse(curr).currency, 'USD', mapper.global.XRATE)
                return prev
            }, 0)
            return prev
        }, 0)

        return {
            totaldiscount: Math.round(totaldiscount),
            totalamount: Math.round(totalamount),
            totalorders: totalCompletedOrders,
            accountstatus: accountstatus,
            lastlogin: lastlogin,
            totalSR: totalSR,
            noofdoc: noofdoc,
            totalamount30: Math.round(totalamount30),
            totaldiscount30: Math.round(totaldiscount30)
        }

    } catch (err) {
        return err;
    }
}

exports.getInvoiceListAdmin = async (client, mapper, options, param, body, headers) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        let account = (await DBredirects.findAllData(mapper.mapper.forModel('Account'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows
        let inv = (await DBredirects.findAllData(mapper.mapper.forModel('Invoice'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows

        let returnINV = inv.filter((c) => {
            let acname = account.filter(a => { return a.customerid == c.accntid })
            c['accntname'] = acname[0].fullname;

            if (c.type.toLowerCase() == 'origional' && c.status != 'CANCELLED')
                return c

        })

        return returnINV
    } catch (err) {
        return err;
    }
}

exports.getLedgerListAdmin = async (client, mapper, options, param, body, headers) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }


        return (await DBredirects.findAllData(mapper.mapper.forModel('Ledger'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows


    } catch (err) {
        return err;
    }
}


exports.getOrderListAdmin = async (client, mapper, options, param, body, headers) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        let orderArray = (await DBredirects.findAllData(mapper.mapper.forModel('Orders'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows
        let orderItemArray = (await DBredirects.findAllData(mapper.mapper.forModel('Orderitem'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows
        let retOrderArray = []
        orderArray.forEach(order => {
            retOrderArray.push({ Orders: order })
        }
        )
        orderItemArray.filter(item => {
            retOrderArray.push({ Orderitem: item })
        })

        return retOrderArray
    } catch (err) {
        return err;
    }
}

exports.getLoginListAdmin = async (client, mapper, options, param, body, headers) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        return (await DBredirects.findData(mapper.mapper.forModel('Communication'), {}, {}, { consistency: options.consistency, prepare: true }))._rs.rows.filter(item => {
            return item.type == 'LOGIN'
        });
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * Account update functionlity 
 * 
 * ***********************************/
/*
returns updated fields
{
        "fieldsU": [
            "field1 ",
            "field2 ",...
        ]
    }
*/
exports.updateAccountAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        //Generate key value pair to match client and database column definition and verify the payload
        let accntobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Account');
        //Return error if no object is present

        if (!accntobj.status) {
            return new Error(`Validation failed. ${accntobj.msg}`);
        }

        //set null to unset to avoid tombstone if missed anywhere
        accntobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, accntobj.data);

        //validate data before inserting into Database. Check if account already exists.
        let checkCustomer = await DBredirects.findSingleData(mapper.mapper.forModel('Account'), { customerid: accntobj.data[0]['Account']['customerid'], email: accntobj.data[0]['Account']['email'], id: accntobj.data[0]['Account']['id'] }, {}, { consistency: options.consistency });

        if (([checkCustomer].length == 0)) {
            return new Error('Account does not exist');
        } else {

            checkCustomer.dob = checkCustomer.dob.toString()

            let fieldsU = mapper.global.VALIDATE.matchObj(checkCustomer, accntobj.data[0].Account, mapper.global.TRAVERSEJSON)
            if (fieldsU.length == 0) {
                return new Error("No change detected.");
            }

            let updateS = (await DBredirects.updateData(mapper.mapper.forModel('Account'), accntobj.data[0].Account, {}, { consistency: options.consistency }))._rs.rows

            //callback audit
            let value = { 'new_data': JSON.stringify(accntobj.data[0].Account), 'action_result': JSON.stringify(updateS) }
            auditFunc(value, 'CHANGE', 'UPDATE', 'Account', `${fieldsU}`, mapper, DBredirects, options, system_user, 'Account');

            //insert communication
            let msg = `Your account has been successfully updated. Following entities have been updated ${fieldsU}.`
            let com = (
                await mapper.global.communicationEntry('INBOUND', accntobj.data[0].Account.customerid, userR.userid, msg, mapper, DBredirects, options, system_user)
            )
            //callback audit
            value = {
                'new_data': JSON.stringify(com.data),
                'action_result': JSON.stringify(
                    com.rs
                )
            }

            auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

            return { fieldsU }
        }
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * Order item update functionlity 
 * 
 * ***********************************/
exports.postOrderitemAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        //Generate key value pair to match client and database column definition and verify the payload
        let orderitemobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Orderitem');

        //Return error if no object is present
        if (!orderitemobj.status) {
            return new Error(`Validation failed. ${orderitemobj.msg}`);
        }
        //set null to unset to avoid tombstone if missed anywhere
        orderitemobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, orderitemobj.data);

        //loop
        for (let i = 0; i < orderitemobj.data.length; i++) {

            let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('Orderitem'), { accntid: orderitemobj.data[i].Orderitem.accntid, type: orderitemobj.data[i].Orderitem.type, id: orderitemobj.data[i].Orderitem.id, itemid: orderitemobj.data[i].Orderitem.itemid, ordernum: orderitemobj.data[i].Orderitem.ordernum }, {}, { consistency: options.consistency });

            let activeOldData = [oldData]


            let fieldsU = [];
            if (activeOldData.length != 0) {

                if (!mapper.global.VALIDATE.orderItemValidation(oldData, userR.type)) {
                    return new Error('Order item data validation failed.')
                }

                fieldsU = mapper.global.VALIDATE.matchObj(oldData, orderitemobj.data[i].Orderitem, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    return new Error("No change detected.");
                }

            } else {
                return new Error('Order item could not be found')
            }

            let updateS = (await DBredirects.updateData(mapper.mapper.forModel('Orderitem'), orderitemobj.data[i].Orderitem, {}, { consistency: options.consistency }))

            let value = {
                'new_data': JSON.stringify(orderitemobj.data[i].Orderitem),
                'action_result': JSON.stringify(updateS._rs)
            }

            //console.log(value);

            auditFunc(value, 'CHANGE', 'UPDATE', 'Orders', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-ORDERITEM');


            //insert communication
            let msg = `Your Order item ${orderitemobj.data[i].Orderitem.itemid} has been successfully updated. Following entities have been updated ${fieldsU}.`
            let com = (
                await mapper.global.communicationEntry('INBOUND', userR.accntid, orderitemobj.data[i].Orderitem.updatedby, msg, mapper, DBredirects, options, system_user)
            )
            //callback audit
            value = {
                'new_data': JSON.stringify(com.data),
                'action_result': JSON.stringify(
                    com.rs
                )
            }
            auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
            //lop end
        }
        return orderitemobj.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * Key table upsert functionlity 
 * 
 * ***********************************/
exports.postKeyAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        //Generate key value pair to match client and database column definition and verify the payload
        let keyobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Key');

        //Return error if no object is present
        if (!keyobj.status) {
            return new Error(`Validation failed. ${keyobj.msg}`);
        }
        //set null to unset to avoid tombstone if missed anywhere
        keyobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, keyobj.data);

        //loop
        let itemChanges = []
        for (let i = 0; i < keyobj.data.length; i++) {

            let oldData = (await (new DBredirects.mapperFunc(mapper.mapper.forModel('Key'), 'key')).getResults(keyobj.data[i].Key.id))._rs.rows;

            let activeOldData = oldData.filter(obj => {
                return (obj && obj.status)
            })

            let fieldsU = [];
            if (activeOldData.length != 0) {

                fieldsU = mapper.global.VALIDATE.matchObj(oldData, keyobj.data[i].Key, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    console.log('No change detected')
                } else {

                    itemChanges.push(mapper.mapper.forModel('Key')
                        .batching.update(keyobj.data[i].Key, { ifExists: false }
                            , { consistency: options.consistency }))
                }

            } else {
                keyobj.data[i].Key.created = keyobj.data[i].Key.updated;
                keyobj.data[i].Key.createdby = keyobj.data[i].Key.updatedby;

                itemChanges.push(mapper.mapper.forModel('Key')
                    .batching.update(keyobj.data[i].Key, { ifExists: false }
                        , { consistency: options.consistency }))
            }

        }
        let updateS = await mapper.mapper.batch(itemChanges);
        let value = {
            'new_data': JSON.stringify(keyobj.data),
            'action_result': JSON.stringify(updateS._rs)
        }

        auditFunc(value, 'CHANGE', 'UPSERT', 'Key', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-KEY');

        //insert communication
        let msg = `Your Key has been successfully updated. Following entities have been updated ${fieldsU}.`
        let com = (
            await mapper.global.communicationEntry('INBOUND', userR.accntid, userR.userid, msg, mapper, DBredirects, options, system_user)
        )
        //callback audit
        value = {
            'new_data': JSON.stringify(com.data),
            'action_result': JSON.stringify(
                com.rs
            )
        }
        auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

        //write to file
        let writeData = {}
        for (let n = 0; n < keyobj.data.length; n++) {
            writeData[keyobj.data[n].Key.keyname] = keyobj.data[n].Key.keyvalue
        }
        accessSync(mapper.global.keyfile, constants.W_OK);
        writeFileSync(path.join(path.dirname(mapper.global.keyfile), path.basename(mapper.global.keyfile) + '.' + String(Date.now())), JSON.stringify(mapper.global.KEY))
        writeFileSync(mapper.global.keyfile, JSON.stringify(writeData))

        return keyobj.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * Lov table upsert functionlity 
 * 
 * ***********************************/
exports.postLovAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        //Generate key value pair to match client and database column definition and verify the payload
        let lovobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Lov');

        //Return error if no object is present
        if (!lovobj.status) {
            return new Error(`Validation failed. ${lovobj.msg}`);
        }
        //set null to unset to avoid tombstone if missed anywhere
        lovobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, lovobj.data);

        //loop
        let itemChanges = []
        let lovnames = []
        for (let i = 0; i < lovobj.data.length; i++) {

            let oldData = (await (new DBredirects.mapperFunc(mapper.mapper.forModel('Lov'), 'lov')).getResults(lovobj.data[i].Lov.id))._rs.rows;

            let activeOldData = oldData.filter(obj => {
                return (obj && obj.status)
            })

            let fieldsU = [];
            if (activeOldData.length != 0) {
                fieldsU = mapper.global.VALIDATE.matchObj(oldData, lovobj.data[i].Lov, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    console.log('No change detected')
                } else {

                    itemChanges.push(mapper.mapper.forModel('Lov')
                        .batching.update(lovobj.data[i].Lov, { ifExists: false }
                            , { consistency: options.consistency }))
                }

            } else {
                lovobj.data[i].Lov.created = lovobj.data[i].Lov.updated;
                lovobj.data[i].Lov.createdby = lovobj.data[i].Lov.updatedby

                itemChanges.push(mapper.mapper.forModel('Lov')
                    .batching.update(lovobj.data[i].Lov, { ifExists: false }
                        , { consistency: options.consistency }))
            }

            //for file write
            lovnames.push(lovobj.data[i].Lov.lovname);
        }
        let updateS = await mapper.mapper.batch(itemChanges);
        let value = {
            'new_data': JSON.stringify(lovobj.data),
            'action_result': JSON.stringify(updateS._rs)
        }

        auditFunc(value, 'CHANGE', 'UPSERT', 'Lov', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-LOV');
        //insert communication
        let msg = `Your lovs has been successfully updated.`
        let com = (
            await mapper.global.communicationEntry('INBOUND', userR.accntid, userR.userid, msg, mapper, DBredirects, options, system_user)
        )
        //callback audit
        value = {
            'new_data': JSON.stringify(com.data),
            'action_result': JSON.stringify(
                com.rs
            )
        }
        auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

        //write to file
        let writeData = {}
        let keyval = []
        for (let k = 0; k < lovnames.length; k++) {

            for (let n = 0; n < lovobj.data.length; n++) {
                if (lovobj.data[n].Lov.lovname == lovnames[k] && lovobj.data[n].Lov.status == 'ACTIVE') {
                    keyval.push(lovobj.data[n].Lov.lovvalue)
                }

            }
            writeData[lovnames[k]] = keyval
            keyval = []
        }
        accessSync(mapper.global.lovfile, constants.W_OK);
        writeFileSync(path.join(path.dirname(mapper.global.lovfile), path.basename(mapper.global.lovfile) + '.' + String(Date.now())), JSON.stringify(mapper.global.LOV))
        writeFileSync(mapper.global.lovfile, JSON.stringify(writeData))

        return lovobj.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * company table upsert functionlity 
 * 
 * ***********************************/
exports.postCompanyAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        //Generate key value pair to match client and database column definition and verify the payload
        let cobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Company');

        //Return error if no object is present
        if (!cobj.status) {
            return new Error(`Validation failed. ${cobj.msg}`);
        }
        //set null to unset to avoid tombstone if missed anywhere
        cobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, cobj.data);

        //loop
        for (let i = 0; i < cobj.data.length; i++) {

            let oldData = (await (new DBredirects.mapperFunc(mapper.mapper.forModel('Company'), 'company')).getResults(cobj.data[i].Company.id))._rs.rows;

            let activeOldData = oldData.filter(obj => {
                return (obj && obj.status)
            })

            let fieldsU = [];
            if (activeOldData.length != 0) {

                fieldsU = mapper.global.VALIDATE.matchObj(oldData, cobj.data[i].Company, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    console.log("No change detected.");
                }
                itemChanges.push(mapper.mapper.forModel('Company')
                    .batching.update(cobj.data[i].Company, { ifExists: false }
                        , { consistency: options.consistency }))
            } else {
                cobj.data[i].Company.created = cobj.data[i].Company.updated;
                cobj.data[i].Company.createdby = cobj.data[i].Company.updatedby

                itemChanges.push(mapper.mapper.forModel('Company')
                    .batching.update(cobj.data[i].Company, { ifExists: false }
                        , { consistency: options.consistency }))
            }
        }

        let updateS = await mapper.mapper.batch(itemChanges);
        let value = {
            'new_data': JSON.stringify(cobj.data),
            'action_result': JSON.stringify(updateS._rs)
        }

        auditFunc(value, 'CHANGE', 'UPSERT', 'Company', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-COMPANY');


        //insert communication
        let msg = `Your company detail  has been successfully updated. Following entities have been updated ${fieldsU}.`
        let com = (
            await mapper.global.communicationEntry('INBOUND', userR.accntid, userR.userid, msg, mapper, DBredirects, options, system_user)
        )
        //callback audit
        value = {
            'new_data': JSON.stringify(com.data),
            'action_result': JSON.stringify(
                com.rs
            )
        }

        auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

        return cobj.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * Invoice update functionlity 
 * 
 * ***********************************/
exports.postInvoiceAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    var createPdf = require('./pdfCreate');

    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        //Generate key value pair to match client and database column definition and verify the payload
        let invobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Invoice');

        //Return error if no object is present
        if (!invobj.status) {
            return new Error(`Validation failed. ${invobj.msg}`);
        }
        //set null to unset to avoid tombstone if missed anywhere
        invobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, invobj.data);

        //loop
        for (let i = 0; i < invobj.data.length; i++) {

            let oldData = await DBredirects.findData(mapper.mapper.forModel('Invoice'), { accntid: invobj.data[i].Invoice.accntid, id: invobj.data[i].Invoice.id, invid: invobj.data[i].Invoice.invid, ordernum: invobj.data[i].Invoice.ordernum, type: mapper.q.in_(['ORIGIONAL']) }, { where: { status: 'NEW' } }, { consistency: options.consistency });

            let fieldsU = [];
            if (oldData.length != 0) {

                fieldsU = mapper.global.VALIDATE.matchObj(oldData, invobj.data[i].Invoice, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    return new Error("No change detected.");
                }
            }

            let updateS = (await DBredirects.upsertData(mapper.mapper.forModel('Invoice'), invobj.data[i].Invoice, {}, { consistency: options.consistency }))
            console.log('creating Invoice doc')
            createPdf('companyDetail', invobj.data[i].Invoice);
            invobj.data[i].Invoice.type = 'DUPLICATE'
            let updateS1 = (await DBredirects.upsertData(mapper.mapper.forModel('Invoice'), invobj.data[i].Invoice, {}, { consistency: options.consistency }))
            console.log('creating Invoice doc')
            createPdf('companyDetail', invobj.data[i].Invoice);
            invobj.data[i].Invoice.type = 'TRIPLICATE'
            let updateS2 = (await DBredirects.upsertData(mapper.mapper.forModel('Invoice'), invobj.data[i].Invoice, {}, { consistency: options.consistency }))
            console.log('creating Invoice doc')
            createPdf('companyDetail', invobj.data[i].Invoice);
            let value = {
                'new_data': JSON.stringify(invobj.data[i].Invoice),
                'action_result': JSON.stringify(updateS._rs) + JSON.stringify(updateS1._rs) + JSON.stringify(updateS2._rs)
            }

            auditFunc(value, 'CHANGE', 'UPDATE', 'Invoice', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-INVOICE');

            //insert communication
            let msg = `Your invoice ${JSON.stringify(invobj.data[i].Invoice.invid)} has been successfully updated. Following entities have been updated ${fieldsU}.`
            let com = (
                await mapper.global.communicationEntry('INBOUND', userR.accntid, invobj.data[i].Invoice.updatedby, msg, mapper, DBredirects, options, system_user)
            )
            //callback audit
            value = {
                'new_data': JSON.stringify(com.data),
                'action_result': JSON.stringify(
                    com.rs
                )
            }
            auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
            //lop end



        }
        return invobj.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * User update functionlity 
 * 
 * ***********************************/
exports.postUserAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        //Generate key value pair to match client and database column definition and verify the payload
        let usrobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'User');

        //Return error if no object is present
        if (!usrobj.status) {
            return new Error(`Validation failed. ${usrobj.msg}`);
        }
        //set null to unset to avoid tombstone if missed anywhere
        usrobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, usrobj.data);

        //loop
        for (let i = 0; i < usrobj.data.length; i++) {

            let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('User'), { userid: usrobj.data[i].User.userid, id: usrobj.data[i].User.id }, {}, { consistency: options.consistency });

            let fieldsU = [];

            if ([oldData].length != 0) {

                fieldsU = mapper.global.VALIDATE.matchObj(oldData, usrobj.data[i].User, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    return new Error("No change detected.");
                }
            }

            //removing password, logintoken and apitoken from the object
            let t = {
                id: usrobj.data[i].User.id,
                status: usrobj.data[i].User.status,
                type: usrobj.data[i].User.type,
                userid: usrobj.data[i].User.userid,
                accntid: usrobj.data[i].User.accntid,
                lastlogin: usrobj.data[i].User.lastlogin,
                updatedby: usrobj.data[i].User.updatedby,
                createdby: usrobj.data[i].User.createdby,
                created: usrobj.data[i].User.created,
                updated: usrobj.data[i].User.updated
            }
            usrobj.data[i].User = t


            let updateS = (await DBredirects.updateData(mapper.mapper.forModel('User'), usrobj.data[i].User, {}, { consistency: options.consistency }))
            let value = {
                'new_data': JSON.stringify(usrobj.data[i].User),
                'action_result': JSON.stringify(updateS._rs)
            }

            auditFunc(value, 'CHANGE', 'UPDATE', 'User', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-USER');

            //insert communication
            let msg = `User ${JSON.stringify(usrobj.data[i].User.userid)} has been successfully updated. Following entities have been updated ${fieldsU}.`
            let com = (
                await mapper.global.communicationEntry('INBOUND', userR.accntid, usrobj.data[i].User.updatedby, msg, mapper, DBredirects, options, system_user)
            )
            //callback audit
            value = {
                'new_data': JSON.stringify(com.data),
                'action_result': JSON.stringify(
                    com.rs
                )
            }
            auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
            //lop end
        }
        return usrobj.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * Product upsert functionlity 
 * 
 * ***********************************/
exports.postProduct = async (client, mapper, options, param, body, headers, auditFunc) => {

    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        if (userR.type != 'ADMIN') {
            return new Error('Only admin is allowed to change products.')
        }
        //Generate key value pair to match client and database column definition and verify the payload
        let prodobj = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Product');

        //Return error if no object is present
        if (!prodobj.status) {
            return new Error(`Validation failed. ${prodobj.msg}`);
        }
        //console.log(prodobj.data)
        //set null to unset to avoid tombstone if missed anywhere
        prodobj.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, prodobj.data);

        //loop

        for (let i = 0; i < prodobj.data.length; i++) {
            //getContact
            let oldData = await DBredirects.findSingleData(mapper.mapper.forModel('Product'), { id: prodobj.data[i].Product.id, prodid: prodobj.data[i].Product.prodid }, {}, { consistency: options.consistency });
            let activeOldData = [oldData].filter(obj => {
                return (obj && (obj.status == 'INACTIVE' || obj.status == 'ACTIVE'))
            })

            let fieldsU = [];
            if (activeOldData.length != 0) {
                fieldsU = mapper.global.VALIDATE.matchObj(oldData, prodobj.data[i].Product, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    return new Error("No change detected.");
                }

            } else {
                prodobj.data[i].Product.created = prodobj.data[i].Product.updated;
                prodobj.data[i].Product.createdby = prodobj.data[i].Product.updatedby
            }


            let upsertS = (await DBredirects.upsertData(mapper.mapper.forModel('Product'), prodobj.data[i].Product, {}, { consistency: options.consistency }))
            let value = {
                'new_data': JSON.stringify(prodobj.data[i].Product),
                'action_result': JSON.stringify(upsertS._rs)
            }

            if (activeOldData.length != 0) {
                auditFunc(value, 'CHANGE', 'UPDATE', 'Product', `${fieldsU}`, mapper, DBredirects, options, system_user, 'UPDATE-PRODUCT');
            } else {
                auditFunc(value, 'CURRENT', 'CREATE', 'Product', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-PRODUCT');
            }

            //insert communication
            let msg = `Your product ${prodobj.data[i].Product.prodid} has been successfully updated. Following entities have been updated ${fieldsU}.`
            let com = (
                await mapper.global.communicationEntry('INBOUND', userR.accntid, prodobj.data[i].Product.updatedby, msg, mapper, DBredirects, options, system_user)
            )
            //callback audit
            value = {
                'new_data': JSON.stringify(com.data),
                'action_result': JSON.stringify(
                    com.rs
                )
            }
            auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
            //lop end
        }
        return prodobj.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
 * 
 * ledger table upsert functionlity 
 * 
 * ***********************************/
exports.ledger = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        //Generate key value pair to match client and database column definition and verify the payload
        let ledger = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings(body), 'Ledger');

        //Return error if no object is present
        if (!ledger.status) {
            return new Error(`Validation failed. ${ledger.msg}`);
        }
        //set null to unset to avoid tombstone if missed anywhere
        ledger.data = mapper.global.VALIDATE.avoidTombstone(mapper.cassandra, ledger.data);
        let itemChanges = []
        //loop
        for (let i = 0; i < ledger.data.length; i++) {

            let oldData = (await (new DBredirects.mapperFunc(mapper.mapper.forModel('Ledger'), 'ledger')).getResults(ledger.data[i].Ledger.id))._rs.rows;


            let activeOldData = oldData.filter(obj => {

                return (obj && obj.status)
            })

            let fieldsU = [];

            if (activeOldData.length != 0) {

                fieldsU = mapper.global.VALIDATE.matchObj(oldData, ledger.data[i].Ledger, mapper.global.TRAVERSEJSON)
                if (fieldsU.length == 0) {
                    console.log("No change detected.");
                }
                if (ledger.data[i].Ledger.status == 'Delete') {
                    itemChanges.push(mapper.mapper.forModel('Ledger')
                        .batching.remove(ledger.data[i].Ledger, { ifExists: true }
                            , { consistency: options.consistency }))



                } else {


                    itemChanges.push(mapper.mapper.forModel('Ledger')
                        .batching.update(ledger.data[i].Ledger, { ifExists: false }
                            , { consistency: options.consistency }))



                }
            } else {


                ledger.data[i].Ledger.created = ledger.data[i].Ledger.updated;
                ledger.data[i].Ledger.createdby = ledger.data[i].Ledger.updatedby

                itemChanges.push(mapper.mapper.forModel('Ledger')
                    .batching.update(ledger.data[i].Ledger, { ifExists: false }
                        , { consistency: options.consistency }))


            }
        }

        let updateS = await mapper.mapper.batch(itemChanges);
        //console.log(itemChanges, updateS)
        let value = {
            'new_data': JSON.stringify(ledger.data),
            'action_result': JSON.stringify(updateS._rs)
        }

        auditFunc(value, 'CHANGE', 'UPSERT', 'Ledger', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-LEDGER');


        //insert communication
        let msg = `Your ledger detail has been successfully updated`
        let com = (
            await mapper.global.communicationEntry('INBOUND', userR.accntid, userR.userid, msg, mapper, DBredirects, options, system_user)
        )
        //callback audit
        value = {
            'new_data': JSON.stringify(com.data),
            'action_result': JSON.stringify(
                com.rs
            )
        }

        auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

        return ledger.data.length
    } catch (err) {
        return err;
    }
}

/*************************************
* 
* Exchange file upsert functionlity 
* 
* ***********************************/
exports.postXchangeAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        accessSync(mapper.global.xchangefile, constants.W_OK);
        writeFileSync(path.join(path.dirname(mapper.global.xchangefile), path.basename(mapper.global.xchangefile) + '.' + String(Date.now())), JSON.stringify(mapper.global.XRATE))
        writeFileSync(mapper.global.xchangefile, JSON.stringify(body.XRATE))

        return Object.keys(body.XRATE).length
    } catch (err) {
        return err;
    }
}
/*************************************
* 
* GST file upsert functionlity 
* 
* ***********************************/
exports.postGSTAdmin = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        accessSync(mapper.global.gstfile, constants.W_OK);
        writeFileSync(path.join(path.dirname(mapper.global.gstfile), path.basename(mapper.global.gstfile) + '.' + String(Date.now())), JSON.stringify(mapper.global.GSTMAP))
        writeFileSync(mapper.global.gstfile, JSON.stringify(body.GSTMAP))

        return Object.keys(body.GSTMAP).length
    } catch (err) {
        return err;
    }
}
/*************************************
* 
* Bulk products upload
* 
* ***********************************/
exports.bulkproducts = async (client, mapper, options, param, req, headers, auditFunc) => {
    let files = req.files;
    let body = { 'System': JSON.parse(req.body.System) }
    const xlsx = require('xlsx')
    let error = ''
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        files.forEach(f => {
            if (f.mimetype == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet') {
                let dataBuf = xlsx.read(f.buffer)
                let sheetnames = xlsx.read(f.buffer).SheetNames
                var aoa = xlsx.utils.sheet_to_json(dataBuf.Sheets[sheetnames[0]], { raw: true }).reduce((prev, cur) => {
                    if (cur) {
                        for (let i = 0; i < Object.keys(cur).length; i++) {
                            if (cur[Object.keys(cur)[i]] == "'") {
                                cur[Object.keys(cur)[i]] = ''
                            }
                        }

                        if (typeof cur['quantity'] == 'number') {
                            cur['quantity'] = cur['quantity'].toString()
                        }

                        if (typeof cur['priceperunit'] == 'number') {
                            cur['priceperunit'] = cur['priceperunit'].toString()
                        }

                        prev.push({ Product: cur })
                    }
                    return prev;
                }, [])
                // console.log(aoa)
                body['data'] = aoa

                mapper.global.callFunction('postProductAdmin', client, mapper, options, param, body, headers, auditFunc)
                    .then(data => {
                        if (data instanceof Error) {
                            console.log(data);
                            //res.json({ 'errorMsg': "Failed to upsert products." })
                            error = 'Some or all product entry failed.'
                        } else {
                            console.log('success')
                        }
                    })
            }
            return error
        })
    } catch (err) {
        return err;
    }
}
/*************************************
* 
* Bulk gst upload
* 
* ***********************************/
exports.bulkgst = async (client, mapper, options, param, req, headers, auditFunc) => {
    let files = req.files;
    let body = { 'System': JSON.parse(req.body.System) }
    const xlsx = require('xlsx')
    let error = ''
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        let GST = await mapper.global.callFunction('getGST', client, mapper, options, param, body, headers)
        // console.log(GST)

        files.forEach(f => {
            if (f.mimetype == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet') {
                let dataBuf = xlsx.read(f.buffer)
                let sheetnames = xlsx.read(f.buffer).SheetNames
                var aoa = xlsx.utils.sheet_to_json(dataBuf.Sheets[sheetnames[0]], { raw: true })
                GST.forEach(g => {
                    let temp = aoa.filter(f => {
                        return (f['casno'] == g['casno'])
                    }
                    )
                    if (temp.length > 0) {
                        g['cgst'] = temp[0]['cgst']
                        g['sgst'] = temp[0]['sgst']
                        g['igst'] = temp[0]['igst']
                        g['tax'] = temp[0]['tax']
                        g['discount'] = temp[0]['discount']
                    }
                })
                let tGST = GST.reduce((pre, cur) => {
                    pre[cur['prodid']] = {
                        'cgst': cur['cgst'].toString(),
                        'sgst': cur['sgst'].toString(),
                        'igst': cur['igst'].toString(),
                        'tax': cur['tax'].toString(),
                        'discount': cur['discount'].toString()
                    }
                    return pre
                }, {})
                body['GSTMAP'] = tGST
                mapper.global.callFunction('postGSTAdmin', client, mapper, options, param, body, headers, auditFunc)
                    .then(data => {
                        if (data instanceof Error) {
                            console.log(data);
                            //res.json({ 'errorMsg': "Failed to upsert products." })
                            error = 'GST update failed.'
                        } else {
                            console.log('success')
                        }
                    })
            }
            return error
        })
    } catch (err) {
        return err;
    }
}

/*************************************
* 
* Bulk inv upload
* 
* ***********************************/
exports.bulkinv = async (client, mapper, options, param, req, headers, auditFunc) => {
    let files = req.files;
    let body = { 'System': JSON.parse(req.body.System) }
    const xlsx = require('xlsx')
    let error = ''
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        files.forEach(f => {
            if (f.mimetype == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet') {
                let dataBuf = xlsx.read(f.buffer)
                let sheetnames = xlsx.read(f.buffer).SheetNames
                var aoa = xlsx.utils.sheet_to_json(dataBuf.Sheets[sheetnames[0]], { raw: true }).reduce((prev, cur) => {

                    return prev;
                }, [])
                console.log(aoa)

            }
            return error
        })
    } catch (err) {
        return err;
    }
}

/*************************************
* 
* get Orders and status
* 
* ***********************************/
exports.adminDataOrders = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }

        let data = (
            await DBredirects.findAllData(mapper.mapper.forModel('Orders'), {},
                { fields: ["status", "finalamount", "taxamount", "cgstamount", "sgstamount", "igstamount", "updated"] },
                { consistency: options.consistency, prepare: true })
        )._rs.rows

        //console.log(data);

        let totalSumFinal = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'COMPLETED')
                prev = Number(prev) + Number(curr.finalamount)
            return prev
        }, 0)

        let totalTaxFinal = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'COMPLETED')
                prev = Number(prev) + Number(curr.taxamount) + Number(curr.cgstamount) + Number(curr.sgstamount) + Number(curr.igstamount)
            return prev
        }, 0)

        let totalSumNew = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'IN_PROGRESS')
                prev = Number(prev) + Number(curr.finalamount)
            return prev
        }, 0)

        let totalTaxNew = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'IN_PROGRESS')
                prev = Number(prev) + Number(curr.taxamount) + Number(curr.cgstamount) + Number(curr.sgstamount) + Number(curr.igstamount)
            return prev
        }, 0)

        let totalSumPending = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'PENDING')
                prev = Number(prev) + Number(curr.finalamount)
            return prev
        }, 0)

        let totalTaxPending = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'PENDING')
                prev = Number(prev) + Number(curr.taxamount) + Number(curr.cgstamount) + Number(curr.sgstamount) + Number(curr.igstamount)
            return prev
        }, 0)
        
        return {
            list: data,
            ordercomplete: totalSumFinal,
            ordertaxcomplete: totalTaxFinal,
            orderinprogress: totalSumNew,
            ordertaxinprogress: totalTaxNew,
            orderpending: totalSumPending,
            ordertaxpending: totalTaxPending
        }

    } catch (err) {
        return err;
    }
}



/******************************
 * 
 * Get data for graphs invoice
 * 
 ******************************/
exports.adminDataInvoice = async (client, mapper, options, param, body, headers, auditFunc) => {
    try {
        let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
        if (userR.type != 'ADMIN') {
            return 'You are not an administrator. Operation is not permitted.'
        }
        let data = (
            await DBredirects.findAllData(mapper.mapper.forModel('Invoice'), {},
                { fields: ["status", "finalamount", "tax", "csgt", "sgst", "igst", "invdate", "type"] },
                { consistency: options.consistency, prepare: true })
        )._rs.rows.filter(
            inv => {
                return inv.type.toUpperCase() == 'ORIGIONAL' && (inv.status.toUpperCase() == 'NEW' || inv.status.toUpperCase() == 'FINAL')
            })
        console.log(data)


        let totalSumFinal = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'FINAL')
                prev = Number(prev) + Number(curr.finalamount)
            return prev
        }, 0)

        let totalTaxFinal = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'FINAL')
                prev = Number(prev) + Number(curr.tax) + Number(curr.cgst) + Number(curr.sgst) + Number(curr.igst)
            return prev
        }, 0)

        let totalSumNew = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'NEW')
                prev = Number(prev) + Number(curr.finalamount)
            return prev
        }, 0)

        let totalTaxNew = data.reduce((prev, curr) => {
            if (curr.status.toUpperCase() == 'NEW')
                prev = Number(prev) + Number(curr.tax) + Number(curr.cgst) + Number(curr.sgst) + Number(curr.igst)
            return prev
        }, 0)

        return {
            list: data,
            invoicefinal: totalSumFinal,
            invoicetaxfinal: totalTaxFinal,
            invoicenew: totalSumNew,
            invoicetaxnew: totalTaxNew
        }
    } catch (err) {
        return err;
    }
}


/*******************
 * 
 * send email
 * 
 *******************/
exports.sendemail = async (client, mapper, options, param, body, headers, auditFunc) => {

    try {
        /* let userR = await mapper.global.checkUser(client, mapper, options, param, body, headers, DBredirects)
         if (userR.type != 'ADMIN') {
             return 'You are not an administrator. Operation is not permitted.'
         }*/
        const SE = require('./mail');

        await SE(body.data.email, mapper.global.ENV.ENQUIRY, 'Product Enquiry', body.data, mapper);
        //insert communication
        let com = (
            await mapper.global.communicationEntry('ENQUIRY', body.data.email,
                body.data.email, JSON.stringify(body.data)
                , mapper, DBredirects, options, 'ANONYMOUS')
        )
        //callback audit
        let value = {
            'new_data': JSON.stringify(com.data),
            'action_result': JSON.stringify(
                com.rs
            )
        }
        auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')
        //lop end
        return '1'
    } catch (err) {
        return err;
    }

}