/*
Created by: Peeyush
Created on: July 16 2021
Email: j.k.peeyush@gmail.com
Description: declares all global vaiables
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
//Controller functions
const controller = require('./controller');
const controllerAdmin = require('./controllerAdmin')

//Custom js file having JSON tree pasrsing functions
const TRAVERSEJSON = require('./traverseJSON');

//Custom js file having validation functions
const VALIDATE = require('./validation');

//Custom js file for utility functions
const UTILITY = require('./utility');

//Custom js file for GST fuction
const GST = require('./gst')

//File read module
const { accessSync, constants, readFileSync } = require('fs')

//To resolve relativ path
const path = require('path');

const jwt = require('jsonwebtoken');

const Schema = require(path.resolve('middleware/', './ObjectType'));

var RDBMSJSON = null;
var ENV = null;
var FIELDCOLMAP = null;
var KEY = null;
var LOV = null;
var MAPPER = null;
var TEMP = null;
var GSTMAP = null;
var XRATE = null;
var CLIST = null;
try {
    //Country list configuration file
    accessSync(path.resolve('config', './countryCode.json'), constants.R_OK);
    CLIST = JSON.parse(readFileSync(path.resolve('config', './countryCode.json')));

    //Table constraints configuration file
    accessSync(path.resolve('config/', './rdbms.json'), constants.R_OK);
    RDBMSJSON = JSON.parse(readFileSync(path.resolve('config/', './rdbms.json')));

    //Envrionment configuration file
    accessSync(path.resolve('config/', './env.json'), constants.R_OK);
    ENV = JSON.parse(readFileSync(path.resolve('config/', './env.json')));

    //Field Column map configuration file
    accessSync(path.resolve('config/', './fieldcolumn.json'), constants.R_OK);
    FIELDCOLMAP = JSON.parse(readFileSync(path.resolve('config/', './fieldcolumn.json')));

    //GST product map configuration file
    accessSync(path.resolve('config/', './prodgstmapping.json'), constants.R_OK);
    GSTMAP = JSON.parse(readFileSync(path.resolve('config/', './prodgstmapping.json')));

    //GST product map configuration file
    accessSync(path.resolve('config/', './exchangeRate.json'), constants.R_OK);
    XRATE = JSON.parse(readFileSync(path.resolve('config/', './exchangeRate.json')));

    //Key definition configuration file
    accessSync(path.resolve('config/', './key.json'), constants.R_OK);
    KEY = JSON.parse(readFileSync(path.resolve('config/', './key.json')));

    //List of values configuration file
    accessSync(path.resolve('config/', './lov.json'), constants.R_OK);
    LOV = JSON.parse(readFileSync(path.resolve('config/', './lov.json')));

    //List of values mapper file
    accessSync(path.resolve('config/', './mapper.json'), constants.R_OK);
    MAPPER = JSON.parse(readFileSync(path.resolve('config/', './mapper.json')));

    //List of values statecode file
    accessSync(path.resolve('config/', './stateCode.json'), constants.R_OK);
    STCD = JSON.parse(readFileSync(path.resolve('config/', './stateCode.json')));

    //Temp folder
    TEMP = path.resolve('temp/');

} catch (error) {
    console.log(error);
}

exports.TRAVERSEJSON = TRAVERSEJSON;
exports.VALIDATE = VALIDATE;
exports.UTILITY = UTILITY;
exports.GST = GST;
exports.gstfile = path.resolve('config/', './prodgstmapping.json')
exports.lovfile = path.resolve('config/', './lov.json')
exports.keyfile = path.resolve('config/', './key.json')
exports.xchangefile = path.resolve('config/', './exchangeRate.json')
exports.STCD = STCD
exports.RDBMSJSON = RDBMSJSON;
exports.GSTMAP = GSTMAP;
exports.XRATE = XRATE;
exports.ENV = ENV;
exports.FIELDCOLMAP = FIELDCOLMAP;
exports.KEY = KEY;
exports.LOV = LOV;
exports.CLIST = CLIST;
exports.MAPPER = MAPPER;
exports.PROD_QUERY_FIELD = 'id,prodid,status,productname,casno,unitselltype,priceperunit,type,makecountry,makecompany,reactivity,storgetype,structure,chemicalformula,chemicalname,currency,updated';
exports.LOV_QUERY_FIELD = 'id,status,lovname,lovvalue,updatedby,updated';
exports.KEY_QUERY_FIELD = 'id,status,keyname,keyvalue,updatedby,updated';
exports.COMP_QUERY_FIELD = 'id,status,name,address,phone,mobile,fax,idtype,idnum,country,email,defaultcurrency,registrationnum,registrationdate,license,updatedby,updated';
exports.TEMP = TEMP;
exports.Schema = Schema;

//audit function
exports.AUDIT = (valueObj, type, operation, tablename, fieldname, mapper, DBredirects, options, system_user, context) => {
    let audit = mapper.global.Schema.Audit;
    audit.type = type
    audit.operation = operation
    audit.tablename = tablename
    audit.fieldname = fieldname
    audit.value = valueObj
    let au = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings({
        data: [{
            Audit: audit
        }], System: {
            user: system_user
        }
    }), 'Audit');
    if (!au.status) {
        console.log(au.msg)
        return au.msg
    }
    au.data[0].Audit.created = au.data[0].Audit.updated
    au.data[0].Audit.createdby = au.data[0].Audit.updatedby
    DBredirects.insertData(mapper.mapper.forModel('Audit'), {}, au.data[0].Audit, { consistency: options.consistency }).then(data => {

        console.log('INFO' + ' ' + context + ' ' + new Date().toISOString() + ' ' + JSON.stringify(data))
    }).catch(err => {
        console.log(err);
    })

}

//communication entry
exports.communicationEntry = async (comtype, accntid, user, msg, mapper, DBredirects, options, system_user) => {
    //insert communication
    let comm = mapper.global.Schema.Communication
    comm.type = comtype
    comm.accntid = accntid
    comm.user = user
    comm.createdby = system_user
    comm.message = `${msg}. Thanks, PRASI LABS`
    let f = mapper.global.VALIDATE.Settings.getValidationStatus(mapper, new mapper.global.VALIDATE.Settings({
        data: [{
            Communication: comm
        }], System: {
            user: system_user
        }
    }), 'Communication');
    f.data[0].Communication.created = f.data[0].Communication.updated
    return {
        data: f.data[0].Communication
        , rs: (await DBredirects.insertData(mapper.mapper.forModel('Communication'), {}, f.data[0].Communication, { consistency: options.consistency }))._rs
    }
}

//Check user
exports.checkUser = async (client, mapper, options, param, body, headers, DBredirects) => {

    if (client === undefined) {
        throw new Error("Client can not be null");
    }

    if (!(await mapper.global.validateT(mapper, headers.authorization, body.System.user, 'API', DBredirects))) {
        throw new Error('Authentication failed. Please relogin using password');
    }

    let userR = await mapper.global.VALIDATE.userTyepValiation(DBredirects, mapper, body.System.user)

    if (!userR) {
        throw new Error('Something is worng. User not found.')
    }
    return userR
}

//Encryption
const encrypt = (payload, secret, algorithm, expiretime) => {
    if (!expiretime) {
        return jwt.sign({ data: payload }, secret, {
            algorithm: algorithm
        });
    }
    return jwt.sign({ data: payload }, secret, {
        algorithm: algorithm,
        expiresIn: expiretime // if ommited, the token will not expire
    });
}

//Decryption
const decrypt = (signed, secret, algorithm) => {
    return jwt.verify(signed, secret, {
        // Never forget to make this explicit to prevent
        // signature stripping attacks
        algorithms: [algorithm],
    });
}

//Login Validation
exports.validateT = async (mapper, input, userid, context, DBredirects) => {

    if (!mapper || !input || !userid) {
        return new Error('required parameters are missing')
    }
    switch (context) {
        case 'GENERATE':
            let inputObj = {}
            inputObj['passwordtoken'] = encrypt(input, mapper.global.KEY.LOGIN_KEY, mapper.global.KEY.ALGORITHM, '');
            inputObj['logintoken'] = encrypt(input + userid, mapper.global.KEY.API_USER_KEY, mapper.global.KEY.ALGORITHM, mapper.global.KEY.LOGIN_EXPIRE_TIME);
            inputObj['apitoken'] = encrypt(input + userid, mapper.global.KEY.API_CALL_KEY, mapper.global.KEY.ALGORITHM, mapper.global.KEY.API_EXPIRE_TIME);
            // console.log(inputObj)
            return inputObj;
        case 'API':
            let data1 = await DBredirects.findData(mapper.mapper.forModel('User'), { userid: userid }, { fields: ["userid", "apitoken", "status"] }, {})
            if (data1._rs.rows.filter(val => { return val.apitoken == input && val.status == 'ACTIVE' }).length == 1) {
                return true
            }
            return false
        case 'LOGIN':
            let data2 = await DBredirects.findData(mapper.mapper.forModel('User'),
                { userid: userid }, { fields: ["userid", "logintoken", "apitoken", "status", "type"] }, {})
            if (data2._rs.rows.filter(val => {
                return decrypt(val.logintoken,
                    mapper.global.KEY.API_USER_KEY,
                    mapper.global.KEY.ALGORITHM).data == decrypt(input,
                        mapper.global.KEY.API_USER_KEY,
                        mapper.global.KEY.ALGORITHM).data && val.status == 'ACTIVE'
            }).length == 1) {
                return {
                    userid: data2._rs.rows[0].userid,
                    logintoken: data2._rs.rows[0].logintoken,
                    apitoken: data2._rs.rows[0].apitoken,
                    type: data2._rs.rows[0].type
                }
            }
            return false
        case 'PASSWORD':
            return decrypt(input, mapper.global.KEY.LOGIN_KEY, mapper.global.KEY.ALGORITHM);
        default:
            return false
    }

}

/*************************************
 * 
 * Phone create functionlity 
 * 
 * ***********************************/
exports.validatePhone = async (searchObj, mapper, options, auditFunc, context, DBredirects, audit_all_fields, system_user) => {
    //Phone validtion
    let checkPhone = await DBredirects.findAllData(mapper.mapper.forModel('Phone'), {}, { fields: ['pid', 'status', 'phone', 'mobile', 'id'] }, { consistency: options.consistency })
    if (checkPhone._rs.rows.filter(obj => {
        return (searchObj.Phone.mobile == obj.mobile && searchObj.Phone.phone == obj.phone) && obj.status == 'ACTIVE'
    }).length < 1) {
        searchObj.Phone.created = searchObj.Phone.updated
        searchObj.Phone.createdby = searchObj.Phone.updatedby
        let r = await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Phone'), {}, searchObj.Phone, { consistency: options.consistency });
        //callback audit
        let value = { 'new_data': JSON.stringify(searchObj.Phone), 'action_result': JSON.stringify(r._rs.rows) }
        auditFunc(value, 'CURRENT', 'CREATE', 'Phone', audit_all_fields, mapper, DBredirects, options, system_user, `${context} Phone`);
    } else {
        return checkPhone.first()
    }
}

/*************************************
 * 
 * zip create functionlity 
 * 
 * ***********************************/
exports.validateZip = async (searchObj, mapper, options, auditFunc, context, DBredirects, audit_all_fields, system_user) => {
    let checkZip = await DBredirects.findAllData(mapper.mapper.forModel('Zip'), {}, {}, { consistency: options.consistency })
    if (checkZip._rs.rows.filter(obj => {
        return (searchObj.Zip.zip == obj.zip && searchObj.Zip.state == obj.state && searchObj.Zip.country == obj.country) && obj.status == 'ACTIVE'
    }).length < 1) {
        searchObj.Zip.created = searchObj.Zip.updated
        searchObj.Zip.createdby = searchObj.Zip.updatedby
        let r = await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Zip'), {}, searchObj.Zip, { consistency: options.consistency });

        //callback audit
        let value = { 'new_data': JSON.stringify(searchObj.Zip), 'action_result': JSON.stringify(r._rs.rows) }
        auditFunc(value, 'CURRENT', 'CREATE', 'Zip', audit_all_fields, mapper, DBredirects, options, system_user, `${context} Zip`);
    } else {
        return checkZip.first()
    }
}

/*************************************
 * 
 * Email create functionlity 
 * 
 * ***********************************/
exports.validateEmail = async (searchObj, mapper, options, auditFunc, context, DBredirects, audit_all_fields, system_user) => {
    //Get list of data of tables
    let checkEmail = await DBredirects.findAllData(mapper.mapper.forModel('Email'), {}, { fields: ['eid', 'status', 'email', 'id'] }, { consistency: options.consistency })

    //Email validation      
    if (checkEmail._rs.rows.filter(obj => {
        return (searchObj.Email.email == obj.email) && obj.status == 'ACTIVE'
    }).length < 1) {
        searchObj.Email.created = searchObj.Email.updated
        searchObj.Email.createdby = searchObj.Email.updatedby
        let r = (await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Email'), {}, searchObj.Email, { consistency: options.consistency }))._rs.rows;

        //callback audit
        let value = { 'new_data': JSON.stringify(searchObj.Email), 'action_result': JSON.stringify(r) }
        auditFunc(value, 'CURRENT', 'CREATE', 'Email', audit_all_fields, mapper, DBredirects, options, system_user, `${context}  Email`);
    } else {
        return checkEmail.first()
    }
}

//function returnig called function 
exports.callFunction = function (funcObjNameContext, ...args) {
    try {
        switch (funcObjNameContext) {
            /* Test connectivity */
            case 'connectDB':
                return controller.connectDB(...args);
            case 'testDB':
                return controller.checkDB(...args).then((data) => {
                    return data
                }).catch((err) => {
                    return err;
                })
            /*common calls*/
            case 'getAllProducts':
                return controller.getAllProducts(...args).then((data) => {
                    return data
                }).catch((err) => {
                    return err;
                })
            case 'getProductsBySearch':
                return controller.getProductsBySearch(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getProductsByName':
                return controller.getProductsByName(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getProductsByCasno':
                return controller.getProductsByCasno(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getProductsByPartner':
                return controller.getProductsByPartner(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getProductsByCountry':
                return controller.getProductsByCountry(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getProductsByCategory':
                return controller.getProductsByCategory(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getGSTList':
                return controller.getGSTList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getPartnerList':
                return controller.getPartnerList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getCategoryList':
                return controller.getCategoryList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getCountryList':
                return controller.getCountryList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getCurrencyList':
                return controller.getCurrencyList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getXchangeList':
                return controller.getXchangeList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getCurrencyList':
                return controller.getCurrencyList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            /* login logout calls */
            case 'login':
                return controller.login(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'logout':
                return controller.logout(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            /*user calls*/
            case 'getAccountList':
                return controller.getAccountList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getContactList':
                return controller.getContactList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getAddressList':
                return controller.getAddressList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getCommunicationList':
                return controller.getCommunicationList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getInvoiceList':
                return controller.getInvoiceList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getStat':
                return controller.getStat(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getSRList':
                return controller.getSRList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getLoginList':
                return controller.getLoginList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getOrderList':
                return controller.getOrderList(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'addToCart':
                return controller.addToCart(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postAccount':
                return controller.postAccount(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'updateAccount':
                return controller.updateAccount(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postContact':
                return controller.postContact(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postAddress':
                return controller.postAddress(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postCommunication':
                return controller.postCommunication(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postSR':
                return controller.postSR(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postUser':
                return controller.postUser(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postOrder':
                return controller.postOrder(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            /*admin calls*/
            case 'getAllAdminProducts':
                return controllerAdmin.getAllAdminProducts(...args).then((data) => {
                    return data
                }).catch((err) => {
                    return err;
                })
            case 'postInvoiceAdmin':
                return controllerAdmin.postInvoiceAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postProductAdmin':
                return controllerAdmin.postProduct(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getKeyAdmin':
                return controllerAdmin.getKeyAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getGST':
                return controllerAdmin.getGST(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getLovAdmin':
                return controllerAdmin.getLovAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getCompanyAdmin':
                return controllerAdmin.getCompanyAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postKeyAdmin':
                return controllerAdmin.postKeyAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postLovAdmin':
                return controllerAdmin.postLovAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postCompanyAdmin':
                return controllerAdmin.postCompanyAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postXchangeAdmin':
                return controllerAdmin.postXchangeAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postGSTAdmin':
                return controllerAdmin.postGSTAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })

            case 'bulkproducts':
                return controllerAdmin.bulkproducts(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'bulkgst':
                return controllerAdmin.bulkgst(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'bulkinv':
                return controllerAdmin.bulkinv(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getSRListAdmin':
                return controllerAdmin.getSRListAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getCommunicationListAdmin':
                return controllerAdmin.getCommunicationListAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getOrderListAdmin':
                return controllerAdmin.getOrderListAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getInvoiceListAdmin':
                return controllerAdmin.getInvoiceListAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'getLedgerListAdmin':
                return controllerAdmin.getLedgerListAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'ledger':
                return controllerAdmin.ledger(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'postOrderitemAdmin':
                return controllerAdmin.postOrderitemAdmin(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'adminDataOrders':
                return controllerAdmin.adminDataOrders(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'adminDataInvoice':
                return controllerAdmin.adminDataInvoice(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'adminDataLedger':
                return controllerAdmin.adminDataLedger(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'adminDataComm':
                return controllerAdmin.adminDataComm(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'adminDataSR':
                return controllerAdmin.adminDataSR(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'sendemail':
                return controllerAdmin.sendemail(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            case 'downloadInvoice':
                return controller.downloadInvoice(...args).then((data) => {
                    return data;
                }).catch((err) => {
                    return err;
                })
            default:
                return "Nothing to do."
        }
    } catch (err) {
        return err;
    }
}