
/*
Created by: Peeyush
Created on: July 22 2021
Email: j.k.peeyush@gmail.com
Description: GST Calculation
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
class GST {
    #baseprice = 0;
    #igstrate = 0;
    #cgstrate = 0;
    #sgstrate = 0;
    #othertax = 0;
    #discount = 0;
    #shipcharge = 0;
    #isurance = 0;
    #others = 0;
    #isCalled = true;
    country = '';
    state = '';
    //constructor
    constructor(c, s) {
        this.country = c
        this.state = s
    }
    //fixing price to two decimal places
    twoDec(val) {
        return Number(Number(val).toFixed(2))
    }

    //function to set values of each price
    setValues(baseprice, igstrate, cgstrate, sgstrate, othertax, discount, shipcharge, isurance, others) {
        'use strict';
        this.#isCalled = false;
        typeof baseprice === 'number' ? this.#baseprice = baseprice : (() => { throw "number type expected"; })();
        typeof igstrate === 'number' ? this.#igstrate = igstrate : (() => { throw "number type expected"; })();
        typeof cgstrate === 'number' ? this.#cgstrate = cgstrate : (() => { throw "number type expected"; })();
        typeof sgstrate === 'number' ? this.#sgstrate = sgstrate : (() => { throw "number type expected"; })();
        typeof othertax === 'number' ? this.#othertax = othertax : (() => { throw "number type expected"; })();
        typeof discount === 'number' ? this.#discount = discount : (() => { throw "number type expected"; })();
        typeof shipcharge === 'number' ? this.#shipcharge = shipcharge : (() => { throw "number type expected"; })();
        typeof isurance === 'number' ? this.#isurance = isurance : (() => { throw "number type expected"; })();
        typeof others === 'number' ? this.#others = others : (() => { throw "number type expected"; })();
    }
    //return net tax of each order item
    netItemTax() {
        'use strict';
        if (this.#isCalled) {
            throw "First set values by calling setValues(baseprice, igstrate, cgstrate, sgstrate, othertax, discount, shipcharge, isurance, others) on the object instance.";
        }

        if (this.country.toUpperCase() != 'IN') {

            return ((this.#othertax) * (this.#baseprice - this.#discount) / 100 * 1.0);
        }
        else if ((this.country.toUpperCase() == 'IN') && (this.state.toUpperCase() == 'TS')) {

            return (((this.#cgstrate + this.#sgstrate) / 100) * (this.#baseprice - this.#discount) * 1.0);
        }
        else {
            return (((this.#igstrate)) * (this.#baseprice - this.#discount) / 100 * 1.0);
        }

    }
    //return gross tax of each item
    grossItemTax() {
        'use strict';
        if (this.#isCalled) {
            throw "First set values by calling setValues(baseprice, igstrate, cgstrate, sgstrate, othertax, discount, shipcharge, isurance, others) on the object instance.";
        }
        if (this.country.toUpperCase() != 'IN') {

            return ((this.#othertax) * (this.#baseprice) / 100 * 1.0);
        }
        else if ((this.country.toUpperCase() == 'IN') && (this.state.toUpperCase() == 'TS')) {
            return (((this.#cgstrate + this.#sgstrate) / 100) * (this.#baseprice) * 1.0);
        }
        else {
            return (((this.#igstrate)) * (this.#baseprice) / 100 * 1.0);
        }
    }
    //return gross amount of each item
    grossItemAmont() {
        'use strict';
        return (this.#baseprice + this.grossItemTax());
    }
    //return net amount of each item
    netItemAmont() {
        'use strict';
        return ((this.#baseprice - this.#discount) + this.netItemTax());
    }
    //return total gross amount of the order
    getGrossAmount(obj, insurance, shipcharge, others) {
        'use strict';
        if (!Array.isArray(obj)) {
            throw "Expected array of JSON object";
        }
        else {
            let grossAmout = 0
            obj.forEach(element => {
                this.setValues(element['baseprice'], element['igstrate'], element['cgstrate'], element['sgstrate'], element['othertax'], element['discount'], shipcharge, insurance, others);
                grossAmout = grossAmout + this.grossItemAmont();
            });
//assume to take last set value of taxes and all products have same type of taxes charges in a single invoice
            if (this.country.toUpperCase() != 'IN') {

                grossAmout = grossAmout + ((this.#othertax) * (this.#shipcharge) / 100 * 1.0);
            }
            else if ((this.country.toUpperCase() == 'IN') && (this.state.toUpperCase() == 'TS')) {

                grossAmout = grossAmout + (((this.#cgstrate + this.#sgstrate) / 100) * (this.#shipcharge) * 1.0);
            }
            else {
                grossAmout = grossAmout + (((this.#igstrate)) * (this.#shipcharge) / 100 * 1.0);
            }

            return (grossAmout + this.#others + this.#isurance + this.#shipcharge);
        }
    }

    //return total net amount of the order
    getNetAmount(obj, insurance, shipcharge, others) {
        'use strict';
        if (!Array.isArray(obj)) {
            throw "Expected array of JSON object";
        }
        else {
            let netAmout = 0;
            obj.forEach(element => {
                this.setValues(element['baseprice'], element['igstrate'], element['cgstrate'], element['sgstrate'], element['othertax'], element['discount'], shipcharge, insurance, others);
                netAmout = netAmout + this.netItemAmont();
            });
            
//assume to take last set value of taxes and all products have same type of taxes charges in a single invoice

            if (this.country.toUpperCase() != 'IN') {

                netAmout = netAmout + ((this.#othertax) * (this.#shipcharge) / 100 * 1.0);
            }
            else if ((this.country.toUpperCase() == 'IN') && (this.state.toUpperCase() == 'TS')) {

                netAmout = netAmout + (((this.#cgstrate + this.#sgstrate) / 100) * (this.#shipcharge) * 1.0);
            }
            else {
                netAmout = netAmout + (((this.#igstrate)) * (this.#shipcharge) / 100 * 1.0);
            }


            return (netAmout + this.#others + this.#isurance + this.#shipcharge);
        }
    }

    //get net GST amount
    getNetGst(obj, insurance, shipcharge, others) {
        'use strict';
        if (!Array.isArray(obj)) {
            throw "Expected array of JSON object";
        }
        else {
            let netGst = 0;
            obj.forEach(element => {
                this.setValues(element['baseprice'], element['igstrate'], element['cgstrate'], element['sgstrate'], element['othertax'], element['discount'], shipcharge, insurance, others);
                netGst = netGst + this.netItemTax();
            });
            return (netGst);
        }
    }

    //get gross GST amount
    getgrossGst(obj, insurance, shipcharge, others) {
        'use strict';
        if (!Array.isArray(obj)) {
            throw "Expected array of JSON object";
        }
        else {
            let grossGst = 0;
            obj.forEach(element => {
                this.setValues(element['baseprice'], element['igstrate'], element['cgstrate'], element['sgstrate'], element['othertax'], element['discount'], shipcharge, insurance, others);
                grossGst = grossGst + this.grossItemTax();
            });
            return (grossGst);
        }
    }

    //Get total order amount segrigation
    getAllAmount(obj, insurance, shipcharge, others) {
        if (!Array.isArray(obj)) {
            throw "Expected array of JSON object";
        }
        else {
            let allAmout = { grossGst: 0, netGst: 0, grossAmount: 0, netAmount: 0 };
            allAmout['grossGst'] = this.getgrossGst(obj, insurance, shipcharge, others);
            allAmout['netGst'] = this.getNetGst(obj, insurance, shipcharge, others);
            allAmout['grossAmount'] = this.getGrossAmount(obj, insurance, shipcharge, others);
            allAmout['netAmount'] = this.getNetAmount(obj, insurance, shipcharge, others);
            return allAmout;
        }
    }
    //get item amount segrigation
    getItemAllAmount() {
        let allAmout = { grossGst: 0, netGst: 0, grossAmount: 0, netAmount: 0 };
        allAmout['grossGst'] = this.grossItemTax();
        allAmout['netGst'] = this.netItemTax();
        allAmout['grossAmount'] = this.grossItemAmont()
        allAmout['netAmount'] = this.netItemAmont();
        return allAmout;
    }
}

exports.GST = GST

//convert currency
exports.calculteXrate = (fromAmount, fromCurrency, toCUrrency, rateChartObj) => {
    let amount = -1
    let status = false
    if (rateChartObj) {
        if (rateChartObj[fromCurrency]) {
            if (rateChartObj[fromCurrency][toCUrrency]) {
                amount = fromCurrency == toCUrrency ? fromAmount : rateChartObj[fromCurrency][toCUrrency] * fromAmount;
                status = true
            }
        } else {
            if (rateChartObj[toCUrrency]) {
                if (rateChartObj[toCUrrency][fromCurrency]) {
                    amount = fromCurrency == toCUrrency ? fromAmount : (1.0 / rateChartObj[toCUrrency][fromCurrency]) * fromAmount;
                    status = true
                }
            }
            else {
                for (let i = 0; i < 1; i++) {

                    let F = rateChartObj[Object.keys(rateChartObj)[0]][fromCurrency]
                    let T = rateChartObj[Object.keys(rateChartObj)[0]][toCUrrency]

                    if (F && T) {
                        amount = fromAmount * (T / F)
                        status = true
                    }
                }
            }
        }
    }

    if (!status) {
        throw 'conversion failed.'
    }
    return amount
}

exports.caclulateInvoiceData = (invobj, prodlist, gstmap, gstObj, xrate, traveresPath, GST, state) => {


    let gst = new GST.GST(invobj.bill.country, invobj.bill.state);

    gstObj.shipcharge = Number(invobj.shippingcharges)
    gstObj.isurance = Number(invobj.othercharges)
    gstObj.others = Number(invobj.surcharge)

    let itemLIst = []

    for (let i = 0; i < invobj.item.length; i++) {

        let x = prodlist.filter(pitem => {
            return pitem.prodid == JSON.parse(invobj.item[i]).itemprodid
        })[0]

        let baseprice = 0,
            igstrate = 0,
            cgstrate = 0,
            sgstrate = 0,
            othertax = 0,
            discount = 0
        let gstrate = [];
        baseprice = gst.twoDec((JSON.parse(invobj.item[i]).itemqty) * (Number(x.priceperunit) != 0 ? Number(x.priceperunit) : Number(JSON.parse(invobj.item[i]).itemprice)))
        //Note in this section plase do not use  / or . as separator
        //   console.log(invobj.item[i].itemprice)
        if ((state == "36") && (invobj.bill.country.toUpperCase() == "IN")) {

            traveresPath(gstmap, `${x.prodid}#cgst`, '#', gstrate)
            cgstrate = Number(gstrate[0])

            traveresPath(gstmap, `${x.prodid}#sgst`, '#', gstrate)
            sgstrate = Number(gstrate[0])

            igstrate = 0
            othertax = 0

        } else if (state != "36" && invobj.bill.country.toUpperCase() == "IN") {

            traveresPath(gstmap, `${x.prodid}#igst`, '#', gstrate)
            igstrate = Number(gstrate[0])

            sgstrate = 0
            cgstrate = 0
            othertax = 0

        } else if (invobj.bill.country.toUpperCase() != "IN") {

            traveresPath(gstmap, `${x.prodid}#tax`, '#', gstrate)

            othertax = Number(gstrate[0])

            sgstrate = 0
            cgstrate = 0
            igstrate = 0

        } else {

            return new Error('Bill location is not specified. Invoice could not be generated. Please generte manually.')

        }

        discount = gst.twoDec(Number(JSON.parse(invobj.item[i]).itemdiscount) * Number(JSON.parse(invobj.item[i]).itemqty))

        gst.setValues(baseprice, igstrate, cgstrate, sgstrate, othertax, discount, gstObj.shipcharge, gstObj.isurance, gstObj.others)

        let amount = gst.getItemAllAmount();

        // console.log(amount)
        if (
            gst.twoDec(amount.netGst)
            !=

            (Number(JSON.parse(invobj.item[i]).itemcgst)
                + Number(JSON.parse(invobj.item[i]).itemsgst)
                + Number(JSON.parse(invobj.item[i]).itemigst)
                + Number(JSON.parse(invobj.item[i]).itemtax))
        ) {

            return new Error('There is a mismatch in tax amount sent and server calculated value. Invoice could not be generated. Please generte manually.')
        }
        if (
            gst.twoDec(amount.netAmount - amount.netGst)
            !=
            Number(JSON.parse(invobj.item[i]).itemamount)
        ) {
            return new Error('There is a mismatch in item amount sent and server calculated value. Invoice could not be generated. Please generte manually.')
        }

        itemLIst.push({
            baseprice: baseprice,
            igstrate: igstrate,
            cgstrate: cgstrate,
            sgstrate: sgstrate,
            discount: discount,
            othertax: othertax,
            isurance: gstObj.isurance,
            shipcharge: gstObj.shipcharge,
            others: gstObj.others
        });

    }
    // console.log(itemLIst)
    let totalAmount = gst.getAllAmount(itemLIst, gstObj.isurance, gstObj.shipcharge, gstObj.others)

    //console.log(totalAmount,invobj.finalamount)
    if (
        Math.round(totalAmount.netAmount)
        !=
        Math.round(Number(invobj.finalamount))
    ) {
        return new Error('There is a mismatch in total order amount sent and server calculated value. Invoice could not be generated. Please generte manually.')
    }

    if (
        gst.twoDec(totalAmount.netGst)
        !=
        (Number(invobj.csgt) + Number(invobj.sgst) + Number(invobj.igst) + Number(invobj.tax))
    ) {
        return new Error('There is a mismatch in total tax amount sent and server calculated value. Invoice could not be generated. Please generte manually.')
    }
    //console.log(state)
    return 1
}