/*
Created by: Peeyush
Created on: March 23 2022
Email: j.k.peeyush@gmail.com
Description: invoice creation function
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////

//Invoice creation
const createPdf = require('./pdfCreate');
const createPO = require('./testPDFCreate');
const path = require('path');
const createInvoice = async (mapper, options, obj, bodySystem, auditFunc, DBredirects, audit_all_fields, system_user) => {
    console.log('Creating Invoice...')
    let invoiceObj = mapper.global.Schema.Invoice;
    //initializes item array
    invoiceObj.item = [];
    let orderObj = obj;
    let insertS1 = ''
    let insertS2 = ''
    let insertS3 = ''
    //assing order level detail to invoice
    invoiceObj.currency = orderObj.Orders.currency
    invoiceObj.invdate = new Date().toISOString();
    invoiceObj.accntid = orderObj.Orders.accntid
    invoiceObj.ordernum = orderObj.Orders.orderid
    invoiceObj.poordernum = orderObj.Orders.poordernum
    invoiceObj.poorderdate = orderObj.Orders.poorderdate
    invoiceObj.status = 'NEW'
    invoiceObj.quotenum = orderObj.Orders.quotenum
    invoiceObj.quotedate = orderObj.Orders.quotedate
    invoiceObj.ispo = orderObj.Orders.direction == 'OUT' ? true : false
    invoiceObj.bill = orderObj.Orders.bill
    invoiceObj.ship = orderObj.Orders.ship
    invoiceObj.totalamount = Number(orderObj.Orders.orderamount.toFixed(2))
    invoiceObj.finalamount = Math.round(Number(orderObj.Orders.finalamount.toFixed(2)))
    invoiceObj.surcharge = Number(orderObj.Orders.surcharge.toFixed(2))
    invoiceObj.otherdiscount = Number(orderObj.Orders.discount.toFixed(2))
    invoiceObj.othercharges = Number(orderObj.Orders.insurenceamount.toFixed(2))
    invoiceObj.shippingcharges = Number(orderObj.Orders.shippingamount.toFixed(2))
    invoiceObj.shippingdetail = orderObj.Orders.shippingdetail
    invoiceObj.shipgst = orderObj.Orders.shipgst,
        invoiceObj.csgt = Number(orderObj.Orders.cgstamount.toFixed(2))
    invoiceObj.sgst = Number(orderObj.Orders.sgstamount.toFixed(2))
    invoiceObj.igst = Number(orderObj.Orders.igstamount.toFixed(2))
    invoiceObj.tax = Number(orderObj.Orders.taxamount.toFixed(2))
    invoiceObj.orderdate = new Date(orderObj.Orders.updated).toLocaleDateString();

    //get company detail for invoicing
    let companyDetail = (await DBredirects.findAllData(mapper.mapper.forModel('Company'), {
        fields: ["name", "address", "phone", "mobile", "country", "email", "idtype", "idnum"]
    }, {}, { consistency: options.consistency }))._rs.rows.filter(obj => {
        return obj.name == 'PRASI LAB INDIA'
    })

    //map data to invoice object
    let item = mapper.global.Schema.invoiceitem;
    let prodlist = [];
    let p;
    for (let element = 0; element < orderObj.Orderitem.length; element++) {
        if (orderObj.Orderitem[element].pcid) {
            p = await DBredirects.findSingleData(mapper.mapper.forModel('Product'), {
                id: orderObj.Orderitem[element].pcid,
                prodid: orderObj.Orderitem[element].prodid
            }, {
                fields: ['status', 'prodid', 'casno', 'productname', 'makecompany', 'priceperunit', 'unitselltype', 'type', 'currency']
            }, { consistency: options.consistency })

        } else {
            p = (await DBredirects.findAllData(mapper.mapper.forModel('Product'), {}, {
                fields: ['status', 'prodid', 'casno', 'productname', 'makecompany', 'priceperunit', 'unitselltype', 'type', 'currency']
            }, { consistency: options.consistency }))._rs.rows.filter((item) => {

                return item.prodid == orderObj.Orderitem[element].prodid && item.status.toUpperCase() == 'ACTIVE'
            })[0]

        }

        prodlist.push(p)
        //generate product detail


        //extract item detail into invoice
        item.itemname = orderObj.Orderitem[element].itemname
        item.itemqty = orderObj.Orderitem[element].quantity
        item.itemprice = Number(orderObj.Orderitem[element].itemprice.toFixed(2))
        item.itemcgst = Number(orderObj.Orderitem[element].cgstamount.toFixed(2))
        item.itemsgst = Number(orderObj.Orderitem[element].sgstamount.toFixed(2))
        item.itemigst = Number(orderObj.Orderitem[element].igstamount.toFixed(2))
        item.itemtax = Number(orderObj.Orderitem[element].taxamount.toFixed(2))
        item.itemcgstrate = Number(orderObj.Orderitem[element].cgstrate.toFixed(2))
        item.itemsgstrate = Number(orderObj.Orderitem[element].sgstrate.toFixed(2))
        item.itemigstrate = Number(orderObj.Orderitem[element].igstrate.toFixed(2))
        item.itemtaxrate = Number(orderObj.Orderitem[element].taxrate.toFixed(2))
        item.itemdescription = orderObj.Orderitem[element].itemdetail
        item.itemsellunit = p.unitselltype
        item.currency = orderObj.Orderitem[element].currency
        item.itemamount = Number(orderObj.Orderitem[element].itemamount.toFixed(2))
        item.itemdiscount = Number(orderObj.Orderitem[element].discount.toFixed(2))
        item.itemprodid = orderObj.Orderitem[element].prodid
        //push to invoice item property array
        invoiceObj.item.push(JSON.stringify(item))
    }

    let stateCode = []
    let state = 0

    mapper.global.TRAVERSEJSON.traveres(global.STCD, '', stateCode);

    for (var g = 0; g < Object.keys(stateCode).length; g++) {
        if (stateCode['.' + g + '.short'] == invoiceObj.bill.state) {
            state = stateCode['.' + g + '.code']
        }
    }

    //invoice calculation
    let z = '';
    try {
        z = mapper.global.GST.caclulateInvoiceData(invoiceObj, prodlist, mapper.global.GSTMAP, mapper.global.Schema.gstobj, mapper.global.XRATE, mapper.global.TRAVERSEJSON.traveresPath, mapper.global.GST, state);

        if (z != 1) {

            return z
        }
    } catch (error) {
        console.trace(error)
        return error
    }


    invoiceObj.bill['statecode'] = state;
    //validate structure
    let finalInv = mapper.global.VALIDATE.Settings.getValidationStatus(mapper,
        new mapper.global.VALIDATE.Settings({
            data: [{
                Invoice:
                    invoiceObj
            }],
            System: bodySystem
        }), 'Invoice');

    if (!finalInv.status) {
        console.trace(finalInv.status)
        return new Error(`Validation failed. Invoice gneration failed. ${finalInv.msg}`);
    }

    finalInv.data[0].Invoice.created = finalInv.data[0].Invoice.updated
    finalInv.data[0].Invoice.createdby = finalInv.data[0].Invoice.updatedby
    finalInv.data[0].Invoice.invmaildetail = mapper.global.Schema.invoiceheaders;
    finalInv.data[0].Invoice.invmaildetail.itemheader = JSON.stringify(companyDetail);
    finalInv.data[0].Invoice.invmaildetail.itemfooter = mapper.global.KEY.INV_SIG_NOTE;

    /*
    create origional invoice
    */
    finalInv.data[0].Invoice.type = 'ORIGIONAL';
    //create invoice


    insertS1 = (await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Invoice'), {

    }, finalInv.data[0].Invoice, { consistency: options.consistency }))._rs.rows
    //callback audit

    let value = { 'new_data': JSON.stringify(finalInv.data[0].Invoice), 'action_result': JSON.stringify(insertS1) }
    auditFunc(value, 'CURRENT', 'CREATE', 'Invoice', audit_all_fields, mapper, DBredirects, options, system_user, 'Invoice');


    if (!invoiceObj.ispo) {

        console.log('creating origional Invoice doc')
        // console.log(finalInv.data[0].Invoice);
        createPdf(companyDetail, finalInv.data[0].Invoice);
        /*
          create duplicate invoice
        */
        console.log('creating duplicate Invoice doc')
        finalInv.data[0].Invoice.type = 'DUPLICATE';
        //create invoice
        insertS2 = (await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Invoice'), {}, finalInv.data[0].Invoice, { consistency: options.consistency }))._rs.rows
        //callback audit
        value = { 'new_data': JSON.stringify(finalInv.data[0].Invoice), 'action_result': JSON.stringify(insertS2) }
        auditFunc(value, 'CURRENT', 'CREATE', 'Invoice', audit_all_fields, mapper, DBredirects, options, system_user, 'Invoice');

        // console.log('creating Invoice doc')
        createPdf(companyDetail, finalInv.data[0].Invoice);

        /*
          create triplicate invoice
        */
        console.log('creating triplicate Invoice doc')
        finalInv.data[0].Invoice.type = 'TRIPLICATE';
        //create invoice
        insertS3 = (await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Invoice'), {}, finalInv.data[0].Invoice, { consistency: options.consistency }))._rs.rows
        //callback audit
        value = { 'new_data': JSON.stringify(finalInv.data[0].Invoice), 'action_result': JSON.stringify(insertS3) }
        auditFunc(value, 'CURRENT', 'CREATE', 'Invoice', audit_all_fields, mapper, DBredirects, options, system_user, 'Invoice');

        // console.log('creating Invoice doc')
        createPdf(companyDetail, finalInv.data[0].Invoice);
    } else {
        console.log('creating PO document')
        createPO(companyDetail, finalInv.data[0].Invoice);
    }
    //create communication
    let msg = `Your invoice has been successfully generated.`
    let com = (
        await mapper.global.communicationEntry('INBOUND', finalInv.data[0].Invoice.invid, finalInv.data[0].Invoice.updatedby, msg, mapper, DBredirects, options, system_user)
    )
    //callback audit
    value = {
        'new_data': JSON.stringify(com.data),
        'action_result': JSON.stringify(
            com.rs
        )
    }
    auditFunc(value, 'CURRENT', 'CREATE', 'Communication', audit_all_fields, mapper, DBredirects, options, system_user, 'Communication')

    //entry in ledger
    let ledger = mapper.global.Schema.Ledger;

    ledger.id = ''
    ledger.status = 'Open'
    ledger.type = invoiceObj.ispo ? 'Debit' : 'Credit'
    ledger.source = 'Order'
    ledger.currency = finalInv.data[0].Invoice.currency
    ledger.amount = finalInv.data[0].Invoice.finalamount
    ledger.dsc = `Amount for order ${finalInv.data[0].Invoice.ordernum} and PO ${finalInv.data[0].Invoice.poordernum}`

    let finalL = mapper.global.VALIDATE.Settings.getValidationStatus(mapper,
        new mapper.global.VALIDATE.Settings({
            data: [{
                Ledger:
                    ledger
            }],
            System: bodySystem
        }), 'Ledger');

    if (!finalL.status) {
        console.trace(finalL.status)
    } else {
        finalL.data[0].Ledger.created = finalL.data[0].Ledger.updated
        finalL.data[0].Ledger.createdby = finalL.data[0].Ledger.updatedby
        let updateS = (await DBredirects.insertDataIfNotPresent(mapper.mapper.forModel('Ledger'), {}, finalL.data[0].Ledger, { consistency: options.consistency }))._rs.rows;
        let valuel = {
            'new_data': JSON.stringify(ledger),
            'action_result': JSON.stringify(updateS)
        }

        auditFunc(valuel, 'CHANGE', 'UPSERT', 'Ledger', audit_all_fields, mapper, DBredirects, options, system_user, 'UPSERT-LEDGER');


    }

    //
    console.log('Invoice creation completed.')
    return JSON.stringify({ insertS1, insertS2, insertS3 })
}
module.exports = createInvoice;