/*
Created by: Peeyush
Created on: June 15 2022
Email: j.k.peeyush@gmail.com
Description: pdf invoice creation function
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////

const fs = require('fs');
const path = require('path');

const numToWords = require('./numToWord')

var fonts = {
    Roboto: {
        normal: `${path.resolve(__dirname + '/fonts/Roboto-Regular.ttf')}`,
        bold: `${path.resolve(__dirname + '/fonts/Roboto-Medium.ttf')}`,
        italics: `${path.resolve(__dirname + '/fonts/Roboto-Italic.ttf')}`,
        bolditalics: `${path.resolve(__dirname + '/fonts/Roboto-MediumItalic.ttf')}`
    }
};

const PdfPrinter = require('pdfmake');
const printer = new PdfPrinter(fonts);

//Main document definition
var docDefinition = {
    info: {
        title: 'PO document',
        author: 'PRASI LABS',
        subject: 'Purchase Order',
        keywords: 'Purchase Order',
        creator: 'PRASI LABS',
        producer: 'PRASI LABS',
        modDate: `${new Date().toISOString()}`
    },
    pageSize: 'A4',
    pageOrientation: 'portrait',
    pageMargins: [25, 25, 25, 25],
    pageBreakBefore: function (currentNode, followingNodesOnPage, nodesOnNextPage, previousNodesOnPage) {
        return currentNode.headlineLevel === 1 && followingNodesOnPage.length === 0;
    },
    header: {
        layout: "noBorders",
        fontSize: 8,
        margin: [40, 2, 40, 1],
        table: {
            widths: ["*"],
            body: [
                [{ text: "Purchase Order" }]
            ]
        }
    },
    footer: (currentPage, pageCount) => {
        var t = {
            layout: "noBorders",
            fontSize: 8,
            margin: [40, 1, 40, 1],
            table: {
                widths: ["*"],
                body: [
                    [
                        { text: "Page  " + currentPage.toString() + " of " + pageCount }
                    ]
                ]
            }
        };

        return t;
    }
    ,
    watermark: { text: 'PRASI LABS', opacity: 0.05 },
    styles: {
        header: {
            fontSize: 9,
            bold: false,
            //   margin: 1
        },
        subheader: {
            fontSize: 9,
            bold: false,
            //  margin: 1,
            color: 'black',
            fillColor: 'white'
        },
        tableHeader: {
            bold: true,
            fontSize: 9,
            color: 'white',
            fillColor: 'black'
        },
        tableContent: {
            bold: false,
            fontSize: 9,
            color: 'black',
            fillColor: 'white'
        },
        tableContent1: {
            bold: false,
            fontSize: 9,
            color: 'black'
        },
        tableExample: {
            widths: '*',
        }
    },
    defaultStyle: {
        alignment: 'justify'
    },
    content: [
        {
            style: 'tableExample',
            table: {
                // headerRows: 1,
                // dontBreakRows: true,
                // keepWithHeaderRows: 1,
                widths: ['16%', '12%', '12%', '12%', '12%', '12%', '12%', '12%'],
                body: []
            },
            layout: {
                hLineWidth: function (i, node) {
                    return (i === 0 || i === node.table.body.length) ? 2 : 1;
                },
                vLineWidth: function (i, node) {
                    return (i === 0 || i === node.table.widths.length) ? 2 : 1;
                },
                hLineColor: function (i, node) {
                    return (i === 0 || i === node.table.body.length) ? 'black' : 'gray';
                },
                vLineColor: function (i, node) {
                    return (i === 0 || i === node.table.widths.length) ? 'black' : 'gray';
                },
                fillColor: function (rowIndex, node, columnIndex) {
                    return (rowIndex % 2 === 0) ? '#eeeeee' : null;
                },
                // hLineStyle: function (i, node) { return {dash: { length: 10, space: 4 }}; },
                // vLineStyle: function (i, node) { return {dash: { length: 10, space: 4 }}; },
                // paddingLeft: function(i, node) { return 4; },
                // paddingRight: function(i, node) { return 4; },
                // paddingTop: function(i, node) { return 2; },
                // paddingBottom: function(i, node) { return 2; },
            }
        }
    ]
};

//pdf option
var options = {
    fontLayoutCache: true,
    bufferPages: false
}

generateotherInfoText = (otherinfo) => {
    return Object.keys(otherinfo)
        .reduce((pre, cur) => {
            return pre = pre + cur + ': ' + otherinfo[cur] + '\n'
        }, '')
}
//function to transform invoice data into pdfmake style
transformData = (obj) => {
    let data = []
    obj.item.forEach((item) => {
        let tItem = JSON.parse(item)
        data.push([
            {
                text: `${tItem.itemname}\n
                CAS#: ${tItem.itemdescription.casno}  Rate: ${tItem.itemprice}/${tItem.itemsellunit}
                CGST:${tItem.itemcgstrate}  SGST:${tItem.itemsgstrate}  IGST:${tItem.itemigstrate}  TAX(Others):${tItem.itemtaxrate} `,
                style: 'tableContent1', colSpan: 3,
                alignment: 'left'
            },
            '',
            '',
            { text: `${tItem.itemdescription.hsncode}`, style: 'tableContent1', alignment: 'left' },
            { text: `${tItem.itemqty}`, style: 'tableContent1', alignment: 'left' },
            { alignment: 'left', text: `${Number(tItem.itemcgst) + Number(tItem.itemsgst) + Number(tItem.itemigst) + Number(tItem.itemtax)}`, style: 'tableContent1' },
            { text: `${tItem.itemamount}`, style: 'tableContent1', alignment: 'left' },
            { text: `${tItem.currency}`, style: 'tableContent1', alignment: 'left' }
        ]
        )
    }
    );
    return data
}

//function to create odf document
createPDFDocPO = (invnum, custnum, type, docDefinition, options) => {
    var pdfDoc = printer.createPdfKitDocument(docDefinition, options);
    try {
        if (fs.existsSync(path.join(__dirname, `../fs/PO_${invnum}_${custnum}.pdf`))) {
            fs.renameSync(path.join(__dirname, `../fs/PO_${invnum}_${custnum}.pdf`), path.join(__dirname, `../fsold/Old_${Date.now()}_PO_${invnum}_${custnum}.pdf`))
            console.log('Moved old file')
        }

        pdfDoc.pipe(fs.createWriteStream(path.join(__dirname, `../fs/PO_${invnum}_${custnum}.pdf`)));
        pdfDoc.end();
        console.log('created pdf');

    } catch (error) {
        console.log(error);
    }

}

//exported main function
const createPdfPO = (companyDetailObj, invObj) => {
    try {
        //variable definition
        image = path.join(__dirname, '/fonts/Prasilabs.png')
        image1 = path.join(__dirname, '/fonts/makeininda.png')
        image2 = path.join(__dirname, '/fonts/start-up.png')

        quotenum = invObj.quotenum
        quotedate = invObj.quotedate ? (new Date(invObj.quotedate)).toLocaleDateString() : ''

        freight = invObj.shippingcharges
        freightDetail = invObj.shippingdetail ? invObj.shippingdetail : ''

        invoiceNum = invObj.invid;
        customerId = invObj.accntid;
        shippedTo = `${invObj.ship.contact}\n${invObj.ship.address}`;
        billedTo = `${invObj.bill.contact}\n${invObj.bill.address}`;
        compInfo = 'PRASI LABS PVT LIMITED\nPlot no: 3647, Sai Krishana Layout, Near Pragati Nagar Arch,\nPragati Nagar, Hyderabad, Telangana - 500090, Bharat (India)\nM: +91 9160006515\nE: info@prasilabs.com'

        payAmount = Number(invObj.totalamount.toFixed(2));
        netAmount = Math.round(invObj.finalamount);
        totalDiscount = Number(invObj.otherdiscount.toFixed(2));
        netTax = Number(((Number(invObj.shipgst) * Number(invObj.shippingcharges)) + Number(invObj.csgt) + Number(invObj.sgst) + Number(invObj.igst) + Number(invObj.tax)).toFixed(2));
        grossAmount = Number((Number(invObj.totalamount) - Number(netTax)).toFixed(2))

        wayBill = invObj.waybillnum;
        orderNum = invObj.poordernum;
        date = new Date(invObj.invdate).toLocaleDateString();
        amountInWords = numToWords(Math.round(netAmount)).toUpperCase() + ' only';
        PODate = (new Date(invObj.poorderdate)).toLocaleDateString();
        wDate = (new Date(invObj.waybilldate)).toLocaleDateString();
        GSTN = '36AALCP8001G1ZG';
        MSME = 'UDYAM-TS-09-0012881';
        CGSTIN = invObj.bill.GSTIN;
        stateCode = invObj.bill.statecode;
        otherinfo = invObj.otherinfo

        /***************
         * Important information. DO NOT DO MISTAKE. WRITE CORRECT VALUE
         */
        policyNo = 'OG-21-1933-1005-00000012'
        insurance = 'Marine Cargo Insurance From Bajaj Allianz'
        bankName = 'IDFC FIRST BANK'
        acNum = '10068649634'
        ifsc = 'IDFB0080222'
        branch = 'HITECH CITY'
        /**************** */

        //data array
        var tBody = [
            [{
                image: `${image}`,
                fit: [70, 40],
                fillColor: 'white',
                alignment: 'center',
                border: [true, true, false, false],
                // colspan: 3
            }, {
                text: '',
                fillColor: 'white',

                border: [false, true, false, false]
            }, {
                text: '',
                fillColor: 'white',

                border: [false, true, false, false]
            }, {
                text: '',
                fillColor: 'white',

                border: [false, true, false, false]
            }, {
                image: `${image1}`,
                fit: [70, 40],
                fillColor: 'white',
                alignment: 'center',
                border: [false, true, false, false],
                colspan: 3
            }, {
                text: '',
                fillColor: 'white',

                border: [false, true, false, false]
            }, {
                text: '',
                fillColor: 'white',

                border: [false, true, false, false]
            }, {
                image: `${image2}`,
                fit: [70, 50],
                fillColor: 'white',
                alignment: 'center',
                border: [false, true, true, false],
                margin: [0, 0, 5, 0],
                colspan: 2
            }],
            [{
                text: `${compInfo}`,
                colSpan: 4,
                fillColor: '#ffffff',
                style: 'header',
                border: [true, true, true, false],
                alignment: 'left',
                rowSpan: 4
            }, '', '', '', {
                fillColor: '#ffffff',
                style: 'header',
                text: 'P.O.#:',
                border: [true, true, false, false],
                colSpan: 2,
                bold: true
            }, '', {
                fillColor: '#ffffff',
                style: 'header',
                text: `${orderNum}`,
                border: [false, true, true, false],
                colSpan: 2,
                bold: true
            }, ''],
            ['', '', '', '', {
                fillColor: '#ffffff',
                style: 'header',
                text: ``,
                colSpan: 2,
                border: [true, false, false, false],
                bold: false,
                alignment: 'left'
            }, '', {
                    fillColor: '#ffffff',
                    style: 'header',
                    text: ``,
                    colSpan: 2,
                    bold: false,
                    border: [false, false, true, false],
                    alignment: 'left'
                }, ''],
            ['', '', '', '', {
                fillColor: '#ffffff',
                style: 'header',
                text: 'P.O. Date:',
                bold: true,
                colSpan: 2,
                border: [true, false, false, false]
            }, '', {
                    fillColor: '#ffffff',
                    style: 'header',
                    text: `${PODate}`,
                    bold: true,
                    colSpan: 2,
                    border: [false, false, true, false]
                }, ''],
            ['', '', '', '', {
                fillColor: '#ffffff',
                style: 'header',
                text: ``,
                colSpan: 2,
                bold: false,
                border: [true, false, false, false],
                alignment: 'left'
            }, '', {
                    fillColor: '#ffffff',
                    style: 'header',
                    text: ``,
                    colSpan: 2,
                    bold: false,
                    border: [false, false, true, false],
                    alignment: 'left'
                }, ''],

            [{
                text: 'GST', style: 'header', fillColor: 'white',
                alignment: 'right',
                border: [true, false, false, false],
            }, {
                text: `${GSTN}`, style: 'header', fillColor: 'white',
                alignment: 'left',
                border: [false, false, true, false],
                colSpan: 3
            }, '', '', {
                fillColor: '#ffffff',
                style: 'header',
                text: 'Quote#:',
                bold: true,
                colSpan: 2,
                border: [true, false, false, false]
            }, '', {
                fillColor: '#ffffff',
                style: 'header',
                text: `${quotenum}`,
                bold: true,
                colSpan: 2,
                border: [false, false, true, false]
            }, ''],
            [{
                text: 'MSME', style: 'header', fillColor: 'white',
                alignment: 'right',
                border: [true, false, false, true],
            }, {
                text: `${MSME}`, style: 'header', fillColor: 'white',
                alignment: 'left',
                border: [false, false, true, true],
                colSpan: 3
            }, '', '', {
                fillColor: '#ffffff',
                style: 'header',
                text: 'Quote date:',
                colSpan: 2,
                bold: true,
                border: [true, false, false, true],
                alignment: 'left'
            }, '', {
                fillColor: '#ffffff',
                style: 'header',
                text: `${quotedate}`,
                colSpan: 2,
                bold: true,
                border: [false, false, true, true],
                alignment: 'left'
            }, ''],
            [{

                text: `Shipped to:\n${shippedTo}`,
                colSpan: 4,
                fillColor: '#ffffff',
                style: 'subheader',
            }
                , {}, {}, '',
            {

                colSpan: 4,
                fillColor: '#ffffff',
                style: 'subheader',
                stack: [
                    { text: `To:\n${billedTo}` },
                    { text: `GSTIN: ${CGSTIN}, State code: ${stateCode}`, style: 'subheader', fillColor: '#ffffff', },
                ]
            }, {}, {}, ''
            ],
            [
                {
                    text: 'Item',
                    style: 'tableHeader',
                    fillColor: '#111111',
                    colSpan: 3
                }, '', '',
                {
                    text: 'HSN', style: 'tableHeader',
                    fillColor: '#111111'
                }, {
                    text: 'Quantity',
                    style: 'tableHeader'
                }, { text: 'Tax', style: 'tableHeader' },
                {
                    text: 'Amount',

                    style: 'tableHeader'
                },
                {
                    text: 'Currency', style: 'tableHeader'
                }
            ]
        ]

        //Push return invoice data
        transformData(invObj).forEach((i) => {
            tBody.push(i)
        })

        //Push remaining in the body
        tBody.push([{
            rowSpan: 6,
            colSpan: 5,
            text: `Freight gst:${invObj.shipgst}\nFreight detail: ${freightDetail}\n\nAdditional information:\n\n${generateotherInfoText(otherinfo)}`,
            alignment: 'justify',
            style: 'subheader',
            bold: false
        }, '', '', '', '', {
            colSpan: 2,
            text: `Net amount`,
            alignment: 'justify',
            style: 'subheader',
            bold: false
        }, '', {
            text: `${grossAmount - freight}`,
            alignment: 'justify',
            style: 'tableContent',
        }],
            ['', '', '', '', '', {
                colSpan: 2,
                text: `Freight`,
                alignment: 'justify',
                style: 'subheader',
                bold: false
            }, '', {
                    text: `${freight}`,
                    alignment: 'justify',
                    style: 'tableContent',
                }],
            ['', '', '', '', '', {
                colSpan: 2,
                text: `Net tax`,
                alignment: 'justify',
                style: 'subheader',
                bold: false
            }, '', {
                    text: `${netTax}`,
                    alignment: 'justify',
                    style: 'tableContent',
                }],
            ['', '', '', '', '', {
                colSpan: 2,
                text: `Gross Amount`,
                alignment: 'justify',
                style: 'subheader',
                bold: false
            }, '', {
                    text: `${payAmount}`,
                    alignment: 'justify',
                    style: 'tableContent',
                }],
            ['', '', '', '', '', {
                colSpan: 3,
                text: `Rounding off`,
                alignment: 'justify',
                style: 'subheader',
                bold: false
            }, '', {
                    // text: `${totalDiscount}`,
                    // alignment: 'justify',
                    // style: 'tableContent',
                }],
            ['', '', '', '', '', {
                colSpan: 2,
                text: `Payable amount`,
                alignment: 'justify',
                style: 'subheader',
                bold: false
            }, '', {
                    text: `${netAmount}`,
                    alignment: 'justify',
                    style: 'tableContent',
                }],

            [{
                colSpan: 8,
                text: `Payable amount in words (Payable in ${invObj.currency}. Miscellaneous charges are extra. See additional information section.):\n${amountInWords}`,
                style: 'subheader',
                alignment: 'left',
                color: 'white',
                fillColor: 'black'
            }, '', '', '', '', '', '', ''],
            /*    [{
                    colSpan: 4,
                    text: `Policy No.: ${policyNo}\nInsurance: ${insurance}`,
                    style: 'subheader',
                    alignment: 'justify',
                    color: 'black',
                    fillColor: 'white'
                }, '', '', '', {
                    colSpan: 4,
                    text: `PRASI LABS bank detail:\nBank name:  ${bankName}\nA/C no:  ${acNum}\nIFSC Code:  ${ifsc}\nBranch:  ${branch}`,
                    style: 'subheader',
                    alignment: 'justify',
                    color: 'black',
                    fillColor: 'white'
                }, '', '', ''],*/
            [{
                colSpan: 5,
                stack: [{ text: 'Note:\n', }, {
                    ul: [
                        'It is a computer generated Purchase Order. Physical signature is not required.',
                        'This PO shows the actual price of the goods described.',
                        'Subject to Hyderabad (India) jurisdiction.'
                    ]
                }]
                ,
                border: [true, true, false, true],
                style: 'subheader'
            }, '', '', '', '', { text: 'For,\nPrasi Labs pvt ltd.', style: 'subheader', colSpan: 2 }, '',
            {
                qr: `${orderNum}\n${netAmount}\n${date}\nhttps://prasilabs.com/`,
                version: '10',
                eccLevel: 'H',
                fit: '100',
                border: [false, true, true, true],
                fillColor: 'white',
                alignment: 'center',
            }
            ]
        )

        //assiging content to the document
        docDefinition.content[0].table.body = tBody;

        //create doc
        ///  docDefinition.header.table.body[0][0].text = invObj.type + ' COPY'
        try {

            console.log('calling function to generate pdf')

            createPDFDocPO(invObj.poordernum.replace(/\//g, '_'), invObj.ordernum.replace(/\//g, '_'), invObj.type, docDefinition, options);

        } catch (error) {
            console.log(error)
        }
    } catch (error) {
        console.log(error)
    }
}

module.exports = createPdfPO;