
/*
Created by: Peeyush
Created on: July 22 2021
Email: j.k.peeyush@gmail.com
Description: various utility functions
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
//Function definition to generate Id
const crypto = require('crypto');
const path = require('path')
const fs = require('fs')

const baseid = (len) => {
    try {
        //Gets Next Sequance value from Database
        const id = (Date.now()) + crypto.randomInt(247);

        let maxLen = '1'
        let str = ''
        for (let i = 0; i < Math.floor(Math.log10(id)); i++) { maxLen = maxLen + '0' }

        return (Math.ceil(Math.log10(id)) >= len || id % parseInt(maxLen) == 0) ? id :
            (() => {
                for (let j = 0; j < (len - Math.floor(Math.log10(id))); j++) {
                    str = str + '0'
                }
                return (str + id)
            }
            )()
    } catch (error) {
        return error
    }
};
//update sequence number from the file
const seqNum = (en) => {
    let retVal = 0
    try {
        switch (en) {
            case 'PO':
                retVal = fs.readFileSync(path.resolve('config/', './porderseq'))
                fs.writeFileSync(path.resolve('config/', './porderseq'), (Number(retVal) + 1).toString())
                break;
            case 'IN':
                retVal = fs.readFileSync(path.resolve('config/', './invseq'))
                fs.writeFileSync(path.resolve('config/', './invseq'), (Number(retVal) + 1).toString())
                break
            default:
                console.debug('No suitable entity found. Returning 0')
                break;
        }
        console.log(retVal)
    }
    catch (err) {
        console.error(err)
    }

    return retVal;
}
//Generate customer Id
const generateCustomerId = (baseid, len, config) => {
    return (config['CUSTID'] + baseid(len))
};
//generate address id
const generateAddressId = (baseid, len, config) => {
    return (config['ADDRID'] + baseid(len))
};
//Generate Contact id
const generateContactId = (baseid, len, config) => {
    return (config['CONID'] + baseid(len))
};
//Generate Invoice number
const generateInvoiceId = (baseid, len, config) => {
    const dt = new Date()
    const initials = config['INVID'] + dt.getFullYear().toString().slice(-2) + "-" + (parseInt(dt.getFullYear().toString().slice(-2)) + 1) + '/'
    return (initials + '' + seqNum('IN'))
};
//Generate Order number
const generateOrderId = (baseid, len, config) => {
    const dt = new Date()
    const initials = config['ODRID'] + dt.getFullYear().toString().slice(-2) + "-" + (parseInt(dt.getFullYear().toString().slice(-2)) + 1) + '/P'
    return (initials + baseid(len))
};
//Generate Order number
const generatePOrderId = (baseid, len, config) => {
    const dt = new Date()
    const initials = config['POID'] + dt.getFullYear().toString().slice(-2) + "-" + (parseInt(dt.getFullYear().toString().slice(-2)) + 1) + '/'
   // return (initials + '' + seqNum('PO'))
   return (initials)
};
//Generate Order item id
const generateOrderItemId = (baseid, len, config) => {
    return (config['ODRITEMID'] + baseid(len))
};
//Generate Product id
const generateProductId = (baseid, len, config) => {
    return (config['PRODID'] + baseid(len))
};
//Generate Phone id
const generatePhoneId = (baseid, len, config) => {
    return (config['PHONEID'] + baseid(len))
};
//Generate Email id
const generateEmailId = (baseid, len, config) => {
    return (config['EMAILID'] + baseid(len))
};
//Generate Zip id
const generateZipId = (baseid, len, config) => {
    return (config['ZIPID'] + baseid(len))
};
//Generate Waybill id
const generateWaybillId = (baseid, len, config) => {
    return (config['WAYBILLID'] + baseid(len))
};

//Generate communication id
const generateCommId = (baseid, len, config) => {
    return (config['COMM'] + baseid(len))
};

//Generate service request id
const generateSrId = (baseid, len, config) => {
    return (config['SR'] + baseid(len))
};

//Generate user request id
const generateUserId = (baseid, len, config) => {
    return (config['USER'] + baseid(len))
};

exports.generateCustomerId = generateCustomerId
exports.generateInvoiceId = generateInvoiceId
exports.generateOrderId = generateOrderId
exports.generateAddressId = generateAddressId
exports.generateContactId = generateContactId
exports.generateOrderItemId = generateOrderItemId
exports.generateProductId = generateProductId
exports.generatePhoneId = generatePhoneId
exports.generateWaybillId = generateWaybillId
exports.generateZipId = generateZipId
exports.generateEmailId = generateEmailId
exports.baseid = baseid
exports.generateSrId = generateSrId
exports.generateCommId = generateCommId
exports.generateUserId = generateUserId
exports.generatePOrderId = generatePOrderId