
/*
Created by: Peeyush
Created on: July 16 2021
Email: j.k.peeyush@gmail.com
Description: Funtion to validate table constraints defined in configuration
Version: 1.0
Project: PRASI Labs
Security: Confidential
License: NA
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
//Function to compare two JSONobjects of same structure but different values. Returns unmatched fields in an array
const matchObj = (obj1, obj2, TRAVERSEJSON) => {
    let keylist1 = {}
    let keylist2 = {}
    let diff = []
    TRAVERSEJSON.traveres(obj1, '', keylist1);
    TRAVERSEJSON.traveres(obj2, '', keylist2);
    Object.keys(keylist2).forEach(key => {
        if (key != '.id' && key != '.created' && key != '.createdby' && key != '.updated' && key != '.updatedby' && keylist1[key] != keylist2[key]) {
            diff.push(key.split('.').reduce((oldval, newval) => {
                return newval + ' ' + oldval
            }))
        }
    })
    return diff
}
/*
Funtion to validate table constraints defined in rdbms.json.
            This function requires two json objects parameters in the following format:
            First parameter should be JSON object of configuraiton
                {"<table name1>":{"column1":["<NOTNULL/UNIQUE>"],...}...}
            Second parameter should be an array of JSON object of data passed to validate
                {"<table name>":{"column1":"<value>",...}}
            Funcition return a object in the format 
                { isFailed: <boolean>, error: "any valid error message"}
*/
const validateConstraint = (
    // JSON object that  needs validaion
    dataJson,
    // JSON object of constraint configuration file
    rdbmsJson,
    TRAVERSEJSON, element) => {
    'use strict';
    let val = { isFailed: false, error: '' }
    let traveresedRDBS = rdbmsJson;
    let traversedDataArray = [];

    traversedDataArray.push(dataJson);
    // console.log(traversedDataArray)
    //  console.log(dataJson)
    if (!traveresedRDBS instanceof Object || !traversedDataArray instanceof Object || !Array.isArray(traversedDataArray)) {
        val = { isFailed: true, error: 'Passed values are not valid.' }
    }
    else {
        let output = {}
        let dataOutput = []
        TRAVERSEJSON.traveres(rdbmsJson, '', output);
        let object_keys = Object.keys(output).filter(val => {
            return (val.indexOf(`.${element}.`) != -1)
        });
        //console.log(object_keys);
        for (let key = 0; key < object_keys.length; key++) {
            switch (output[object_keys[key]]) {
                case 'NOTNULL':
                    traversedDataArray.forEach(dataObj => {
                        dataOutput = []
                        //console.log(dataObj)
                        TRAVERSEJSON.traveresKey(dataObj, object_keys[key].split('.')[2], dataOutput);
                        if ((dataOutput[0] == 'KEY_NOT_FOUND' || !dataOutput[0])) {
                            val['isFailed'] = true
                            val['error'] = `${object_keys[key].split('.')[1]} -> ${object_keys[key].split('.')[2]} is mandatory`

                            return val;
                        }
                    })
                    break;
                case 'ISARRAY':
                    traversedDataArray.forEach(dataObj => {
                        dataOutput = []
                        TRAVERSEJSON.traveresKey(dataObj, object_keys[key].split('.')[2], dataOutput);

                        if (!Array.isArray(dataOutput[0])) {
                            val['isFailed'] = true
                            val['error'] = `${object_keys[key].split('.')[1]} -> ${object_keys[key].split('.')[2]} should be an array`
                            return val;
                        }
                    })
                    break;
                case 'ISMAP':
                    traversedDataArray.forEach(dataObj => {
                        dataOutput = []
                        TRAVERSEJSON.traveresKey(dataObj, key.split('.')[2], dataOutput);

                        if (!(dataOutput[0] instanceof Map)) {
                            val['isFailed'] = true
                            val['error'] = `${object_keys[key].split('.')[1]} -> ${object_keys[key].split('.')[2]} should be a map`
                            return val;
                        }
                    })
                    break;
                case 'ISSET':
                    traversedDataArray.forEach(dataObj => {
                        dataOutput = []
                        //console.log(dataObj)
                        TRAVERSEJSON.traveresKey(dataObj, key.split('.')[2], dataOutput);

                        if (!(dataOutput[0] instanceof Set)) {
                            val['isFailed'] = true
                            val['error'] = `${object_keys[key].split('.')[1]} -> ${object_keys[key].split('.')[2]} should be a set`
                            return val;
                        }
                    })
                    break;
                case 'ISLIST':
                    traversedDataArray.forEach(dataObj => {
                        dataOutput = []
                        //console.log(dataObj)
                        TRAVERSEJSON.traveresKey(dataObj, object_keys[key].split('.')[2], dataOutput);

                        if (!(dataOutput[0] instanceof List)) {
                            val['isFailed'] = true
                            val['error'] = `${object_keys[key].split('.')[1]} -> ${object_keys[key].split('.')[2]} should be a list`
                            return val;
                        }
                    })
                    break;
                case 'ISDATE':
                    traversedDataArray.forEach(dataObj => {
                        dataOutput = []
                        //console.log(dataObj)
                        TRAVERSEJSON.traveresKey(dataObj, object_keys[key].split('.')[2], dataOutput);

                        if (!(dataOutput[0] instanceof Date)) {
                            val['isFailed'] = true
                            val['error'] = `${object_keys[key].split('.')[1]} -> ${object_keys[key].split('.')[2]} should be a date`
                            return val;
                        }
                    })
                    break;

                case 'ISBLOB':
                    traversedDataArray.forEach(dataObj => {
                        dataOutput = []
                        //console.log(dataObj)
                        TRAVERSEJSON.traveresKey(dataObj, object_keys[key].split('.')[2], dataOutput);

                        if (!(dataOutput[0] instanceof Blob)) {
                            val['isFailed'] = true
                            val['error'] = `${object_keys[key].split('.')[1]} -> ${object_keys[key].split('.')[2]} should be a blob`
                            return val;
                        }
                    })
                    break;
                default:
                    val['isFailed'] = true
                    val['error'] = 'Validation is defined but not checked.'

            }

        }
    }
    return val;
}

//Function to check field - coulumn mapping
//It converts Field - value JSON to COlumn - value JSON 
//Returns JSON object
//valid only for JSON object in the format [{"KEY1":{"SUBKEY1":"","SUBKEY2":['a','1',..],..},"KEY2":{"SUBKEY1":"","SUBKEY2":['a','1',..],..},...}]
const validateMapping = (
    //pass the request data in field - value JSON objec
    jsonobjectArray,
    //pass the map file JSON object
    mapfileObject) => {
    'use strict';
    let returnJsonArray = []
    if (!jsonobjectArray instanceof Object || !mapfileObject instanceof Object || !Array.isArray(jsonobjectArray)) {
        return returnJsonArray
    }
    jsonobjectArray.forEach(jsonobject => {
        let returnJson = {}
        Object.getOwnPropertyNames(jsonobject).forEach(element => {
            if (jsonobject[element] instanceof Object && mapfileObject.hasOwnProperty(element)) {
                returnJson[element] = {}
                Object.getOwnPropertyNames(jsonobject[element]).forEach(key => {
                    if (mapfileObject[element].hasOwnProperty(key) && mapfileObject[element][key].trim()) {
                        returnJson[element][mapfileObject[element][key]] = jsonobject[element][key];

                    }
                })
            }
        })
        returnJsonArray.push(returnJson);

    });
    // console.log(returnJsonArray)
    return returnJsonArray;
}

//function to set avoid tombstone
const avoidTombstone = (cassandra, jsonObject) => {
    'use strict';
    if (jsonObject === undefined) {
        return null;
    }
    if (jsonObject instanceof Object) {
        if (Array.isArray(jsonObject)) {
            jsonObject.forEach(element => {
                if (element instanceof Object) {
                    avoidTombstone(cassandra, element)
                }
            })
        }
        else {
            Object.getOwnPropertyNames(jsonObject).forEach(key => {
                if (jsonObject[key] instanceof Object) {
                    avoidTombstone(cassandra, jsonObject[key])
                } else {
                    if (jsonObject[key]== undefined || jsonObject[key]==null) {
                        jsonObject[key] = cassandra.types.unset;
                    }
                }
            })
        }
    };
    return jsonObject;
}

//assign values to mandatory values
const assignMandatoryValues = (mapper, jsonObject) => {
    'use strict';
    if (jsonObject === undefined) {
        return null;
    }
    if (jsonObject instanceof Object) {
        if (Array.isArray(jsonObject)) {
            jsonObject.forEach(element => {
                if (element instanceof Object) {
                    assignMandatoryValues(mapper, element)
                }
            })
        }
        else {
            Object.getOwnPropertyNames(jsonObject).forEach(key => {
                if (jsonObject[key] instanceof Object) {
                    assignMandatoryValues(mapper, jsonObject[key])
                } else {

                    switch (key) {
                        case 'id':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.cassandra.types.Uuid.random();

                            }
                            break;
                        case 'customerid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateCustomerId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'addressid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateAddressId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'commid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateCommId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'contactid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateContactId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'eid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateEmailId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'orderid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateOrderId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'itemid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateOrderItemId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'pid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generatePhoneId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'prodid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateProductId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'srno':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateSrId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'userid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateUserId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                        case 'zid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateZipId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }

                            break;
                        case 'invid':
                            if (!(jsonObject[key] || jsonObject[key].trim())) {
                                jsonObject[key] = mapper.global.UTILITY.generateInvoiceId(mapper.global.UTILITY.baseid, mapper.global.KEY.LENGTH, mapper.global.KEY)

                            }
                            break;
                    }
                }
            })
        }
    };
    return jsonObject;
}
//check LOV
const checkLovValues = (mapper, jsonObject, context, lovstat) => {
    'use strict';
    if (jsonObject === undefined) {
        return null;
    }
    if (jsonObject instanceof Object) {
        if (Array.isArray(jsonObject)) {
            for (var element = 0; element < jsonObject.length; element++) {

                if (jsonObject[element] instanceof Object) {
                    checkLovValues(mapper, jsonObject[element], element, lovstat)
                }
            }
        }
        else {

            for (var key = 0; key < Object.keys(jsonObject).length; key++) {

                if (jsonObject[Object.keys(jsonObject)[key]] instanceof Object) {
                    checkLovValues(mapper, jsonObject[Object.keys(jsonObject)[key]], Object.keys(jsonObject)[key], lovstat)
                } else {
                    let audit_TYPE = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'audit_TYPE', '', audit_TYPE)

                    let audit_OPERATION = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'audit_OPERATION', '', audit_OPERATION)

                    let communication_TYPE = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'communication_TYPE', '', communication_TYPE)

                    let user_TYPE = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'user_TYPE', '', user_TYPE)

                    let order_STATUS = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'order_STATUS', '', order_STATUS)

                    let contact_TYPE = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'contact_TYPE', '', contact_TYPE)

                    let address_TYPE = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'audit_TYaddress_TYPEPE', '', address_TYPE)

                    let item_STATUS = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'item_STATUS', '', item_STATUS)

                    let account_TYPE = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'account_TYPE', '', account_TYPE)

                    let account_STATUS = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'account_STATUS', '', account_STATUS)

                    let STATUS = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'STATUS', '', STATUS)

                    let invoice_CANEDIT = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'invoice_CANEDIT', '', invoice_CANEDIT)

                    let invoice_STATUS = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'invoice_STATUS', '', invoice_STATUS)

                    let invoice_TYPE = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'invoice_TYPE', '', invoice_TYPE)

                    let anonymous_USER = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'anonymous_USER', '', anonymous_USER)

                    let sr_type = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'sr_type', '', sr_type)

                    let sr_status = {}
                    mapper.global.TRAVERSEJSON.traveresKeyPath(mapper.global.LOV, 'sr_status', '', sr_status)


                    switch (context) {
                        case 'Account':
                            if (Object.keys(jsonObject)[key] == 'type') {

                                if (!account_TYPE['account_TYPE'].filter(k => { return k == jsonObject[Object.keys(jsonObject)[key]] }).length) {
                                    lovstat['status'] = false
                                    lovstat['error'] = `Account type value is incorrect. It should be in ${account_TYPE['account_TYPE']}`

                                }
                            }
                            break;
                        case 'Contact':

                            break;
                        case 'Address':

                            break;
                        case 'Audit':

                            break;
                        case 'Communication':

                            break;
                        case 'Servicerequest':

                            break;
                        case 'Email':

                            break;
                        case 'Phone':

                            break;
                        case 'Zip':

                            break;
                        case 'Invoice':

                            break;
                        case 'User':
                            if (Object.keys(jsonObject)[key] == 'type') {

                                if (!user_TYPE['user_TYPE'].filter(k => { return k == jsonObject[Object.keys(jsonObject)[key]] }).length) {
                                    lovstat['status'] = false
                                    lovstat['error'] = `User type value is incorrect. It should be in ${user_TYPE['user_TYPE']}`

                                }
                            }
                            break;
                        case 'Orders':
                            break;
                        case 'Orderitem':

                            break;
                        case 'Key':

                            break;
                        case 'Lov':

                            break;
                        case 'Company':

                            break;
                        default:
                    }
                }
            }
        };
    }
}

//assign values to table object
const assignValues = (tableObject, jsonObject) => {
    'use strict';
    if (jsonObject === undefined || tableObject === undefined || !jsonObject instanceof Object || !tableObject instanceof Object || Array.isArray(jsonObject) || Array.isArray(tableObject)) {
        return {};
    }

    Object.getOwnPropertyNames(tableObject).forEach(key => {
        tableObject[key] = jsonObject[key]

    })
    // console.log(tableObject)
    return tableObject;
}

//remove columns values from the table object
const removeColumnsValues = (jsonObject) => {
    'use strict';
    let temp = {}
    if (jsonObject === undefined || !jsonObject instanceof Object || Array.isArray(jsonObject)) {
        return {};
    }
    Object.getOwnPropertyNames(jsonObject).forEach(key => {
        if (jsonObject[key] != undefined && jsonObject[key] != null)
            temp[key] = jsonObject[key]
    })
    return temp;
}


//Class definition to get configuration values from various json files
//at the moment 
class Settings {
    #incomingObj = {};
    #time = null;
    constructor(
        //JSON object of incoming data
        incomingObj,
    ) {
        this.#incomingObj = incomingObj;
        this.#time = new Date().toISOString();

    }

    #validtae(
        mapper,
        //key
        key) {
        //validate mapping
        let tempObj = mapper.global.VALIDATE.validateMapping(this.#incomingObj['data'], mapper.global.FIELDCOLMAP)
        //assign system column values
        //console.log(tempObj)
        for (var item = 0; item < tempObj.length; item++) {
            for (var element = 0; element < Object.keys(tempObj[item]).length; element++) {

                tempObj[item][Object.keys(tempObj[item])[element]]['updatedby'] = this.#incomingObj['System']['user']
                tempObj[item][Object.keys(tempObj[item])[element]]['updated'] = this.#time

                //assign all mandatory values
                tempObj[item][Object.keys(tempObj[item])[element]] = mapper.global.VALIDATE.assignMandatoryValues(mapper, tempObj[item][Object.keys(tempObj[item])[element]]);
                //transform into schema object
                //console.log(tempObj[item][Object.keys(tempObj[item])[element]])

                tempObj[item][Object.keys(tempObj[item])[element]] = mapper.global.VALIDATE.assignValues(mapper.global.Schema[Object.keys(tempObj[item])[element]], tempObj[item][Object.keys(tempObj[item])[element]])

                //check constraints
                let validateCheckObj = mapper.global.VALIDATE.validateConstraint(tempObj[item][Object.keys(tempObj[item])[element]], mapper.global.RDBMSJSON, mapper.global.TRAVERSEJSON, Object.keys(tempObj[item])[element])

                if (validateCheckObj['isFailed']) {

                    return { 'error': validateCheckObj['error'], 'data': [] }

                }

                //remove null columns to avoid Cassandra tombstone
                tempObj[item][Object.keys(tempObj[item])[element]] = mapper.global.VALIDATE.removeColumnsValues(mapper.global.Schema[Object.keys(tempObj[item])[element]], tempObj[item][Object.keys(tempObj[item])[element]])

            }
        }
        let l = {}
        mapper.global.VALIDATE.checkLovValues(mapper, tempObj, '', l)

        if (l.status !== undefined && !l.status) {
            return { 'error': l.error, 'data': [] }
        }

        switch (key) {
            case 'Account':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Account') && Object.keys(tempObj[key]['Account']) &&
                        //  tempObj[key].hasOwnProperty('Contact') && Object.keys(tempObj[key]['Contact']) &&
                        tempObj[key].hasOwnProperty('Email') && Object.keys(tempObj[key]['Email']) &&
                        //  tempObj[key].hasOwnProperty('Address') && Object.keys(tempObj[key]['Address']) &&
                        tempObj[key].hasOwnProperty('Zip') && Object.keys(tempObj[key]['Zip']) &&
                        tempObj[key].hasOwnProperty('Phone') && Object.keys(tempObj[key]['Phone'])
                    )) {
                        return { 'error': 'Mandatory account data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Contact':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Contact') && Object.keys(tempObj[key]['Contact']) &&
                        tempObj[key].hasOwnProperty('Email') && Object.keys(tempObj[key]['Email']) &&

                        tempObj[key].hasOwnProperty('Phone') && Object.keys(tempObj[key]['Phone'])
                    )) {
                        return { 'error': 'Mandatory contact data is missing.', 'data': [] }
                    }
                    if (tempObj[key].hasOwnProperty('Account')) {
                        if (!Object.keys(tempObj[key]['Account'])) {
                            return { 'error': 'When account detail is sent it can not be empty' }
                        }
                    }
                }
                break;
            case 'Address':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Address') && Object.keys(tempObj[key]['Address']) &&
                        tempObj[key].hasOwnProperty('Zip') && Object.keys(tempObj[key]['Zip'])

                    )) {
                        return { 'error': 'Mandatory address data is missing.', 'data': [] }
                    }
                    if (tempObj[key].hasOwnProperty('Account')) {
                        if (!Object.keys(tempObj[key]['Account'])) {
                            return { 'error': 'When account detail is sent it can not be empty' }
                        }
                    }
                }
                break;
            case 'Audit':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Audit') && Object.keys(tempObj[key]['Audit'])

                    )) {
                        return { 'error': 'Mandatory audit data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Communication':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Communication') && Object.keys(tempObj[key]['Communication'])
                    )) {
                        return { 'error': 'Mandatory communication data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Servicerequest':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Servicerequest') && Object.keys(tempObj[key]['Servicerequest'])
                    )) {
                        return { 'error': 'Mandatory service request data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Email':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(
                        tempObj[key].hasOwnProperty('Email') && Object.keys(tempObj[key]['Email'])
                    )) {
                        return { 'error': 'Mandatory email data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Phone':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(
                        tempObj[key].hasOwnProperty('Phone') && Object.keys(tempObj[key]['Phone'])
                    )) {
                        return { 'error': 'Mandatory phone data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Zip':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(
                        tempObj[key].hasOwnProperty('Zip') && Object.keys(tempObj[key]['Zip'])

                    )) {
                        return { 'error': 'Mandatory zip data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Invoice':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Invoice') && Object.keys(tempObj[key]['Invoice'])
                    )) {
                        return { 'error': 'Mandatory Invoice data is missing.', 'data': [] }
                    }
                }
                break;
            case 'User':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('User') && Object.keys(tempObj[key]['User'])
                    )) {
                        return { 'error': 'Mandatory user data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Key':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Key') && Object.keys(tempObj[key]['Key'])
                    )) {
                        return { 'error': 'Mandatory key item data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Lov':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Lov') && Object.keys(tempObj[key]['Lov'])
                    )) {
                        return { 'error': 'Mandatory lov item data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Company':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Company') && Object.keys(tempObj[key]['Company'])
                    )) {
                        return { 'error': 'Mandatory company item data is missing.', 'data': [] }
                    }
                }
                break;
                case 'Ledger':
                    for (let key = 0; key < Object.keys(tempObj).length; key++) {
    
                        if (!(tempObj[key].hasOwnProperty('Ledger') && Object.keys(tempObj[key]['Ledger'])
                        )) {
                            return { 'error': 'Mandatory ledger item data is missing.', 'data': [] }
                        }
                    }
                    break;
            case 'Orders':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Orders') && Object.keys(tempObj[key]['Orders'])
                    )) {
                        return { 'error': 'Mandatory order data is missing.', 'data': [] }
                    }
                }
                break;
            case 'Orderitem':
                for (let key = 0; key < Object.keys(tempObj).length; key++) {

                    if (!(tempObj[key].hasOwnProperty('Orderitem') && Object.keys(tempObj[key]['Orderitem'])
                    )) {
                        return { 'error': 'Mandatory order item data is missing.', 'data': [] }
                    }
                }
                break;
            default:
        }
        return { 'error': '', 'data': tempObj }
    }
    //get key value by passing intances of this calss on different keys.
    static getValidationStatus(
        mapper,
        //object of this instantiated class
        obj,
        //key to serch
        key) {
        let returnObject = {
            status: '',
            data: '',
            msg: ''
        }
        if (!(obj && key)) {
            returnObject.status = false
            returnObject.data = {}
            returnObject.msg = "required entity object and/or key is missing"
            return returnObject;
        }
        let v = obj.#validtae(mapper, key)

        v.data.length != 0 ? (
            returnObject.status = true,
            returnObject.msg = ''
        ) : (
            returnObject.status = false,
            returnObject.msg = v.error
        )
        returnObject.data = v.data;
        return returnObject
    }
}


exports.validateConstraint = validateConstraint
exports.validateMapping = validateMapping
exports.avoidTombstone = avoidTombstone
exports.assignMandatoryValues = assignMandatoryValues
exports.assignValues = assignValues
exports.removeColumnsValues = removeColumnsValues
exports.checkLovValues = checkLovValues
exports.matchObj = matchObj
exports.Settings = Settings

//check and return logged in user type, and accntid
exports.userTyepValiation = async (DBredirects, mapper, userid) => {
    let result = (await DBredirects.findData(mapper.mapper.forModel('User'),
        { userid: userid }, { fields: ["userid","accntid", "type"] }, {}))
        .first()
    let count = mapper.global.LOV.user_TYPE.filter(item => {
        return (item == result.type)
    }).length

    return (count == 0 ? undefined : result)
}

//Check order constraints
exports.orderValidation = (orderObject, usertype) => {
    //check order status for update possibility
    let result = false
    switch (usertype) {
        case 'GEN':
            if (orderObject.status == 'PENDING') {
                result = true
            } else {
                console.log('Order is freezed. Please contact us.')
            }
            break;
        case 'ADMIN':
            if (orderObject.status != 'COMPLETED' && orderObject.status != 'CANCELLED') {
                result = true
            }
            else {
                console.log('Order is freezed. Please contact the superuser.')
            }
            break;
        default:
            console.log('No valid user type found for order upsert operation.')
    }
    return result
}

//Check order item constraints
exports.orderItemValidation = (orderItemObject, usertype) => {
    //check order status for update possibility
    let result = false
    switch (orderItemObject.status) {
        case 'COMPLETED':
            if (usertype == 'ADMIN') {
                result = true
            }
            else {
                console.log('Order item is freezed. Please contact the superuser.')
            }
            break;
        case 'CANCELLED':
            if (usertype == 'ADMIN') {
                result = true
            }
            else {
                console.log('Order item is freezed. Please contact the superuser.')
            }
            break;
        default:
            result = true
    }
    return result
}